﻿//******************************************************************************************************
//  SignalToNoiseRatio.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  08/26/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using TimeSeriesFramework;
using TimeSeriesFramework.Adapters;
using TVA;
using TVA.PhasorProtocols;
using SynchrophasorAnalytics.DataConditioning;

namespace SynchrophasorAnalytics.Adapters
{
    /// <summary>
    /// Represents an adapter that calculates signal to noise ratio for magnitude and angle components of the phasor measurement.
    /// </summary>
    [Description("SNR Calculator : Calculates SNR for magnitude and angle components of the phasor measurement")]
    public class SignalToNoiseRatio : CalculatedMeasurementBase
    {
        #region [ Private Members ]

        private MeasurementKey[] m_inputMeasurementKeys;
        private string m_configurationPathName;
        private SignalToNoiseRatioIOConfiguration m_ioConfiguration;
        private Dictionary<string, SignalToNoiseRatioIOMapping> m_input;
        private Dictionary<string, SignalToNoiseRatioIOMapping> m_output;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The path name (including the file name) to the configuration file for the SNR I/O Mappings.
        /// </summary>
        [ConnectionStringParameter,
         Description("The path name (including the file name) to the configuration file for the SNR I/O Mappings."),
         DefaultValue("SNR.xml")]
        public string ConfigurationPathName
        {
            get
            {
                return m_configurationPathName;
            }
            set
            {
                m_configurationPathName = value;
            }
        }

        #endregion

        #region [ Class Methods ]

        /// <summary>
        /// The <see cref="PhasorMagnitudeSignalToNoiseRatio.Initialize"/> method initializes the Phasor Magnitude Signal-to-Noise Ratio Calculator application.
        /// </summary>
        /// <remarks>The <see cref="PhasorMagnitudeSignalToNoiseRatio.Initialize"/> method initializes the Phasor Magnitude Signal-to-Noise Ratio Calculator
        /// Application inside of the openPDC and initializes the frame counter for controlling the speed of the console output. It also creates a
        /// moving window for calculating the mean and standard deviation</remarks>
        public override void Initialize()
        {
            // Initialize the adapter in openPDC
            base.Initialize();

            // Initialize input & output dictionaries
            m_input = new Dictionary<string, SignalToNoiseRatioIOMapping>();
            m_output = new Dictionary<string, SignalToNoiseRatioIOMapping>();

            // Bring in the settings from the connection string
            Dictionary<string, string> settings = Settings;
            string setting;

            if (settings.TryGetValue("ConfigurationPathName", out setting))
            {
                m_configurationPathName = setting;
            }

            try
            {
                // Attempt to load the adapters configuration file 
                m_ioConfiguration = SignalToNoiseRatioIOConfiguration.DeserializeFromXml(m_configurationPathName);
                OnStatusMessage("Successfully read configuration...");

                // Put each of the IO Mappings into two dictionaries linked in one by the input measurement key and one by the output measurement key
                foreach (SignalToNoiseRatioIOMapping ioMapping in m_ioConfiguration.SignalMappings)
                {
                    m_input.Add(ioMapping.InputKey, ioMapping);
                    m_output.Add(ioMapping.OutputKey, ioMapping);
                }

                // Validate input measurements
                List<MeasurementKey> validInputMeasurementKeys = new List<MeasurementKey>();

                for (int i = 0; i < InputMeasurementKeys.Length; i++)
                {
                    if (m_input.ContainsKey(InputMeasurementKeys[i].ToString()))
                    {
                        validInputMeasurementKeys.Add(InputMeasurementKeys[i]);
                    }
                }

                if (validInputMeasurementKeys.Count == 0)
                {
                    throw new InvalidOperationException("No valid measurements were specified as inputs to the SNR calculator.");
                }

                InputMeasurementKeys = validInputMeasurementKeys.ToArray();

            }
            catch (Exception exception)
            {
                OnStatusMessage(exception.InnerException.ToString());
            }

            m_inputMeasurementKeys = InputMeasurementKeys;
        }

        /// <summary>
        /// The <see cref="PhasorMagnitudeSignalToNoiseRatio.PublishFrame"/> method is called when there are new measurements to send to the 
        /// Phasor Magnitude Signal-to-Noise Ratio Calculator Application.
        /// </summary>
        /// <remarks>This method calculate the signal-to-noise ratio in phase magnitude signals. It does this by managing a moving window for 
        /// each magnitude signal and calculating the mean and standard deviation of the portion of the signal in the moving window. It then 
        /// divides each mean by the corresponding standard deviation and converts the resulting value to decibels.</remarks>
        /// <param name="frame">An <see cref="IFrame"/> object containing the most recent frame of magnitude measurements.</param>
        /// <param name="index">An integer</param>
        protected override void PublishFrame(IFrame frame, int index)
        {
            // Remove each of the raw measurements from the frame
            foreach (IMeasurement measurement in frame.Measurements.Values)
            {
                SignalToNoiseRatioIOMapping signal = new SignalToNoiseRatioIOMapping();
                if (m_input.TryGetValue(measurement.Key.ToString(), out signal))
                {
                    signal.Data.Add(measurement.Value);
                    if (signal.SignalType == SignalToNoiseRatioMeasurementType.Angle)
                    {
                        UnwrapPhaseAngle(signal);
                    }
                }
            }

            // Remove the oldest set of referenced phase angles from the moving window
            RemoveOldestDataPoints();

            // Calculate SNR Values
            CalculateSignalToNoiseRatio();

            // Prepare to clone the output measurements. 
            IMeasurement[] outputMeasurements = OutputMeasurements;

            // Create a list of IMeasurement objects
            List<IMeasurement> output = new List<IMeasurement>();

            // Cycle through the defined output measurements
            foreach (IMeasurement measurement in outputMeasurements)
            {
                // Retrieve SNR from the dictionary based on its output measurement key
                SignalToNoiseRatioIOMapping signal = new SignalToNoiseRatioIOMapping();
                if (m_output.TryGetValue(measurement.Key.ToString(), out signal))
                {
                    output.Add(Measurement.Clone(measurement, signal.SNR, frame.Timestamp));
                }
                else
                {
                    output.Add(Measurement.Clone(measurement, 0, frame.Timestamp));
                }
            }

            // Output the calculated measurements
            OnNewMeasurements(output);
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Performs unwrapping of phase angles
        /// </summary>
        /// <param name="signal">The phase angle contained inside of the <see cref="SignalToNoiseRatioIOMapping"/> object.</param>
        private void UnwrapPhaseAngle(SignalToNoiseRatioIOMapping signal)
        {
            if (signal.Data.Count > 2)
            {
                if (Math.Abs(signal.Data[signal.Data.Count - 1] - signal.Data[signal.Data.Count - 2]) > 180)
                {
                    if (signal.Data[signal.Data.Count - 1] < signal.Data[signal.Data.Count - 2])
                    {
                        signal.CurrentKValue++;
                    }
                    else
                    {
                        signal.CurrentKValue--;
                    }
                    Console.WriteLine(signal.Data[signal.Data.Count - 1] + "," + signal.Data[signal.Data.Count - 2] + "," + signal.CurrentKValue);
                }
            }
            signal.UnwrappedData.Add(signal.Data[signal.Data.Count - 1] + 360 * signal.CurrentKValue);
        }

        /// <summary>
        /// Calculates the signal-to-noise ratio for each of the magnitudes 
        /// in the moving window
        /// </summary>
        private void CalculateSignalToNoiseRatio()
        {
            // The signal to noise ratio in decibels is 10 times the logarithm of the mean 
            // divided by the standard deviation of the signal
            foreach (SignalToNoiseRatioIOMapping signal in m_input.Values)
            {
                if (signal.SignalType == SignalToNoiseRatioMeasurementType.Magnitude)
                {
                    signal.SNR = 10 * Math.Log10(ArithmeticMeanOfSignal(signal.Data) / StandardDeviationOfSignal(signal.Data));
                }
                else if (signal.SignalType == SignalToNoiseRatioMeasurementType.Angle)
                {
                    List<double> derivativeOfSignal = DerivativeOfSignal(signal.UnwrappedData);
                    signal.SNR = 10 * Math.Log10(ArithmeticMeanOfSignal(derivativeOfSignal) / StandardDeviationOfSignal(derivativeOfSignal));
                }
            }
        }

        /// <summary>
        /// Calculates the delta (or difference) between adjacent data points in the set.
        /// </summary>
        /// <param name="signal">The moving window that contains the signal in question.</param>
        /// <returns>A double representing the 'derivative' of the signal.</returns>
        private List<double> DerivativeOfSignal(List<double> signal)
        {
            List<double> derivativeOfSignal = new List<double>();

            for (int i = 1; i < signal.Count; i++)
            {
                derivativeOfSignal.Add(Math.Abs(signal[i] - signal[i - 1]));
            }

            return derivativeOfSignal;
        }

        /// <summary>
        /// Calculates the arithmetic mean of a duration of the signal in the moving window
        /// </summary>
        /// <param name="signal">The moving window that contains the signal in question.</param>
        /// <returns>A double representing the arithmetic mean of the signal.</returns>
        private double ArithmeticMeanOfSignal(List<double> signal)
        {
            double summation = 0;

            // Add all of the points in the duration of the moving window
            foreach (double point in signal)
            {
                summation += point;
            }

            // Divide by the total number of data points
            return summation / signal.Count;
        }

        /// <summary>
        /// Calculates the standard deviation of the duration of the signal in the moving window.
        /// </summary>
        /// <param name="signal">The moving window that contains the signal in question.</param>
        /// <returns>A double representing the standard deviation of the signal.</returns>
        private double StandardDeviationOfSignal(List<double> signal)
        {
            List<double> residuals = new List<double>();

            // compute the arithmetic mean of the signal
            double mean = ArithmeticMeanOfSignal(signal);

            // compute the standard deviation of the signal
            foreach (double point in signal)
            {
                residuals.Add((point - mean) * (point - mean));
            }

            double standardDeviation = Math.Sqrt(ArithmeticMeanOfSignal(residuals));

            return standardDeviation;
        }

        /// <summary>
        /// Removes the oldest data point of each signal from the moving window
        /// </summary>
        private void RemoveOldestDataPoints()
        {
            foreach (SignalToNoiseRatioIOMapping signal in m_input.Values)
            {
                if (signal.Data.Count == m_ioConfiguration.WindowSize + 1)
                {
                    signal.Data.RemoveAt(m_ioConfiguration.WindowSize - 1);
                }
            }
        }

        #endregion
    }
}
