﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using Extreme.Mathematics.LinearAlgebra.Providers;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Measurements;
using SynchrophasorAnalytics.Matrices;

namespace SynchrophasorAnalytics.Calibration
{
    public class PrimaryCalibrator
    {
        private ComplexMatrix m_rawCurrentMeasurements;
        private ComplexMatrix m_rawVoltageMeasurements;
        private ComplexMatrix m_PerfectPTReference;
        private ComplexMatrix m_completeCalibrationMatrix;
        private ComplexMatrix m_measurementErrors;
        private ComplexMatrix m_DiComplete;
        private ComplexMatrix m_YaEComplete;
        private List<ComplexMatrix> m_Di = new List<ComplexMatrix>();
        private List<ComplexMatrix> m_YaE = new List<ComplexMatrix>();
        private List<Network> m_networkSnapshots = new List<Network>();

        public PrimaryCalibrator(List<Network> networkSnapshots)
        {
            m_networkSnapshots = networkSnapshots;

            foreach (Network network in networkSnapshots)
            {
                m_Di.Add(ComplexMatrix.CreateDiagonal(network.CurrentMeasurementVectorForPrimaryPhasorCalibration.GetColumn(0)));
                m_YaE.Add((new CalibrationMatrix(network)).Matrix * ComplexMatrix.CreateDiagonal(network.VoltageMeasurementVectorForPrimaryPhasorCalibration.GetColumn(0)));
            }
        }

        public void Calibrate()
        {
            Kernel.UseProvider(new ParallelX64LinearAlgebraProvider());
            CombineSamplesBeforeCalibration();
            RemoveColumnsFromPerfectPT(m_networkSnapshots[0]);
            m_completeCalibrationMatrix = MatrixCalculationExtensions.HorizontallyConcatenate(m_DiComplete, m_YaEComplete);
            m_measurementErrors = ((m_completeCalibrationMatrix.Transpose() * m_completeCalibrationMatrix).GetInverse()) * m_completeCalibrationMatrix.Transpose();
        }

        public void SerializeResultsToXml(string pathName)
        {
            List<CalibrationResults> calibrationResults = new List<CalibrationResults>();

            List<DoubleComplex> aPhaseResults = new List<DoubleComplex>();
            List<DoubleComplex> bPhaseResults = new List<DoubleComplex>();
            List<DoubleComplex> cPhaseResults = new List<DoubleComplex>();

            for (int i = 0; i < m_measurementErrors.RowCount; i++)
            {
                if (i % 3 == 0)
                {
                    aPhaseResults.Add(m_measurementErrors[i, 0]);
                }
                else if (i % 3 == 1)
                {
                    bPhaseResults.Add(m_measurementErrors[i, 0]);
                }
                else if (i % 3 == 2)
                {
                    cPhaseResults.Add(m_measurementErrors[i, 0]);
                }
            }

            for (int i = 0; i < cPhaseResults.Count; i++)
            {
                calibrationResults.Add(new CalibrationResults()
                {
                    MeasurementID = i,
                    MeasurementType = PhasorType.CurrentPhasor,
                    APhaseRCF = Math.Sqrt(aPhaseResults[i].Re * aPhaseResults[i].Re + aPhaseResults[i].Im * aPhaseResults[i].Im),
                    BPhaseRCF = Math.Sqrt(bPhaseResults[i].Re * bPhaseResults[i].Re + bPhaseResults[i].Im * bPhaseResults[i].Im),
                    CPhaseRCF = Math.Sqrt(cPhaseResults[i].Re * cPhaseResults[i].Re + cPhaseResults[i].Im * cPhaseResults[i].Im),
                    APhasePACF = Math.Atan2(aPhaseResults[i].Im, aPhaseResults[i].Re),
                    BPhasePACF = Math.Atan2(bPhaseResults[i].Im, bPhaseResults[i].Re),
                    CPhasePACF = Math.Atan2(cPhaseResults[i].Im, cPhaseResults[i].Re)
                });
            }

            try
            {
                // Create an XmlSerializer with the type of Network
                XmlSerializer serializer = new XmlSerializer(typeof(List<CalibrationResults>));

                // Open a connection to the file and path.
                TextWriter writer = new StreamWriter(pathName);

                // Serialize this instance of NetworkMeasurements
                serializer.Serialize(writer, calibrationResults);

                // Close the connection
                writer.Close();
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Serialize the Calibration Results to the Xml File: " + exception.ToString());
            }

        }

        private void CombineSamplesBeforeCalibration()
        {
            m_DiComplete = m_Di[0];
            m_YaEComplete = m_YaE[0];

            for (int i = 1; i < m_Di.Count; i++)
            {
                m_DiComplete = MatrixCalculationExtensions.VerticallyConcatenate(m_DiComplete, m_Di[i]);
                m_YaEComplete = MatrixCalculationExtensions.VerticallyConcatenate(m_YaEComplete, m_YaE[i]);
            }
        }

        private void RemoveColumnsFromPerfectPT(Network network)
        {
            m_PerfectPTReference = m_YaEComplete.GetSubmatrix(0, m_YaEComplete.RowCount - 1,
                                                              3 * (network.Model.ObservedBusses.IndexOf(network.PerfectPtBusForPrimaryPhasorCalibration) - 1),
                                                              3 * (network.Model.ObservedBusses.IndexOf(network.PerfectPtBusForPrimaryPhasorCalibration) - 1) + 2);
            m_YaEComplete = MatrixCalculationExtensions.RemoveColumn(m_YaEComplete, 3 * (network.Model.ObservedBusses.IndexOf(network.PerfectPtBusForPrimaryPhasorCalibration) - 1) + 2);
            m_YaEComplete = MatrixCalculationExtensions.RemoveColumn(m_YaEComplete, 3 * (network.Model.ObservedBusses.IndexOf(network.PerfectPtBusForPrimaryPhasorCalibration) - 1) + 1);
            m_YaEComplete = MatrixCalculationExtensions.RemoveColumn(m_YaEComplete, 3 * (network.Model.ObservedBusses.IndexOf(network.PerfectPtBusForPrimaryPhasorCalibration) - 1));
        }

        private List<CurrentPhasorGroup> DetermineActiveCurrentPhasorsForCalibration(Network network)
        {
            List<CurrentPhasorGroup> currentsForCalibration = new List<CurrentPhasorGroup>();

            // Find which phasor groups should not be included in the calibration
            foreach (CurrentPhasorGroup currentPhasorGroup in network.Model.ActiveCurrentPhasors)
            {
                if (currentPhasorGroup.PhaseA.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                    currentPhasorGroup.PhaseB.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                    currentPhasorGroup.PhaseC.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active)
                {
                    currentsForCalibration.Add(currentPhasorGroup);
                }
            }

            return currentsForCalibration;
        }

        private void MapRawCurrentMeasurements(List<List<CurrentPhasorGroup>> rawCurrentPhasorGroups)
        {
            m_rawCurrentMeasurements = ComplexMatrix.Create(3 * rawCurrentPhasorGroups[0].Count, rawCurrentPhasorGroups.Count, 0);

            for (int i = 0; i < rawCurrentPhasorGroups.Count(); i++)
            {
                for (int j = 0; j < rawCurrentPhasorGroups[0].Count(); i++)
                {
                    m_rawCurrentMeasurements[i, 3 * j] = rawCurrentPhasorGroups[i][3 * j].PhaseA.Measurement.PerUnitComplexPhasor;
                    m_rawCurrentMeasurements[i, 3 * j + 1] = rawCurrentPhasorGroups[i][3 * j + 1].PhaseB.Measurement.PerUnitComplexPhasor;
                    m_rawCurrentMeasurements[i, 3 * j + 2] = rawCurrentPhasorGroups[i][3 * j + 2].PhaseC.Measurement.PerUnitComplexPhasor;
                }
            }
        }
    }
}
