﻿//******************************************************************************************************
//  SignalToNoiseRatioIOConfiguration.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  08/26/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace SynchrophasorAnalytics.DataConditioning
{
    [Serializable()]
    public class SignalToNoiseRatioIOConfiguration
    {
        #region [ Private Members ]

        private List<SignalToNoiseRatioIOMapping> m_signalMappings;
        private int m_windowSize;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// A list of <see cref="SignalToNoiseRatioIOMapping"/> objects.
        /// </summary>
        [XmlArray("SignalToNoiseRatioIOMappings")]
        public List<SignalToNoiseRatioIOMapping> SignalMappings
        {
            get
            {
                return m_signalMappings;
            }
            set
            {
                m_signalMappings = value;
            }
        }

        /// <summary>
        /// The number of measurements to persist in a moving window for calculation of the SNR.
        /// </summary>
        [XmlAttribute("WindowSize")]
        public int WindowSize
        {
            get
            {
                return m_windowSize;
            }
            set
            {
                m_windowSize = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Default constructor for the <see cref="SignalToNoiseRatioIOConfiguration"/> class.
        /// </summary>
        public SignalToNoiseRatioIOConfiguration()
            :this(new List<SignalToNoiseRatioIOMapping>())
        {
        }

        /// <summary>
        /// A constructor for the <see cref="SignalToNoiseRatioIOConfiguration"/> class. Defaults to a 30 frame <see cref="WindowSize"/>.
        /// </summary>
        /// <param name="signalMappings">A list of <see cref="SignalToNoiseRatioIOMapping"/> objects.</param>
        public SignalToNoiseRatioIOConfiguration(List<SignalToNoiseRatioIOMapping> signalMappings)
            :this(signalMappings, 30)
        {
        }

        /// <summary>
        /// Designated constructor for the <see cref="SignalToNoiseRatioIOConfiguration"/> class.
        /// </summary>
        /// <param name="signalMappings">A list of <see cref="SignalToNoiseRatioIOMapping"/> objects.</param>
        /// <param name="windowSize">The number of measurements to persist in a moving window for calculation of the SNR.</param>
        public SignalToNoiseRatioIOConfiguration(List<SignalToNoiseRatioIOMapping> signalMappings, int windowSize)
        {
            m_signalMappings = signalMappings;
            m_windowSize = windowSize;
        }

        #endregion

        #region [ Xml Serialization/Deserialization methods ]

        /// <summary>
        /// Creates a new <see cref="SignalToNoiseRatioIOConfiguration"/> by deserializing the configuration file from the specified location.
        /// </summary>
        /// <param name="pathName">The location of the configuration file including the file name.</param>
        /// <returns>A new <see cref="SignalToNoiseRatioIOConfiguration"/> based on the configuration file.</returns>
        public static SignalToNoiseRatioIOConfiguration DeserializeFromXml(string pathName)
        {
            try
            {
                // Create an empy SignalToNoiseRatioIOConfiguration object reference.
                SignalToNoiseRatioIOConfiguration network = null;

                // Create an XmlSerializer with the type of SignalToNoiseRatioIOConfiguration.
                XmlSerializer deserializer = new XmlSerializer(typeof(SignalToNoiseRatioIOConfiguration));

                // Read the data in from the file.
                StreamReader reader = new StreamReader(pathName);

                // Cast the deserialized data as a SignalToNoiseRatioIOConfiguration object.
                network = (SignalToNoiseRatioIOConfiguration)deserializer.Deserialize(reader);

                // Close the connection.
                reader.Close();

                return network;
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Deserialize the SNR I/O from the Configuration File: " + exception.ToString());
            }
        }

        /// <summary>
        /// Serialize the <see cref="SignalToNoiseRatioIOConfiguration"/> to the specified file.
        /// </summary>
        /// <param name="pathName">The directory name included the file name of the desired location for Xml Serialization.</param>
        public void SerializeToXml(string pathName)
        {
            try
            {
                // Create an XmlSerializer with the type of SignalToNoiseRatioIOConfiguration
                XmlSerializer serializer = new XmlSerializer(typeof(SignalToNoiseRatioIOConfiguration));

                // Open a connection to the file and path.
                TextWriter writer = new StreamWriter(pathName);

                // Serialize this instance of I/O Configuration
                serializer.Serialize(writer, this);

                // Close the connection
                writer.Close();
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Serialize the SNR I/O from the Configuration File: " + exception.ToString());
            }
        }

        #endregion
    }
}
