﻿//******************************************************************************************************
//  SubstationGraph.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//  07/02/2013 - Kevin D. Jones
//       Fixed error in casting switching devices when resolving observed busses.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SynchrophasorAnalytics.Modeling;

namespace SynchrophasorAnalytics.Graph
{
    public class SubstationGraph
    {
        #region [ Private Members ]

        private int m_internalID;
        private List<Node> m_vertexSet;
        private List<SwitchingDeviceBase> m_edgeSet;
        private VertexAdjacencyList m_adjacencyList;

        #endregion 

        #region [ Properties ]

        public List<Node> VertexSet
        {
            get
            {
                return m_vertexSet;
            }
            set
            {
                m_vertexSet = value;
            }
        }

        public List<SwitchingDeviceBase> EdgeSet
        {
            get
            {
                return m_edgeSet;
            }
            set
            {
                m_edgeSet = value;
            }
        }

        public VertexAdjacencyList AdjacencyList
        {
            get
            {
                return m_adjacencyList;
            }
            set
            {
                m_adjacencyList = value;
            }
        }

        #endregion

        #region [ Constructors ]

        public SubstationGraph(Substation substation)
        {
            m_internalID = substation.InternalID;
            m_vertexSet = new List<Node>();
            m_edgeSet = new List<SwitchingDeviceBase>();

            // Build Vertex Set
            foreach (Node node in substation.Nodes)
            {
                m_vertexSet.Add(node);
            }

            // Build Edge Set
            foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
            {
                m_edgeSet.Add(circuitBreaker);
            }
            foreach (Switch circuitSwitch in substation.Switches)
            {
                m_edgeSet.Add(circuitSwitch);
            }

            InitializeAdjacencyList();
        }

        #endregion

        #region [ Public Methods ]

        public void ResolveDirectlyConnectedAdjacencies()
        {
            while (ResolvingDirectlyConnectedAdjacencies()) { }
        }

        public List<ObservedBus> ResolveToObservedBusses()
        {
            List<ObservedBus> observedBusses = new List<ObservedBus>();
            foreach (VertexAdjacencyRow vertexAdjacencyRow in m_adjacencyList.Rows)
            {
                List<Node> nodeCluster = new List<Node>();
                foreach (int nodeInternalID in vertexAdjacencyRow.Header.Vertices)
                {
                    nodeCluster.Add(m_vertexSet.Find(x => x.InternalID == nodeInternalID));
                }
                observedBusses.Add(new ObservedBus(m_internalID, nodeCluster));
            }
            return observedBusses;
        }

        public bool ResolvingDirectlyConnectedAdjacencies()
        {
            foreach (VertexAdjacencyRow row in m_adjacencyList.Rows)
            {
                foreach (VertexCluster adjacency in row.Adjacencies)
                {
                    SwitchingDeviceBase connectingEdge = ConnectingEdgeBetween(row.Header, adjacency);

                    
                    if (connectingEdge is CircuitBreaker)
                    {
                        CircuitBreaker circuitBreaker = (CircuitBreaker)connectingEdge;
                        if (circuitBreaker.IsClosed)
                        {
                            ConnectionEstablished(row.Header, adjacency);
                            return true;
                        }
                    }
                    else if (connectingEdge is Switch)
                    {
                        Switch circuitSwitch = (Switch)connectingEdge;
                        if (circuitSwitch.IsClosed)
                        {
                            ConnectionEstablished(row.Header, adjacency);
                            return true;
                        }
                    }
                }
            }

            return false;
        }

        public void ConnectionEstablished(VertexCluster fromVertexCluster, VertexCluster toVertexCluster)
        {
            List<int> fromVertices = new List<int>();
            List<int> toVertices = new List<int>();

            foreach (int vertex in fromVertexCluster.Vertices)
            {
                fromVertices.Add(vertex);
            }

            foreach (int vertex in toVertexCluster.Vertices)
            {
                toVertices.Add(vertex);
            }

            VertexCluster fromCluster = new VertexCluster(fromVertices);
            VertexCluster toCluster = new VertexCluster(toVertices);

            VertexAdjacencyRow source = m_adjacencyList.RowWithHeader(fromVertexCluster);
            VertexAdjacencyRow target = m_adjacencyList.RowWithHeader(toVertexCluster);

            // Merge the two rows into one
            source.MergeWith(target);

            // Remove the old from the list
            m_adjacencyList.RemoveRow(target);

            // Update the vertices in the rest of the table
            foreach (VertexAdjacencyRow row in m_adjacencyList.Rows)
            {
                foreach (VertexCluster vertexCluster in row.Adjacencies)
                {
                    if (vertexCluster.Equals(fromCluster) || vertexCluster.Equals(toCluster))
                    {
                        vertexCluster.Vertices = source.Header.Vertices;
                    }
                }
                row.RemoveDuplicateVertexClusters();
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="fromVertexCluster"></param>
        /// <param name="toVertexCluster"></param>
        /// <returns></returns>
        public SwitchingDeviceBase ConnectingEdgeBetween(VertexCluster fromVertexCluster, VertexCluster toVertexCluster)
        {
            foreach (SwitchingDeviceBase switchingDevice in m_edgeSet)
            {
                foreach (int fromVertex in fromVertexCluster.Vertices)
                {
                    foreach (int toVertex in toVertexCluster.Vertices)
                    {
                        if (switchingDevice.FromNode.InternalID == fromVertex && switchingDevice.ToNode.InternalID == toVertex)
                        {
                            return switchingDevice;
                        }
                        else if (switchingDevice.FromNode.InternalID == toVertex && switchingDevice.ToNode.InternalID == fromVertex)
                        {
                            return switchingDevice;
                        }
                    }
                }
            }

            return null;
        }

        //private SwitchingDeviceBase ConnectingEdgeBetween(int fromNodeID, int toNodeID)
        //{
        //    foreach (SwitchingDeviceBase switchingDevice in m_edgeSet)
        //    {
        //        if (switchingDevice.FromNode.InternalID == fromNodeID && switchingDevice.ToNode.InternalID == toNodeID)
        //        {
        //            return switchingDevice;
        //        }
        //        else if (switchingDevice.FromNode.InternalID == toNodeID && switchingDevice.ToNode.InternalID == fromNodeID)
        //        {
        //            return switchingDevice;
        //        }
        //        else
        //        {
        //            return null;
        //        }
        //    }

        //    return null;
        //}

        #endregion

        #region [ Private Methods ]

        private void InitializeAdjacencyList()
        {
            m_adjacencyList = new VertexAdjacencyList();

            List<VertexCluster> vertexClusters = new List<VertexCluster>();
            List<VertexAdjacencyRow> adjacencyRows = new List<VertexAdjacencyRow>();

            foreach (Node node in m_vertexSet)
            {
                VertexCluster vertexCluster = new VertexCluster(node.InternalID);
                List<VertexCluster> vertexAdjacencies = new List<VertexCluster>();

                foreach (SwitchingDeviceBase switchingDevice in m_edgeSet)
                {
                    if (node.InternalID == switchingDevice.FromNode.InternalID)
                    {
                        vertexAdjacencies.Add(new VertexCluster(switchingDevice.ToNode.InternalID));
                    }
                    else if (node.InternalID == switchingDevice.ToNode.InternalID)
                    {
                        vertexAdjacencies.Add(new VertexCluster(switchingDevice.FromNode.InternalID));
                    }
                }

                m_adjacencyList.Rows.Add(new VertexAdjacencyRow(vertexCluster, vertexAdjacencies));
            }
        }

        #endregion
    }
}
