﻿//******************************************************************************************************
//  TransmissionLineGraph.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using SynchrophasorAnalytics.Modeling;

namespace SynchrophasorAnalytics.Graph
{
    public class TransmissionLineGraph
    {
        #region [ Private Members ]

        private List<Node> m_vertexSet;
        private List<ITwoTerminal> m_edgeSet;
        private VertexAdjacencyList m_adjacencyList;
        private Tree m_treeWithFromNodeRoot;
        private Tree m_treeWithToNodeRoot;

        #endregion 

        #region [ Properties ]

        public Tree TreeWithFromNodeRoot
        {
            get
            {
                return m_treeWithFromNodeRoot;
            }
            set
            {
                m_treeWithFromNodeRoot = value;
            }
        }

        public Tree TreeWithToNodeRoot
        {
            get
            {
                return m_treeWithToNodeRoot;
            }
            set
            {
                m_treeWithToNodeRoot = value;
            }
        }

        public List<Node> VertexSet
        {
            get
            {
                return m_vertexSet;
            }
            set
            {
                m_vertexSet = value;
            }
        }

        public List<ITwoTerminal> EdgeSet
        {
            get
            {
                return m_edgeSet;
            }
            set
            {
                m_edgeSet = value;
            }
        }

        public VertexAdjacencyList AdjacencyList
        {
            get
            {
                return m_adjacencyList;
            }
            set
            {
                m_adjacencyList = value;
            }
        }

        #endregion

        #region [ Constructors ]

        public TransmissionLineGraph()
            :this(new List<Node>(), new List<ITwoTerminal>())
        {
        }

        public TransmissionLineGraph(List<Node> vertexSet, List<ITwoTerminal> edgeSet)
        {
            m_vertexSet = vertexSet;
            m_edgeSet = edgeSet;
            
            InitializeAdjacencyList();
            
        }

        public TransmissionLineGraph(TransmissionLine transmissionLine)
        {
            m_vertexSet = new List<Node>();
            m_edgeSet = new List<ITwoTerminal>();

            m_vertexSet.Add(transmissionLine.FromNode);
            m_vertexSet.Add(transmissionLine.ToNode);
            foreach (Node node in transmissionLine.Nodes)
            {
                m_vertexSet.Add(node);
            }

            foreach (LineSegment lineSegment in transmissionLine.LineSegments)
            {
                m_edgeSet.Add(lineSegment);
            }
            foreach (Switch lineSwitch in transmissionLine.Switches)
            {
                m_edgeSet.Add(lineSwitch);
            }
        }

        #endregion

        #region [ Private Methods ]

        private void InitializeAdjacencyList()
        {
            m_adjacencyList = new VertexAdjacencyList();

            List<VertexCluster> vertexClusters = new List<VertexCluster>();
            List<VertexAdjacencyRow> adjacencyRows = new List<VertexAdjacencyRow>();

            foreach (Node node in m_vertexSet)
            {
                VertexCluster vertexCluster = new VertexCluster(node.InternalID);
                List<VertexCluster> vertexAdjacencies = new List<VertexCluster>();

                foreach (ITwoTerminal twoTerminalConnection in m_edgeSet)
                {
                    if (node.InternalID == twoTerminalConnection.FromNode.InternalID)
                    {
                        vertexAdjacencies.Add(new VertexCluster(twoTerminalConnection.ToNode.InternalID));
                    }
                    else if (node.InternalID == twoTerminalConnection.ToNode.InternalID)
                    {
                        vertexAdjacencies.Add(new VertexCluster(twoTerminalConnection.FromNode.InternalID));
                    }
                }

                m_adjacencyList.Rows.Add(new VertexAdjacencyRow(vertexCluster, vertexAdjacencies));
            }
        }

        private VertexCluster SearchForDeepestForkFrom(VertexCluster vertexCluster)
        {
            VertexAdjacencyRow adjacencyRow = m_adjacencyList.RowWithHeader(vertexCluster);
            if (adjacencyRow.Adjacencies.Count == 1)
            {
                return SearchForDeepestForkFrom(adjacencyRow.Adjacencies[0]);
            }
            else if (adjacencyRow.Adjacencies.Count > 1)
            {
                return adjacencyRow.Header;
            }
            else
            {
                return null;
            }

        }
        

        #endregion

        public void ResolveDirectlyConnectedAdjacencies()
        {
            while (ResolvingDirectlyConnectedAdjacencies())
            {
            }
        }

        public bool ResolvingDirectlyConnectedAdjacencies()
        {
            foreach (VertexAdjacencyRow row in m_adjacencyList.Rows)
            {
                foreach (VertexCluster adjacency in row.Adjacencies)
                {
                    ITwoTerminal connectingEdge = ConnectingEdgeBetween(row.Header, adjacency);

                    if ((Switch)connectingEdge != null)
                    {
                        Switch circuitSwitch = (Switch)connectingEdge;
                        if (circuitSwitch.IsClosed)
                        {
                            ConnectionEstablished(row.Header, adjacency);
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="fromVertexCluster"></param>
        /// <param name="toVertexCluster"></param>
        /// <returns></returns>
        public SwitchingDeviceBase ConnectingEdgeBetween(VertexCluster fromVertexCluster, VertexCluster toVertexCluster)
        {
            foreach (SwitchingDeviceBase switchingDevice in m_edgeSet)
            {
                foreach (int fromVertex in fromVertexCluster.Vertices)
                {
                    foreach (int toVertex in toVertexCluster.Vertices)
                    {
                        if (switchingDevice.FromNode.InternalID == fromVertex && switchingDevice.ToNode.InternalID == toVertex)
                        {
                            return switchingDevice;
                        }
                        else if (switchingDevice.FromNode.InternalID == toVertex && switchingDevice.ToNode.InternalID == fromVertex)
                        {
                            return switchingDevice;
                        }
                    }
                }
            }

            return null;
        }

        public void ConnectionEstablished(VertexCluster fromVertexCluster, VertexCluster toVertexCluster)
        {
            List<int> fromVertices = new List<int>();
            List<int> toVertices = new List<int>();

            foreach (int vertex in fromVertexCluster.Vertices)
            {
                fromVertices.Add(vertex);
            }

            foreach (int vertex in toVertexCluster.Vertices)
            {
                toVertices.Add(vertex);
            }

            VertexCluster fromCluster = new VertexCluster(fromVertices);
            VertexCluster toCluster = new VertexCluster(toVertices);

            VertexAdjacencyRow source = m_adjacencyList.RowWithHeader(fromVertexCluster);
            VertexAdjacencyRow target = m_adjacencyList.RowWithHeader(toVertexCluster);

            // Merge the two rows into one
            source.MergeWith(target);

            // Remove the old from the list
            m_adjacencyList.RemoveRow(target);

            // Update the vertices in the rest of the table
            foreach (VertexAdjacencyRow row in m_adjacencyList.Rows)
            {
                foreach (VertexCluster vertexCluster in row.Adjacencies)
                {
                    if (vertexCluster.Equals(fromCluster) || vertexCluster.Equals(toCluster))
                    {
                        vertexCluster.Vertices = source.Header.Vertices;
                    }
                }
                row.RemoveDuplicateVertexClusters();
            }
        }

    }
}
