﻿//******************************************************************************************************
//  VertexAdjacencyRows.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SynchrophasorAnalytics.Graph
{
    public class VertexAdjacencyRow
    {
        private VertexCluster m_rowHeader;
        private List<VertexCluster> m_adjacencies;

        public VertexCluster Header
        {
            get
            {
                return m_rowHeader;
            }
            set
            {
                m_rowHeader = value;
            }
        }

        public List<VertexCluster> Adjacencies
        {
            get
            {
                return m_adjacencies;
            }
            set
            {
                m_adjacencies = value;
            }
        }

        public VertexAdjacencyRow()
            :this(new VertexCluster(), new List<VertexCluster>())
        {
        }

        public VertexAdjacencyRow(VertexCluster rowHeader, List<VertexCluster> adjacencies)
        {
            m_rowHeader = rowHeader;
            m_adjacencies = adjacencies;
        }

        public override string ToString()
        {
            string header = m_rowHeader.ToString();
            string adjacencies = "";

            foreach (VertexCluster cluster in m_adjacencies)
            {
                adjacencies += cluster.ToString() + ",";
            }

            if (adjacencies.Length > 1)
            {
                adjacencies = adjacencies.Substring(0, adjacencies.Length - 1);
                return header + " | " + adjacencies;
            }
            else
            {
                return header + " | (null)";
            }
        }

        public void MergeWith(VertexAdjacencyRow adjacencyRow)
        {
            // Remove the headers of the merging rows from their respective adjacency lists.
            this.RemoveVertex(adjacencyRow.Header);
            adjacencyRow.RemoveVertex(this.Header);

            // OR-Wise Merging of Adjacencies
            foreach (VertexCluster adjacency in adjacencyRow.Adjacencies)
            {
                if (!this.ContainsAdjacency(adjacency))
                {
                    m_adjacencies.Add(adjacency);
                }
            }

            m_rowHeader.MergeWith(adjacencyRow.Header);
        }

        public VertexAdjacencyRow MergeToFormNewRow(VertexAdjacencyRow adjacencyRow)
        {
            this.MergeWith(adjacencyRow);
            return this;
        }

        public void RemoveDuplicateVertices()
        {
            m_rowHeader.RemoveDuplicateVertices();

            foreach (VertexCluster cluster in m_adjacencies)
            {
                cluster.RemoveDuplicateVertices();
            }
        }

        public void RemoveDuplicateVertexClusters()
        {
            VertexAdjacencyRow temporaryRow = new VertexAdjacencyRow();

            foreach (VertexCluster vertexCluster in m_adjacencies)
            {
                if (!temporaryRow.ContainsAdjacency(vertexCluster))
                {
                    temporaryRow.Adjacencies.Add(vertexCluster);
                }
            }

            m_adjacencies = temporaryRow.Adjacencies;
        }

        public void RemoveVertex(VertexCluster vertexCluster)
        {
            List<VertexCluster> updatedVertices = new List<VertexCluster>();

            foreach (VertexCluster adjacency in m_adjacencies)
            {
                if (!adjacency.Equals(vertexCluster))
                {
                    updatedVertices.Add(adjacency);
                }
            }

            m_adjacencies = updatedVertices;
        }

        public bool ContainsAdjacency(VertexCluster vertexCluster)
        {
            foreach (VertexCluster adjacency in m_adjacencies)
            {
                if (adjacency.Equals(vertexCluster))
                {
                    return true;
                }
            }
            return false;
        }

        public void Sort()
        {
            m_rowHeader.Sort();

            foreach (VertexCluster cluster in m_adjacencies)
            {
                cluster.Sort();
            }
        }
    }
}
