﻿//******************************************************************************************************
//  CovarianceMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using Extreme.Mathematics.LinearAlgebra.Providers;

namespace SynchrophasorAnalytics.Matrices
{
    public class CovarianceMatrix
    {
        #region [ Private Members ]

        private ComplexMatrix covarianceMatrix;
        private ComplexMatrix upperRightCovarianceMatrix;
        private ComplexMatrix lowerLeftCovarianceMatrix;
        private ComplexMatrix upperLeftCovarianceMatrix;
        private ComplexMatrix lowerRightCovarianceMatrix;

        #endregion

        #region [ Properties ]

        public ComplexMatrix Matrix
        {
            get 
            { 
                return covarianceMatrix; 
            }
        }

        #endregion

        #region [ Constructors ]

        public CovarianceMatrix(Network network)
        {
            BuildMatrices(network);
        }


        #endregion

        #region [ Private Methods ]

        private void BuildMatrices(Network network)
        {
            if (network.Model.PhaseConfiguration == PhaseSelection.PositiveSequence)
            {
                BuildPositiveSequenceMatrix(network);
            }
            else if (network.Model.PhaseConfiguration == PhaseSelection.ThreePhase)
            {
                BuildThreePhaseMatrix(network);
            }
        }

        private void BuildPositiveSequenceMatrix(Network network)
        {
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                    }
                }
            }

            int measuredNodesCount = measuredNodes.Count;
            int measuredCurrentsCount = measuredCurrents.Count;
            covarianceMatrix = ComplexMatrix.Create(measuredNodesCount + measuredCurrentsCount, measuredNodesCount + measuredCurrentsCount);

            for (int i = 0; i < covarianceMatrix.RowCount; i++)
            {
                if (i < measuredNodesCount)
                {
                    covarianceMatrix[i, i] = new DoubleComplex(measuredNodes[i].Voltage.PositiveSequence.Measurement.MeasurementCovariance, 0);
                }
                else
                {
                    covarianceMatrix[i, i] = new DoubleComplex(measuredCurrents[i - measuredNodesCount].PositiveSequence.Measurement.MeasurementCovariance, 0);
                }
            }
        }

        private void BuildThreePhaseMatrix(Network network)
        {
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                    }
                }
            }

            int measuredNodesCount = measuredNodes.Count;
            int measuredCurrentsCount = measuredCurrents.Count;
            covarianceMatrix = ComplexMatrix.Create(3 * (measuredNodesCount + measuredCurrentsCount),3 * (measuredNodesCount + measuredCurrentsCount));

            for (int i = 0; i < measuredNodesCount + measuredCurrentsCount; i++)
            {
                if (i < measuredNodesCount)
                {
                    covarianceMatrix[3 * i, 3 * i] = new DoubleComplex(measuredNodes[i].Voltage.PhaseA.Measurement.MeasurementCovariance, 0);
                    covarianceMatrix[3 * i + 1, 3 * i + 1] = new DoubleComplex(measuredNodes[i].Voltage.PhaseB.Measurement.MeasurementCovariance, 0);
                    covarianceMatrix[3 * i + 2, 3 * i + 2] = new DoubleComplex(measuredNodes[i].Voltage.PhaseC.Measurement.MeasurementCovariance, 0);
                }
                else
                {
                    covarianceMatrix[3 * i, 3 * i] = new DoubleComplex(measuredCurrents[i - measuredNodesCount].PhaseA.Measurement.MeasurementCovariance, 0);
                    covarianceMatrix[3 * i + 1, 3 * i + 1] = new DoubleComplex(measuredCurrents[i - measuredNodesCount].PhaseB.Measurement.MeasurementCovariance, 0);
                    covarianceMatrix[3 * i + 2, 3 * i + 2] = new DoubleComplex(measuredCurrents[i - measuredNodesCount].PhaseC.Measurement.MeasurementCovariance, 0);
                }
            }
        }

        #endregion

        #region [ Public Methods ]

        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            for (int i = 0; i < covarianceMatrix.RowCount; i++)
            {
                string csvRow = "";
                for (int j = 0; j < covarianceMatrix.ColumnCount; j++)
                {
                    if (j < covarianceMatrix.ColumnCount - 1)
                    {
                        csvRow += covarianceMatrix[i, j].Re.ToString() + ",";
                    }
                    else
                    {
                        csvRow += covarianceMatrix[i, j].Re.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\Covariance Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".csv"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
