﻿//******************************************************************************************************
//  CurrentMeasurementBusIncidenceMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// The <see cref="CurrentMeasurementBusIncidenceMatrix"/> is used to create an instance of the current 
    /// measurement-bus incidence matrix for the specified power system. The matrix can be used for either 
    /// a single phase system or a three phase system by specifying the <see cref="PhaseSelection"/> 
    /// to be either <see cref="PhaseSelection.PositiveSequence"/> or 
    /// <see cref="PhaseSelection.ThreePhase"/>.
    /// </summary>
    /// <remarks>
    /// <para>The current measurement-bus incidence matrix is a matrix that shows the location of the current 
    /// flow measurements in the network. It is an <i><b>m</b></i> by <i><b>b</b></i> sized matrix
    /// where <i><b>m</b></i> is the number of current measurements in the network and <i><b>b</b></i> is the 
    /// number of busses which have a current measurement leaving the bus. It is populated using
    /// a few simple rules. These include:</para>
    /// <list type="number">
    /// <item><description>Each row of the matrix corresponds to a current measurement in the 
    /// network.</description></item>
    /// <item><description>Each column of the matrix corresponds to a bus in the system which has a current 
    /// measurement leaving the bus.</description></item>
    /// <item><description>If measurement <i><b>X</b></i> (corresponding to row <i><b>X</b></i>) leaves 
    /// bus <i><b>Y</b></i> (corresponding to column <i><b>Y</b></i>) then the matrix element 
    /// (<i><b>X</b></i>,<i><b>Y</b></i>) will be a 1.</description></item>
    /// <item><description>If measurement <i><b>X</b></i> (corresponding to row <i><b>X</b></i>) leaves 
    /// bus <i><b>Y</b></i> heading towards bus <i><b>Z</b></i> (corresponding to column <i><b>Z</b></i>) then 
    /// the matrix element (<i><b>X</b></i>,<i><b>Z</b></i>) will be a -1.</description></item>
    /// <item><description>All remaining entries will be identically zero.</description></item>
    /// </list>
    /// <para>These rules apply for both single phase and three phase systems. In three phase systems the 
    /// elements described in the above rules are replaced by 3 by 3 blocks that are either positive identity, 
    /// negative identity or zero.</para>
    /// </remarks>
    /// <seealso cref="VoltageMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="SeriesAdmittanceMatrix"/>
    /// <seealso cref="ShuntSusceptanceMatrix"/>
    /// <seealso cref="CovarianceMatrix"/>
    /// <seealso cref="SystemMatrix"/>
    public class CurrentMeasurementBusIncidenceMatrix
    {
        #region [ Private Members ]

        private ComplexMatrix m_A = ComplexMatrix.Create(0, 0, 0);
        private ComplexMatrix m_exhaustiveA = ComplexMatrix.Create(0, 0, 0);
        private List<string> m_csvRowHeaders;
        private List<string> m_csvColumnHeaders;
        private PhaseSelection m_phaseSelection;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The current measurement-bus incidence matrix
        /// </summary>
        public ComplexMatrix Matrix
        {
            get 
            { 
                return m_A; 
            }
        }

        /// <summary>
        /// The current measurement-bus incidence matrix assuming current phasors on each end of each branch
        /// </summary>
        public ComplexMatrix ExhaustiveMatrix
        {
            get 
            { 
                return m_exhaustiveA; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// The designated constructor for the <see cref="CurrentMeasurementBusIncidenceMatrix"/> class.
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/> model.</param>
        public CurrentMeasurementBusIncidenceMatrix(Network network)
        {
            BuildMatrices(network);
            m_phaseSelection = network.Model.PhaseConfiguration;
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Method that decides whether to build the <see cref="PhaseSelection.PositiveSequence"/> matrices or
        /// the <see cref="PhaseSelection.ThreePhase"/> matrices;
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/>.</param>
        private void BuildMatrices(Network network)
        {
            if (network.Model.PhaseConfiguration == PhaseSelection.PositiveSequence)
            {
                BuildPositiveSequenceMatrix(network);
                //BuildExhaustivePositiveSequenceMatrix(network);
            }
            else if (network.Model.PhaseConfiguration == PhaseSelection.ThreePhase)
            {
                BuildThreePhaseMatrix(network);
                //BuildExhaustiveThreePhaseMatrix(network);
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.PositiveSequence"/> version of the <see cref="CurrentMeasurementBusMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildPositiveSequenceMatrix(Network network)
        {
            m_csvRowHeaders = new List<string>();
            m_csvColumnHeaders = new List<string>();
            m_csvColumnHeaders.Add("A Matrix");

            List<ObservedBus> observedBusses = network.Model.ObservedBusses;
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            // Create an empty current measurement-bus incidence matrix rows equal in number to the number of current phasors and
            // columns equal in number to the number of observed busses in the network.
            m_A = ComplexMatrix.Create(measuredCurrents.Count, observedBusses.Count, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                m_csvRowHeaders.Add(measuredCurrent.Description);
                foreach (ObservedBus observedBus in observedBusses)
                {
                    if (measuredCurrents.IndexOf(measuredCurrent) == measuredCurrents.Count - 1)
                    {
                        m_csvColumnHeaders.Add("Sub: " + observedBus.InternalID.ToString());
                    }
                    foreach (Node node in observedBus.Nodes)
                    {
                        if (node.InternalID == (measuredCurrent.MeasuredFromNodeID))
                        {
                            // If measurement X leaves bus Y its element should be 1
                            m_A[measuredCurrents.IndexOf(measuredCurrent), observedBusses.IndexOf(observedBus)] = new DoubleComplex(1, 0);
                        }
                        else if (node.InternalID == (measuredCurrent.MeasuredToNodeID))
                        {
                            // If measurement X points toward bus Z its element should be -1
                            m_A[measuredCurrents.IndexOf(measuredCurrent), observedBusses.IndexOf(observedBus)] = new DoubleComplex(-1, 0);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.ThreePhase"/> version of the <see cref="CurrentMeasurementBusMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildThreePhaseMatrix(Network network)
        {
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            // Create an empty current measurement-bus incidence matrix rows equal in number to the number of current phasors and
            // columns equal in number to the number of observed busses in the network.
            m_A = ComplexMatrix.Create(3 * measuredCurrents.Count, 3 * observedBusses.Count, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                foreach (ObservedBus observedBus in observedBusses)
                {
                    if (observedBus.Nodes.Contains(measuredCurrent.MeasuredFromNode))
                    {
                        // If measurement X leaves bus Y its element should be a 3 by 3 identity matrix.
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)), (3 * observedBusses.IndexOf(observedBus))] = new DoubleComplex(1, 0);
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)) + 1, (3 * observedBusses.IndexOf(observedBus)) + 1] = new DoubleComplex(1, 0);
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)) + 2, (3 * observedBusses.IndexOf(observedBus)) + 2] = new DoubleComplex(1, 0);
                    }
                    else if (observedBus.Nodes.Contains(measuredCurrent.MeasuredToNode))
                    {
                        // If measurement X points toward bus Z its element should be a 3 by 3 negative identity matrix
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)), (3 * observedBusses.IndexOf(observedBus))] = new DoubleComplex(-1, 0);
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)) + 1, (3 * observedBusses.IndexOf(observedBus)) + 1] = new DoubleComplex(-1, 0);
                        m_A[(3 * measuredCurrents.IndexOf(measuredCurrent)) + 2, (3 * observedBusses.IndexOf(observedBus)) + 2] = new DoubleComplex(-1, 0);
                    }
                }
            }
        }

        private void BuildExhaustivePositiveSequenceMatrix(Network network)
        {
            throw new NotImplementedException();
        }

        private void BuildExhaustiveThreePhaseMatrix(Network network)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Converts the <see cref="CurrentMeasurementBusIncidenceMatrix"/> into a string in CSV format.
        /// </summary>
        /// <returns>A CSV formatted string representation of the <see cref="CurrentMeasurementBusIncidenceMatrix"/>.</returns>
        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            if (m_phaseSelection == PhaseSelection.PositiveSequence)
            {
                string columnHeaders = "";
                foreach (string header in m_csvColumnHeaders)
                {
                    columnHeaders += header + ",";
                }

                columnHeaders = columnHeaders.Substring(0, columnHeaders.Length - 1);
                stringBuilder.AppendFormat(columnHeaders + "{0}", Environment.NewLine);
            }

            for (int i = 0; i < m_A.RowCount; i++)
            {
                string csvRow = "";

                if (m_phaseSelection == PhaseSelection.PositiveSequence)
                {
                    csvRow = m_csvRowHeaders[i] + ",";
                }
                for (int j = 0; j < m_A.ColumnCount; j++)
                {
                    if (j < m_A.ColumnCount - 1)
                    {
                        csvRow += m_A[i, j].Re.ToString() + ",";
                    }
                    else
                    {
                        csvRow += m_A[i, j].Re.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Method to print the <see cref="CurrentMeasurementBusIncidenceMatrix"/> to a CSV file named 'A Matrix yyyy-MM-dd  hh-mm-ss.csv' where the date and time are in UTC time.
        /// </summary>
        /// <param name="pathName">The directory name (not including the file name) to the destination for the file to be written to.</param>
        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\A Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".csv"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
