﻿//******************************************************************************************************
//  MatrixCalculationExtensions.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code as part of SystemMatrix class
//  06/01/2013 - Kevin D. Jones
//       Removed from SystemMatrix class and made it its own set of extensions.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using Extreme.Mathematics.LinearAlgebra.Providers;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// 
    /// </summary>
    public static class MatrixCalculationExtensions
    {
        /// <summary>
        /// Vertically concatenates two ComplexMatrix objects into one ComplexMatrix object.
        /// </summary>
        /// <param name="A"> 
        /// Top partition of desired matrix.
        /// </param>
        /// <param name="B"> 
        /// Botton partition of desired matrix.
        /// </param>
        /// <returns>
        /// ComplexMatrix object which is the vertical concatenation of two ComplexMatrix objects.
        /// </returns>
        public static ComplexMatrix VerticallyConcatenate(ComplexMatrix A, ComplexMatrix B)
        {
            if (A.ColumnCount == B.ColumnCount)
            {
                ComplexMatrix X = ComplexMatrix.Create(A.RowCount + B.RowCount, A.ColumnCount, 0);

                for (int i = 0; i < A.RowCount; i++)
                {
                    for (int j = 0; j < A.ColumnCount; j++)
                    {
                        X[i, j] = A[i, j];
                    }
                }

                for (int i = 0; i < B.RowCount; i++)
                {
                    for (int j = 0; j < B.ColumnCount; j++)
                    {
                        X[i + A.RowCount, j] = B[i, j];
                    }
                }
                return X;
            }
            else
            {
                throw new Exception("Matrix dimensions do not align: A - (" + A.RowCount.ToString() + " x " + A.ColumnCount.ToString() + ")  B - ("+ B.RowCount.ToString() + " x " + B.ColumnCount.ToString() + ")");
            }
        }

        /// <summary>
        /// Horizontally concatenates two ComplexMatrix objects into one ComplexMatrix object.
        /// </summary>
        /// <param name="A">
        /// Left partition of desired matrix.
        /// </param>
        /// <param name="B">
        /// Right partition of desired matrix.
        /// </param>
        /// <returns>
        /// ComplexMatrix object which is the horizontal concatenation of two ComplexMatrix objects.
        /// </returns>
        public static ComplexMatrix HorizontallyConcatenate(ComplexMatrix A, ComplexMatrix B)
        {
            if (A.RowCount == B.RowCount)
            {
                ComplexMatrix X = ComplexMatrix.Create(A.RowCount, A.ColumnCount + B.ColumnCount, 0);

                for (int i = 0; i < A.RowCount; i++)
                {
                    for (int j = 0; j < A.ColumnCount; j++)
                    {
                        X[i, j] = A[i, j];
                    }
                }

                for (int i = 0; i < B.RowCount; i++)
                {
                    for (int j = 0; j < B.ColumnCount; j++)
                    {
                        X[i, j + A.ColumnCount] = B[i, j];
                    }
                }

                return X;
            }
            else
            {
                throw new Exception("Matrix dimensions do not align: A - (" + A.RowCount.ToString() + " x " + A.ColumnCount.ToString() + ")  B - (" + B.RowCount.ToString() + " x " + B.ColumnCount.ToString() + ")");
            }
        }

        /// <summary>
        /// Removes undesired column from specified matrix
        /// </summary>
        /// <param name="A">
        /// The matrix to have the column removed from.
        /// </param>
        /// <param name="column">
        /// The integer index of the undesired column
        /// </param>
        /// <returns>
        /// ComplexMatrix object which has had a specified column removed
        /// </returns>
        public static ComplexMatrix RemoveColumn(ComplexMatrix A, int column)
        {
            ComplexMatrix newMatrix = ComplexMatrix.Create(A.RowCount, A.ColumnCount - 1, 0);

            if (column == 0)
            {
                newMatrix = A.GetSubmatrix(0, A.RowCount - 1, 1, A.ColumnCount - 1);
            }
            else if (column == 1)
            {
                ComplexVector leftColumn = A.GetColumn(0);
                ComplexMatrix rightMatrix = A.GetSubmatrix(0, A.RowCount - 1, 2, A.ColumnCount - 1);
                ComplexMatrix leftColumnAsMatrix = ComplexMatrix.Create(leftColumn.Count(), 1, 0);

                for (int i = 0; i < leftColumn.Count(); i++)
                {
                    leftColumnAsMatrix[i, 0] = leftColumn[i];
                }

                newMatrix = HorizontallyConcatenate(leftColumnAsMatrix, rightMatrix);
            }
            else if (column == A.ColumnCount - 2)
            {
                ComplexVector rightColumn = A.GetColumn(A.ColumnCount - 1);
                ComplexMatrix leftMatrix = A.GetSubmatrix(0, A.RowCount - 1, 0, A.ColumnCount - 3);
                ComplexMatrix rightColumnAsMatrix = ComplexMatrix.Create(rightColumn.Count(), 1, 0);

                for (int i = 0; i < rightColumn.Count(); i++)
                {
                    rightColumnAsMatrix[i, 0] = rightColumn[i];
                }

                newMatrix = HorizontallyConcatenate(leftMatrix, rightColumnAsMatrix);
            }
            else if (column == A.ColumnCount - 1)
            {
                newMatrix = A.GetSubmatrix(0, A.RowCount - 1, 0, A.ColumnCount - 2);
            }
            else if (column > 1 && column < A.ColumnCount - 2)
            {
                ComplexMatrix leftMatrix = A.GetSubmatrix(0, A.RowCount - 1, 0, column - 1);
                ComplexMatrix rightMatrix = A.GetSubmatrix(0, A.RowCount - 1, column + 1, A.ColumnCount - 1);

                newMatrix = HorizontallyConcatenate(leftMatrix, rightMatrix);
            }

            return newMatrix;
        }

        /// <summary>
        /// Removes undesired row from specified matrix
        /// </summary>
        /// <param name="A">
        /// The matrix to have the row removed from.
        /// </param>
        /// <param name="row">
        /// The integer index of the undesired row.
        /// </param>
        /// <returns>
        /// ComplexMatrix object which has had a specified row removed.
        /// </returns>
        public static ComplexMatrix RemoveRow(ComplexMatrix A, int row)
        {
            ComplexMatrix newMatrix = ComplexMatrix.Create(A.RowCount - 1, A.ColumnCount, 0);

            if (row == 0)
            {
                newMatrix = A.GetSubmatrix(1, A.RowCount - 1, 0, A.ColumnCount - 1);
            }
            else if (row == 1)
            {
                ComplexVector topRow = A.GetRow(0);
                ComplexMatrix bottomMatrix = A.GetSubmatrix(2, A.RowCount - 1, 0, A.ColumnCount - 1);
                ComplexMatrix topRowAsMatrix = ComplexMatrix.Create(1, topRow.Count(), 0);

                for (int i = 0; i < topRow.Count(); i++)
                {
                    topRowAsMatrix[0, i] = topRow[i];
                }

                newMatrix = VerticallyConcatenate(topRowAsMatrix, bottomMatrix);
            }
            else if (row == A.RowCount - 2)
            {
                ComplexVector bottomRow = A.GetRow(A.RowCount - 1);
                ComplexMatrix topMatrix = A.GetSubmatrix(0, A.RowCount - 3, 0, A.ColumnCount - 1);
                ComplexMatrix bottomRowAsMatrix = ComplexMatrix.Create(1, bottomRow.Count(), 0);

                for (int i = 0; i < bottomRow.Count(); i++)
                {
                    bottomRowAsMatrix[0, i] = bottomRow[i];
                }

                newMatrix = VerticallyConcatenate(topMatrix, bottomRowAsMatrix);
            }
            else if (row == A.RowCount - 1)
            {
                newMatrix = A.GetSubmatrix(0, A.RowCount - 2, 0, A.ColumnCount - 1);
            }
            else if (row > 1 && row < A.RowCount - 2)
            {
                ComplexMatrix topMatrix = A.GetSubmatrix(0, row - 1, 0, A.ColumnCount - 1);
                ComplexMatrix bottomMatrix = A.GetSubmatrix(row + 1, A.RowCount - 1, 0, A.ColumnCount - 1);

                newMatrix = VerticallyConcatenate(topMatrix, bottomMatrix);
            }
            return newMatrix;
        }
       
    }
}
