﻿//******************************************************************************************************
//  SeriesAdmittanceMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// The <see cref="SeriesAdmittanceMatrix"/> is used to create an instance of the series admittance 
    /// matrix for the specified power system. The matrix can be used for either 
    /// a single phase system or a three phase system by specifying the <see cref="PhaseSelection"/> 
    /// to be either <see cref="PhaseSelection.PositiveSequence"/> or 
    /// <see cref="PhaseSelection.ThreePhase"/>.
    /// </summary>
    /// <remarks>
    /// <para>The series admittance matrix is a diagonal matrix (block diagonal in three-phase representation) 
    /// where the diagonal elements are the admittances of the lines being measured. It is an <i><b>m</b></i> 
    /// by <i><b>m</b></i> matrix where <i><b>m</b></i> is the number of current measurements in the network. 
    /// It is populated using a single rule:</para>
    /// <list type="number">
    /// <item><description>For measurement <i><b>X</b></i>, the (<i><b>X</b></i>,<i><b>X</b></i>) matrix element 
    /// is the admittance of the branch being measured.</description></item>
    /// </list>
    /// <para>These rules apply for both single phase and three phase systems. In three phase systems the 
    /// elements described in the above rules are replaced by 3 by 3 blocks that are the three phase impedances
    /// of the lines being measured.</para>
    /// </remarks>
    /// <seealso cref="CurrentMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="VoltageMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="ShuntSusceptanceMatrix"/>
    /// <seealso cref="CovarianceMatrix"/>
    /// <seealso cref="SystemMatrix"/>
    public class SeriesAdmittanceMatrix
    {
        #region [ Private Members ]

        private ComplexMatrix m_Y = ComplexMatrix.Create(0, 0, 0);
        private ComplexMatrix m_exhaustiveY = ComplexMatrix.Create(0, 0, 0);

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The series admittance matrix
        /// </summary>
        public ComplexMatrix Matrix
        {
            get 
            { 
                return m_Y; 
            }
        }

        /// <summary>
        /// The series admittance matrix assuming a current phasor at each end of each branch in the network.
        /// </summary>
        public ComplexMatrix ExhaustiveMatrix
        {
            get 
            { 
                return m_exhaustiveY; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// The designated constructor for the <see cref="SeriesAdmittanceMatrix"/> class.
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/> model.</param>
        public SeriesAdmittanceMatrix(Network network)
        {
            BuildMatrices(network);
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Method that decides whether to build the <see cref="PhaseSelection.PositiveSequence"/> matrices or
        /// the <see cref="PhaseSelection.ThreePhase"/> matrices;
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/>.</param>
        private void BuildMatrices(Network network)
        {
            if (network.Model.PhaseConfiguration == PhaseSelection.PositiveSequence)
            {
                BuildPositiveSequenceMatrix(network);
                //BuildExhaustiveSinglePhaseMatrix(network);
            }
            else if (network.Model.PhaseConfiguration == PhaseSelection.ThreePhase)
            {
                BuildThreePhaseMatrix(network);
                //BuildExhaustiveThreePhaseMatrix(network);
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.PositiveSequence"/> version of the <see cref="SeriesAdmittanceMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildPositiveSequenceMatrix(Network network)
        {
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;
            // Create an empty series admittance matrix with rows and columns equal in number to the number of current phasors in the network.

            // To cut the operations in half for setting the size of the matrix, just count the currents once.
            int measuredCurrentsCount = measuredCurrents.Count;
            m_Y = ComplexMatrix.Create(measuredCurrentsCount, measuredCurrentsCount, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                // To reduce the operatiosn in half for setting the matrix element index of the measured current, just find its index once.
                int indexOfMeasuredCurrent = measuredCurrents.IndexOf(measuredCurrent);

                // The matrix element Z, the number of the current phasor, should be the 
                // positive sequence admittance of the branch measured by measurement Z
                if (measuredCurrent.MeasuredBranch is Transformer)
                {
                    m_Y[indexOfMeasuredCurrent, indexOfMeasuredCurrent] = (measuredCurrent.MeasuredBranch as Transformer).PositiveSequenceSeriesAdmittance;
                }
                else if (measuredCurrent.MeasuredBranch is TransmissionLine)
                {
                    if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).FromSubstationCurrentID)
                    {
                        m_Y[indexOfMeasuredCurrent, indexOfMeasuredCurrent] = 
                            (measuredCurrent.MeasuredBranch as TransmissionLine).FromSideImpedanceToDeepestObservability.PositiveSequenceSeriesAdmittance;
                    }
                    else if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).ToSubstationCurrentID)
                    {
                        m_Y[indexOfMeasuredCurrent, indexOfMeasuredCurrent] =
                            (measuredCurrent.MeasuredBranch as TransmissionLine).ToSideImpedanceToDeepestObservability.PositiveSequenceSeriesAdmittance;
                    }
                }
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.ThreePhase"/> version of the <see cref="SeriesAdmittanceMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildThreePhaseMatrix(Network network)
        {
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;
            // Create an empty series admitanc ematrix with rows and columns equal in number to the number of current phasors in the network. The NetworkModel
            // organizes phasors into their A, B, C, Pos Sequence groups so the count must be multiplied by three for three phase matrices.

            // To cut the operations in half for setting the size of the matrix, just count the currents once.
            int measuredCurrentsCount = 3 * measuredCurrents.Count;
            m_Y = ComplexMatrix.Create(measuredCurrentsCount, measuredCurrentsCount, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                // To reduce the operatiosn in half for setting the matrix element index of the measured current, just find its index once.
                int indexOfMeasuredCurrent = measuredCurrents.IndexOf(measuredCurrent);

                if (measuredCurrent.MeasuredBranch is Transformer)
                {
                    // Get the 3 by 3 three phase admittance matrix for the branch measured by this set of current phasors
                    ComplexMatrix seriesAdmittance = (measuredCurrent.MeasuredBranch as Transformer).ThreePhaseSeriesAdmittance;

                    // The block matrix element Z, the number of the current phasor, should be the 
                    // three phase admittance of the branch measured by measurement Z
                    for (int row = 0; row < 3; row++)
                    {
                        for (int col = 0; col < 3; col++)
                        {
                            m_Y[3 * indexOfMeasuredCurrent + row, 3 * indexOfMeasuredCurrent + col] = seriesAdmittance[row, col];
                        }
                    }
                }
                else if (measuredCurrent.MeasuredBranch is TransmissionLine)
                {
                    if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).FromSubstationCurrentID)
                    {
                        // Get the 3 by 3 three phase admittance matrix for the branch measured by this set of current phasors
                        ComplexMatrix seriesAdmittance = (measuredCurrent.MeasuredBranch as TransmissionLine).FromSideImpedanceToDeepestObservability.ThreePhaseSeriesAdmittance;

                        // The block matrix element Z, the number of the current phasor, should be the 
                        // three phase admittance of the branch measured by measurement Zw
                        for (int row = 0; row < 3; row++)
                        {
                            for (int col = 0; col < 3; col++)
                            {
                                m_Y[3 * indexOfMeasuredCurrent + row, 3 * indexOfMeasuredCurrent + col] = seriesAdmittance[row, col];
                            }
                        }
                    }
                    else if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).ToSubstationCurrentID)
                    {
                        // Get the 3 by 3 three phase admittance matrix for the branch measured by this set of current phasors
                        ComplexMatrix seriesAdmittance = (measuredCurrent.MeasuredBranch as TransmissionLine).ToSideImpedanceToDeepestObservability.ThreePhaseSeriesAdmittance;

                        // The block matrix element Z, the number of the current phasor, should be the 
                        // three phase admittance of the branch measured by measurement Zw
                        for (int row = 0; row < 3; row++)
                        {
                            for (int col = 0; col < 3; col++)
                            {
                                m_Y[3 * indexOfMeasuredCurrent + row, 3 * indexOfMeasuredCurrent + col] = seriesAdmittance[row, col];
                            }
                        }
                    }
                }
            }
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Converts the <see cref="SeriesAdmittanceMatrix"/> into a string in CSV format.
        /// </summary>
        /// <returns>A CSV formatted string representation of the <see cref="SeriesAdmittanceMatrix"/>.</returns>
        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            for (int i = 0; i < m_Y.RowCount; i++)
            {
                string csvRow = "";
                for (int j = 0; j < m_Y.ColumnCount; j++)
                {
                    if (j < m_Y.ColumnCount - 1)
                    {
                        csvRow += m_Y[i, j].Re.ToString() + ", " + m_Y[i, j].Im.ToString() + ",";
                    }
                    else
                    {
                        csvRow += m_Y[i, j].Re.ToString() + "," + m_Y[i, j].Im.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Method to print the <see cref="SeriesAdmittanceMatrix"/> to a CSV file named 'Y Matrix yyyy-MM-dd  hh-mm-ss.csv' where the date and time are in UTC time.
        /// </summary>
        /// <param name="pathName">The directory name (not including the file name) to the destination for the file to be written to.</param>
        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\Y Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".csv"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
