﻿//******************************************************************************************************
//  ShuntSusceptanceMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// The <see cref="ShuntSusceptanceMatrix"/> is used to create an instance of the shunt susceptance
    /// matrix for the specified power system. The matrix can be used for either 
    /// a single phase system or a three phase system by specifying the <see cref="PhaseSelection"/> 
    /// to be either <see cref="PhaseSelection.PositiveSequence"/> or 
    /// <see cref="PhaseSelection.ThreePhase"/>.
    /// </summary>
    /// <remarks>
    /// <para>The shunt susceptance matrix is a martix that relates the lcoation of each current 
    /// measurement to the shunt susceptance of the line that it is measuring. It is an <i><b>m</b></i> 
    /// by <i><b>b</b></i> matrix where <i><b>m</b></i> is the number of current measurements in the 
    /// network and <i><b>b</b></i> is the number of the bus where the current measurement originiates. 
    /// It is populated using a single rule:</para>
    /// <list type="number">
    /// <item><description>For measurement <i><b>X</b></i> (corresponding to row <i><b>X</b></i>) leaving 
    /// bus <i><b>Y</b></i> (corresponding to column <i><b>Y</b></i>), the matrix element 
    /// (<i><b>X</b></i>,<i><b>Y</b></i>) is the shunt susceptance of the side of the line where 
    /// measurement <i><b>X</b></i> was taken.</description></item>
    /// </list>
    /// <para>These rules apply for both single phase and three phase systems. In three phase systems the
    /// elements described in the above rules are replaced by 3 by 3 blocks that are the three phase
    /// shunt susceptance of the branch being measured.</para>
    /// </remarks>
    /// <seealso cref="CurrentMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="VoltageMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="SeriesAdmittanceMatrix"/>
    /// <see cref="CovarianceMatrix"/>
    /// <seealso cref="SystemMatrix"/>
    public class ShuntSusceptanceMatrix
    {
        #region [ Private Members ]

        private ComplexMatrix m_Ys = ComplexMatrix.Create(1, 1, 0);
        private ComplexMatrix m_exhaustiveYs = ComplexMatrix.Create(1, 1, 0);

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The shunt susceptance matrix
        /// </summary>
        public ComplexMatrix Matrix
        {
            get 
            { 
                return m_Ys; 
            }
        }

        /// <summary>
        /// The shunt susceptance matrix assuming there is a current phasor at each end of each branch in the network
        /// </summary>
        public ComplexMatrix ExhaustiveMatrix
        {
            get 
            { 
                return m_exhaustiveYs; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// The designated constructor for the <see cref="ShuntSusceptanceMatrix"/> class.
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/> model.</param>
        public ShuntSusceptanceMatrix(Network network)
        {
            BuildMatrices(network);
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Method that decides whether to build the <see cref="PhaseSelection.PositiveSequence"/> matrices or
        /// the <see cref="PhaseSelection.ThreePhase"/> matrices;
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/>.</param>
        private void BuildMatrices(Network network)
        {
            if (network.Model.PhaseConfiguration == PhaseSelection.PositiveSequence)
            {
                BuildPositiveSequenceMatrix(network);
                //BuildExhaustiveSinglePhaseMatrix(network);
            }
            else if (network.Model.PhaseConfiguration == PhaseSelection.ThreePhase)
            {
                BuildThreePhaseMatrix(network);
                //BuildExhaustiveThreePhaseMatrix(network);
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.PositiveSequence"/> version of the <see cref="ShuntSusceptanceMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildPositiveSequenceMatrix(Network network)
        {
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            m_Ys = ComplexMatrix.Create(measuredCurrents.Count, observedBusses.Count, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                foreach (ObservedBus observedBus in observedBusses)
                {
                    if (observedBus.Nodes.Contains(measuredCurrent.MeasuredFromNode))
                    {
                        if (measuredCurrent.MeasuredBranch is Transformer)
                        {
                            m_Ys[measuredCurrents.IndexOf(measuredCurrent), observedBusses.IndexOf(observedBus)] =
                                (measuredCurrent.MeasuredBranch as Transformer).PositiveSequenceShuntSusceptance / 2;
                        }
                        else if (measuredCurrent.MeasuredBranch is TransmissionLine)
                        {
                            if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).FromSubstationCurrentID)
                            {
                                m_Ys[measuredCurrents.IndexOf(measuredCurrent), observedBusses.IndexOf(observedBus)] =
                                (measuredCurrent.MeasuredBranch as TransmissionLine).FromSideImpedanceToDeepestObservability.PositiveSequenceShuntSusceptance / 2;
                            }
                            else if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).ToSubstationCurrentID)
                            {
                                m_Ys[measuredCurrents.IndexOf(measuredCurrent), observedBusses.IndexOf(observedBus)] =
                                (measuredCurrent.MeasuredBranch as TransmissionLine).ToSideImpedanceToDeepestObservability.PositiveSequenceShuntSusceptance / 2;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.ThreePhase"/> version of the <see cref="ShuntSusceptanceMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildThreePhaseMatrix(Network network)
        {
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;
            List<CurrentPhasorGroup> measuredCurrents = network.Model.ActiveCurrentPhasors;

            m_Ys = ComplexMatrix.Create(3 * measuredCurrents.Count, 3 * observedBusses.Count, 0);

            foreach (CurrentPhasorGroup measuredCurrent in measuredCurrents)
            {
                foreach (ObservedBus observedBus in observedBusses)
                {
                    if (observedBus.Nodes.Contains(measuredCurrent.MeasuredFromNode))
                    {
                        if (measuredCurrent.MeasuredBranch is Transformer)
                        {
                            ComplexMatrix susceptance = (measuredCurrent.MeasuredBranch as Transformer).ThreePhaseShuntSusceptance;
                            for (int row = 0; row < 3; row++)
                            {
                                for (int col = 0; col < 3; col++)
                                {
                                    m_Ys[3 * measuredCurrents.IndexOf(measuredCurrent) + row, 3 * observedBusses.IndexOf(observedBus) + col] = susceptance[row, col] / 2;
                                }
                            }
                        }
                        else if (measuredCurrent.MeasuredBranch is TransmissionLine)
                        {
                            if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).FromSubstationCurrentID)
                            {
                                ComplexMatrix susceptance = (measuredCurrent.MeasuredBranch as TransmissionLine).FromSideImpedanceToDeepestObservability.ThreePhaseShuntSusceptance;
                                for (int row = 0; row < 3; row++)
                                {
                                    for (int col = 0; col < 3; col++)
                                    {
                                        m_Ys[3 * measuredCurrents.IndexOf(measuredCurrent) + row, 3 * observedBusses.IndexOf(observedBus) + col] = susceptance[row, col] / 2;
                                    }
                                }
                            }
                            else if (measuredCurrent.InternalID == (measuredCurrent.MeasuredBranch as TransmissionLine).ToSubstationCurrentID)
                            {
                                ComplexMatrix susceptance = (measuredCurrent.MeasuredBranch as TransmissionLine).ToSideImpedanceToDeepestObservability.ThreePhaseShuntSusceptance;
                                for (int row = 0; row < 3; row++)
                                {
                                    for (int col = 0; col < 3; col++)
                                    {
                                        m_Ys[3 * measuredCurrents.IndexOf(measuredCurrent) + row, 3 * observedBusses.IndexOf(observedBus) + col] = susceptance[row, col] / 2;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Converts the <see cref="ShuntSusceptanceMatrix"/> into a string in CSV format.
        /// </summary>
        /// <returns>A CSV formatted string representation of the <see cref="ShuntSusceptanceMatrix"/>.</returns>
        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            for (int i = 0; i < m_Ys.RowCount; i++)
            {
                string csvRow = "";
                for (int j = 0; j < m_Ys.ColumnCount; j++)
                {
                    if (j < m_Ys.ColumnCount - 1)
                    {
                        csvRow += m_Ys[i, j].Im.ToString() + ",";
                    }
                    else
                    {
                        csvRow += m_Ys[i, j].Im.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Method to print the <see cref="ShuntSusceptanceMatrix"/> to a CSV file named 'Ys Matrix yyyy-MM-dd  hh-mm-ss.csv' where the date and time are in UTC time.
        /// </summary>
        /// <param name="pathName">The directory name (not including the file name) to the destination for the file to be written to.</param>
        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\Ys Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".csv"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
