﻿//******************************************************************************************************
//  SystemMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using Extreme.Mathematics.LinearAlgebra.Providers;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;
using System.Configuration;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// The <see cref="SystemMatrix"/> class is used for managing an instance of matrix which represents a 
    /// linear relationship between voltage and current phasors and the state of the power system.
    /// </summary>
    /// <remarks>The <see cref="SystemMatrix"/> class creates instances of the 
    /// <see cref="CurrentMeasurementBusIncidenceMatrix"/> class, the 
    /// <see cref="VoltageMeasurementBusIncidenceMatrix"/> class, the 
    /// <see cref="SeriesAdmittanceMatrix"/> class, and the 
    /// <see cref="ShuntSusceptanceMatrix"/> class then calculates and arranges 
    /// them in the full matrix. Several class methods used for concatenation and 
    /// matrix manipulation are left public for use if necessary. The matrix can be used for either 
    /// a single phase system or a three phase system by specifying the <see cref="NetworkModel.PhaseConfiguration"/> 
    /// to be either <see cref="Network.PhaseSelection.PositiveSequence"/> or 
    /// <see cref="Network.PhaseSelection.ThreePhase"/>.</remarks>
    /// <seealso cref="CurrentMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="VoltageMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="SeriesAdmittanceMatrix"/>
    /// <seealso cref="CovarianceMatrix"/>
    /// <seealso cref="ShuntSusceptanceMatrix"/>
    public class SystemMatrix
    {
        #region [ Private Members ]

        private CurrentMeasurementBusIncidenceMatrix m_A;
        private VoltageMeasurementBusIncidenceMatrix m_II;
        private SeriesAdmittanceMatrix m_Y;
        private ShuntSusceptanceMatrix m_Ys;
        private CovarianceMatrix m_R;
        private ComplexMatrix m_B = ComplexMatrix.Create(0, 0, 0);
        private ComplexMatrix m_M = ComplexMatrix.Create(0, 0, 0);
        private ComplexMatrix m_K = ComplexMatrix.Create(0, 0, 0);
        private ComplexMatrix m_ExhaustiveK = ComplexMatrix.Create(0, 0, 0);
        #endregion

        #region [ Properties ]

        /// <summary>
        /// The <see cref="CurrentMeasurementBusIncidenceMatrix"/>.
        /// </summary>
        public CurrentMeasurementBusIncidenceMatrix A
        {
            get
            {
                return m_A;
            }
        }

        /// <summary>
        /// The <see cref="VoltageMeasurementBusIncidenceMatrix"/>.
        /// </summary>
        public VoltageMeasurementBusIncidenceMatrix II
        {
            get
            {
                return m_II;
            }
        }

        /// <summary>
        /// The <see cref="SeriesAdmitanceMatrix"/>.
        /// </summary>
        public SeriesAdmittanceMatrix Y
        {
            get
            {
                return m_Y;
            }
        }

        /// <summary>
        /// The <see cref="ShuntSusceptanceMatrix"/>.
        /// </summary>
        public ShuntSusceptanceMatrix Ys
        {
            get
            {
                return m_Ys;
            }
        }

        /// <summary>
        /// An overdefined system which linearly relates voltage and current phasor measurements to state variables in a power system
        /// </summary>
        public ComplexMatrix Matrix
        {
            get 
            { 
                return m_B; 
            }
        }

        /// <summary>
        /// The psuedo inverse of the system matrix which can be used to calculate the system state.
        /// </summary>
        public ComplexMatrix PsuedoInverseOfMatrix
        {
            get 
            { 
                return m_M; 
            }
        }

        /// <summary>
        /// The lower partition of the un-inverted system matrix which relates current measurements to state variables in power system.
        /// </summary>
        public ComplexMatrix LowerPartitionOfMatrix
        {
            get 
            { 
                return m_K; 
            }
        }

        /// <summary>
        /// The lower partition of the system matrix assumign that there is a current phasor at each end of every branch in the network
        /// </summary>
        //public ComplexMatrix ExhaustiveLowerPartitionOfMatrix
        //{
        //    get 
        //    { 
        //        return ExhaustiveK; 
        //    }
        //}

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Creates a new instance of the <see cref="SystemMatrix"/> class.
        /// </summary>
        /// <remarks>
        /// Creates instances of the <see cref="CurrentMeasurementBusIncidenceMatrix"/> class, 
        /// the <see cref="VoltageMeasurementBusIncidenceMatrix"/> class, the 
        /// <see cref="SeriesAdmittanceMatrix"/> class, and the <see cref="ShuntSusceptanceMatrix"/> 
        /// class then calculates and arranges them in the full matrix. It also calculates the 
        /// psuedo-inverse of the system matrix.
        /// </remarks>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        public SystemMatrix(Network network)
        {
            BuildMatrices(network);
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Builds the system matrix
        /// </summary>
        /// <param name="networkModel">The virtualized <see cref="Network"/> model.</param>
        private void BuildMatrices(Network network)
        {

            // Tell the Kernel which provider to use so we can use GetInverse() method

            if (System.Environment.Is64BitOperatingSystem)
            {
                Kernel.UseProvider(new ParallelX64LinearAlgebraProvider());
            }
            else
            {
                //Kernel.UseProvider(new ParallelX86LinearAlgebraProvider());
            }

            BuildLowerPartitionOfSystemMatrix(network);
            BuildSystemMatrix(network);
            ComputePsuedoInverseOfSystemMatrix();
            
        }

        /// <summary>
        /// Builds the <see cref="CurrentMeasurementBusIncidenceMatrix"/>, the
        /// <see cref="SeriesAdmittanceMatrix"/>, the <see cref="ShuntSusceptanceMatrix"/> and
        /// computes the <see cref="LowerPartitionOfMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildLowerPartitionOfSystemMatrix(Network network)
        {
            // Build the Current Measurement-Bus Incidence Matrix
            m_A = new CurrentMeasurementBusIncidenceMatrix(network);

            // Build the Series Admittance Matrix
            m_Y = new SeriesAdmittanceMatrix(network);

            // Build the Shunt Susceptance Matrix
            m_Ys = new ShuntSusceptanceMatrix(network);

            // Compute the lower partition of the system matrix
            m_K = m_Y.Matrix * m_A.Matrix + m_Ys.Matrix;
        }

        /// <summary>
        /// Builds the <see cref="VoltageMeasurmentBusIncidenceMatrix"/>, the
        /// <see cref="CovarianceMatrix"/> and concatenates the <see cref="LowerPartitionOfMatrix"/>
        /// with the <see cref="VoltageMeasurmentBusIncidenceMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildSystemMatrix(Network network)
        {
            // Build the Voltage Measurement-Bus Incidence Matrix
            m_II = new VoltageMeasurementBusIncidenceMatrix(network);

            // Build the Measurement Covariance Matrix
            m_R = new CovarianceMatrix(network);

            // Build the SystemMatrix
            m_B = MatrixCalculationExtensions.VerticallyConcatenate(m_II.Matrix, m_K);
        }

        /// <summary>
        /// Computes the psuedo-inverse of the system matrix
        /// </summary>
        private void ComputePsuedoInverseOfSystemMatrix()
        {
            ComplexMatrix BTranspose = m_B.Transpose();
            m_M = ((BTranspose * m_R.Matrix.GetInverse() * m_B).GetInverse()) * BTranspose * m_R.Matrix.GetInverse();
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Converts the uninverted <see cref="SystemMatrix"/> into a string in CSV format.
        /// </summary>
        /// <returns>A CSV formatted string representation of the <see cref="SystemMatrix"/>.</returns>
        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            for (int i = 0; i < m_M.RowCount; i++)
            {
                string csvRow = "";
                for (int j = 0; j < m_B.ColumnCount; j++)
                {
                    if (j < m_B.ColumnCount - 1)
                    {
                        csvRow += m_B[i, j].Re.ToString() + ", " + m_B[i, j].Im.ToString() + ",";
                    }
                    else
                    {
                        csvRow += m_B[i, j].Re.ToString() + "," + m_B[i, j].Im.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Method to print the <see cref="SystemMatrix"/> to a CSV file named 'System Matrix yyyy-MM-dd  hh-mm-ss.csv' where the date and time are in UTC time.
        /// </summary>
        /// <param name="pathName">The directory name (not including the file name) to the destination for the file to be written to.</param>
        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\System Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".txt"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
