﻿//******************************************************************************************************
//  VoltageMeasurementBusIncidenceMatrix.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  12/30/2009 - Kevin D. Jones
//       Created original algorithms in Matlab
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to accept new parameters in constructor and use new Network Model
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;

namespace SynchrophasorAnalytics.Matrices
{
    /// <summary>
    /// The <see cref="VoltageMeasurementBusIncidenceMatrix"/> is used to create an instance of the voltage 
    /// measuremnet-bus incidence matrix for the specified power system configuration. The matrix can be 
    /// used for either a single phase system or a three phase system by specifying the
    /// <see cref="PhaseSelection"/> to be either <see cref="PhaseSelection.PositiveSequence"/> or 
    /// <see cref="PhaseSelection.ThreePhase"/>.
    /// </summary>
    /// <remarks>
    /// <para>The voltage measurement-bus incidence matrix is very similar to the current measurement-bus 
    /// incidence matrix. It shows the relationship between a voltage measurement and its associated location 
    /// in the network.</para>
    /// <para>Because phasor measurements are used, the voltage measurements are actually direct measurements 
    /// of the system state and therefore only require a simple identity relationship between the measurement 
    /// and the state. The voltage measurement-bus incidence matrix is an <i><b>m</b></i> by <i><b>b</b></i> 
    /// matrix where <i><b>m</b></i> is the number of voltage measurements in the network and <i><b>b</b></i> 
    /// is the number of observed busses or essentially, the number of busses which which have at least one
    /// voltage measurement. Nodes in the power system which are directly connected (such as between
    /// <see cref="SynchrophasorAnalytics.Modeling.CircuitBreaker"/>s or <see cref="SynchrophasorAnalytics.Modeling.Switch"/>es. 
    /// This matrix is populated using the following rules:</para>
    /// <list type="number">
    /// <item><description>Each row of the matrix corresponds to a voltage measurement in the 
    /// network.</description></item>
    /// <item><description>Each column of the matrix corresponds to a bus in the system which has a voltage 
    /// measurement.</description></item>
    /// <item><description>If measurement <i><b>X</b></i> (corresponding to row <i><b>X</b></i>) is located 
    /// at observed bus <i><b>Y</b></i> (corresponding to column <i><b>Y</b></i>) then the matrix element 
    /// (<i><b>X</b></i>,<i><b>Y</b></i>) will be a 1.</description></item>
    /// <item><description>All remaining entries will be identically zero.</description></item>
    /// </list>
    /// <para> These rules apply for both single phase and three phase systems. In three phase systems the 
    /// elements described in the above rules are replaced by 3 by 3 blocks that are positive identity, 
    /// or zero.</para>
    /// </remarks>
    /// <seealso cref="MatrixPhaseType"/>
    /// <seealso cref="CurrentMeasurementBusIncidenceMatrix"/>
    /// <seealso cref="SeriesAdmittanceMatrix"/>
    /// <seealso cref="ShuntSusceptanceMatrix"/>
    /// <seealso cref="CovarianceMatrix"/>
    /// <seealso cref="CalibrationMatrix"/>
    /// <seealso cref="SystemMatrix"/>
    public class VoltageMeasurementBusIncidenceMatrix
    {
        #region [ Private Members ]

        private ComplexMatrix m_II = ComplexMatrix.Create(0, 0, 0);
        private List<string> m_csvColumnHeaders;
        private List<string> m_csvRowHeaders;
        private PhaseSelection m_phaseSelection;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The voltage measurement-bus incidence matrix
        /// </summary>
        public ComplexMatrix Matrix
        {
            get 
            { 
                return m_II; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// The designated constructor for the <see cref="VoltageMeasurementBusIncidenceMatrix"/> class.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        public VoltageMeasurementBusIncidenceMatrix(Network network)
        {
            BuildMatrices(network);
            m_phaseSelection = network.Model.PhaseConfiguration;
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Method that decides whether to build the <see cref="PhaseSelection.PositiveSequence"/> matrices or
        /// the <see cref="PhaseSelection.ThreePhase"/> matrices;
        /// </summary>
        /// <param name="network">The virtualized <see cref="NetworkModel"/>.</param>
        private void BuildMatrices(Network network)
        {
            if (network.Model.PhaseConfiguration == PhaseSelection.PositiveSequence)
            {
                BuildPositiveSequenceMatrix(network);
            }
            else if (network.Model.PhaseConfiguration == PhaseSelection.ThreePhase)
            {
                BuildThreePhaseMatrix(network);
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.PositiveSequence"/> version of the <see cref="VoltageMeasurementBusMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildPositiveSequenceMatrix(Network network)
        {
            m_csvRowHeaders = new List<string>();
            m_csvColumnHeaders = new List<string>();
            m_csvColumnHeaders.Add("II Matrix");

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                string columnHeader = "Sub: " + observedBus.InternalID.ToString() + " with nodes: "; 
                foreach (Node node in observedBus.Nodes)
                {
                    columnHeader += node.InternalID.ToString() + "-";
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                        m_csvRowHeaders.Add(node.Description);
                    }
                }
                m_csvColumnHeaders.Add(columnHeader);
            }

            // Construct an empty matrix to populate
            m_II = ComplexMatrix.Create(measuredNodes.Count, observedBusses.Count, 0);

            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        // If measurement X measures bus Y its element should be 1
                        m_II[measuredNodes.IndexOf(node), observedBusses.IndexOf(observedBus)] = new DoubleComplex(1, 0);
                    }
                }
            }
        }

        /// <summary>
        /// Constructs the <see cref="PhaseSelection.ThreePhase"/> version of the <see cref="VoltageMeasurementBusMatrix"/>.
        /// </summary>
        /// <param name="network">The virtualized <see cref="Network"/> model.</param>
        private void BuildThreePhaseMatrix(Network network)
        {
            m_csvRowHeaders = new List<string>();
            m_csvColumnHeaders = new List<string>();
            m_csvColumnHeaders.Add("II Matrix");

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = network.Model.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                string columnHeader = "Sub: " + observedBus.InternalID.ToString() + " with nodes: "; 
                foreach (Node node in observedBus.Nodes)
                {
                    columnHeader += node.InternalID.ToString() + "-";
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                        m_csvRowHeaders.Add(node.Description);
                    }
                }
                m_csvColumnHeaders.Add(columnHeader);
            }

            // Construct an empty matrix to populate
            m_II = ComplexMatrix.Create(3 * measuredNodes.Count, 3 * observedBusses.Count, 0);

            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        // If measurement X measures bus Y its element should be an identity matrix
                        m_II[(3 * measuredNodes.IndexOf(node)), (3 * observedBusses.IndexOf(observedBus))] = new DoubleComplex(1, 0);
                        m_II[(3 * measuredNodes.IndexOf(node)) + 1, (3 * observedBusses.IndexOf(observedBus)) + 1] = new DoubleComplex(1, 0);
                        m_II[(3 * measuredNodes.IndexOf(node)) + 2, (3 * observedBusses.IndexOf(observedBus)) + 2] = new DoubleComplex(1, 0);
                    }
                }
            }
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Converts the <see cref="VoltageMeasurementBusIncidenceMatrix"/> into a string in CSV format.
        /// </summary>
        /// <returns>A CSV formatted string representation of the <see cref="VoltageMeasurementBusIncidenceMatrix"/>.</returns>
        public string ToCsvString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            if (m_phaseSelection == PhaseSelection.PositiveSequence)
            {
                string columnHeaders = "";
                foreach (string header in m_csvColumnHeaders)
                {
                    columnHeaders += header + ",";
                }
                columnHeaders = columnHeaders.Substring(0, columnHeaders.Length - 1);
                stringBuilder.AppendFormat(columnHeaders + "{0}", Environment.NewLine);
            }

            for (int i = 0; i < m_II.RowCount; i++)
            {
                string csvRow = "";

                if (m_phaseSelection == PhaseSelection.PositiveSequence)
                {
                    csvRow = m_csvRowHeaders[i] + ",";
                }

                for (int j = 0; j < m_II.ColumnCount; j++)
                {
                    if (j < m_II.ColumnCount - 1)
                    {
                        csvRow += m_II[i, j].Re.ToString() + ",";
                    }
                    else
                    {
                        csvRow += m_II[i, j].Re.ToString();
                    }
                }
                stringBuilder.AppendFormat(csvRow + "{0}", Environment.NewLine);
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Method to print the <see cref="VoltageMeasurementBusIncidenceMatrix"/> to a CSV file named 'II Matrix yyyy-MM-dd  hh-mm-ss.csv' where the date and time are in UTC time.
        /// </summary>
        /// <param name="pathName">The directory name (not including the file name) to the destination for the file to be written to.</param>
        public void WriteToDebugFile(string pathName)
        {
            using (StreamWriter outfile = new StreamWriter(pathName + @"\II Matrix " + String.Format("{0:yyyy-MM-dd  hh-mm-ss}", DateTime.UtcNow) + ".csv"))
            {
                outfile.Write(this.ToCsvString());
            }
        }

        #endregion
    }
}
