﻿//******************************************************************************************************
//  CurrentPhasorGroup.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  07/20/2011 - Kevin D. Jones
//       Generated original version of source code.
//  06/01/2013 - Kevin D. Jones
//       Modified to include Xml Serialization and Network Model paramters (Node, etc)
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Modeling;

namespace SynchrophasorAnalytics.Measurements
{
    /// <summary>
    /// Encapsulates a group of current phasors in a +, A, B, C grouping and relates them to the network model.
    /// </summary>
    /// <seealso cref="PhasorGroup"/>
    /// <seealso cref="VoltagePhasorGroup"/>
    /// <seealso cref="PhasorType"/>
    [Serializable()]
    public class CurrentPhasorGroup : PhasorGroup
    {
        #region [ Private Members ]

        private Node m_measuredFromNode;
        private int m_measuredFromNodeID;

        private Node m_measuredToNode;
        private int m_measuredToNodeID;

        private ITwoTerminal m_measuredBranch;
        private int m_measuredBranchID;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The <see cref="SynchrophasorAnalytics.Modeling.Node"/> from which the current measurement originates.
        /// </summary>
        [XmlIgnore()]
        public Node MeasuredFromNode
        {
            get 
            { 
                return m_measuredFromNode; 
            }
            set 
            { 
                m_measuredFromNode = value;
                m_measuredFromNodeID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="SynchrophasorAnalytics.Modeling.InternalID"/> of the <see cref="MeasuredFromNode"/>.
        /// </summary>
        [XmlAttribute("FromNode")]
        public int MeasuredFromNodeID
        {
            get
            {
                return m_measuredFromNodeID;
            }
            set
            {
                m_measuredFromNodeID = value;
            }
        }

        /// <summary>
        /// The <see cref="SynchrophasorAnalytics.Modeling.Node"/> that the current measurement is directed towards.
        /// </summary>
        [XmlIgnore()]
        public Node MeasuredToNode
        {
            get
            {
                return m_measuredToNode;
            }
            set
            {
                m_measuredToNode = value;
                m_measuredToNodeID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="SynchrophasorAnalytics.Modeling.InternalID"/> of the <see cref="MeasuredToNode"/>.
        /// </summary>
        [XmlAttribute("ToNode")]
        public int MeasuredToNodeID
        {
            get
            {
                return m_measuredToNodeID;
            }
            set
            {
                m_measuredToNodeID = value;
            }
        }

        [XmlIgnore()]
        public ITwoTerminal MeasuredBranch
        {
            get
            {
                return m_measuredBranch;
            }
            set
            {
                if (value is Transformer)
                {
                    m_measuredBranch = value;
                    m_measuredBranchID = (value as Transformer).InternalID;
                }
                else if (value is TransmissionLine)
                {
                    m_measuredBranch = value;
                    m_measuredBranchID = (value as TransmissionLine).InternalID;
                }
            }
        }

        [XmlIgnore()]
        public int MeasuredBranchID
        {
            get
            {
                return m_measuredBranchID;
            }
            set
            {
                m_measuredBranchID = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public CurrentPhasorGroup()
            : this(0, 0, "Undefined Name", "Undefined Description")
        {
        }

        /// <summary>
        /// A constructor which specifies only the information required by the <see cref="INetworkDescribable"/> interface. The <see cref="Phasor"/>, <see cref="StatusWord"/>, and <see cref="Node"/> objects are instantiated with default initializers to prevent null references.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="name">A descriptive name for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="description">A description of the <see cref="CurrentPhasorGroup"/> object.</param>
        public CurrentPhasorGroup(int internalID, int number, string name, string description)
            : this(internalID, number, name, description, new Phasor(), new Phasor(), new Phasor(), new Phasor())
        {
        }

        /// <summary>
        /// A constructor which specifies the information required by the <see cref="INetworkDescriptor"/> interface and the <see cref="Phasor"/> objects for +, A, B, & C. 
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="name">A descriptive name for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="description">A description of the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="positiveSequence">The <see cref="Phasor"/> representing positive sequence.</param>
        /// <param name="phaseA">The <see cref="Phasor"/> representing phase A.</param>
        /// <param name="phaseB">The <see cref="Phasor"/> representing phase B.</param>
        /// <param name="phaseC">The <see cref="Phasor"/> representing phase C.</param>
        public CurrentPhasorGroup(int internalID, int number, string name, string description, Phasor positiveSequence, Phasor phaseA, Phasor phaseB, Phasor phaseC)
            :this(internalID, number, name, description, positiveSequence, phaseA, phaseB, phaseC, new StatusWord())
        {
        }

        /// <summary>
        /// A constructor which specifies the information required by the <see cref="INetworkDescriptor"/> interface and the <see cref="Phasor"/> objects for +, A, B, & C and the <see cref="SynchrophasorFramework.Modeling.Node.InternalID"/> of the <see cref="MeasuredFromNode"/> and <see cref="MeasuredToNode"/>.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="name">A descriptive name for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="description">A description of the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="positiveSequence">The <see cref="Phasor"/> representing positive sequence.</param>
        /// <param name="phaseA">The <see cref="Phasor"/> representing phase A.</param>
        /// <param name="phaseB">The <see cref="Phasor"/> representing phase B.</param>
        /// <param name="phaseC">The <see cref="Phasor"/> representing phase C.</param>
        /// <param name="measuredFromNodeID">The <see cref="SynchrophasorFramework.Modeling.InternalID"/> of the <see cref="MeasuredFromNode"/>.</param>
        /// <param name="measuredToNodeID">The <see cref="SynchrophasorFramework.Modeling.InternalID"/> of the <see cref="MeasuredToNode"/>.</param>
        public CurrentPhasorGroup(int internalID, int number, string name, string description, Phasor positiveSequence, Phasor phaseA, Phasor phaseB, Phasor phaseC, int measuredFromNodeID, int measuredToNodeID)
            : this(internalID, number, name, description, positiveSequence, phaseA, phaseB, phaseC, new StatusWord(), measuredFromNodeID, measuredToNodeID)
        {
        }

        /// <summary>
        /// A constructor which specifies the information required by the <see cref="INetworkDescriptor"/> interface and the <see cref="Phasor"/> objects for +, A, B, & C, and the <see cref="StatusWord"/> 
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="name">A descriptive name for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="description">A description of the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="positiveSequence">The <see cref="Phasor"/> representing positive sequence.</param>
        /// <param name="phaseA">The <see cref="Phasor"/> representing phase A.</param>
        /// <param name="phaseB">The <see cref="Phasor"/> representing phase B.</param>
        /// <param name="phaseC">The <see cref="Phasor"/> representing phase C.</param>
        /// <param name="statusWord">The <see cref="StatusWord"/> from the source device for the <see cref="Phasor"/> objects in this <see cref="PhasorGroup"/>.</param>
        public CurrentPhasorGroup(int internalID, int number, string name, string description, Phasor positiveSequence, Phasor phaseA, Phasor phaseB, Phasor phaseC, StatusWord statusWord)
            :this(internalID, number, name, description, positiveSequence, phaseA, phaseB, phaseC, statusWord, 0, 0)
        {
        }

        /// <summary>
        /// The designated constructor for the <see cref="CurrentPhasorGroup"/> class which specifies the information required by the <see cref="INetworkDescriptor"/> interface and the <see cref="Phasor"/> objects for +, A, B, & C, the <see cref="StatusWord"/> and the <see cref="SynchrophasorFramework.Modeling.Node.InternalID"/> of the <see cref="MeasuredFromNode"/> and <see cref="MeasuredToNode"/>.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="name">A descriptive name for the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="description">A description of the <see cref="CurrentPhasorGroup"/> object.</param>
        /// <param name="positiveSequence">The <see cref="Phasor"/> representing positive sequence.</param>
        /// <param name="phaseA">The <see cref="Phasor"/> representing phase A.</param>
        /// <param name="phaseB">The <see cref="Phasor"/> representing phase B.</param>
        /// <param name="phaseC">The <see cref="Phasor"/> representing phase C.</param>
        /// <param name="statusWord">The <see cref="StatusWord"/> from the source device for the <see cref="Phasor"/> objects in this <see cref="PhasorGroup"/>.</param>
        /// <param name="measuredFromNodeID">The <see cref="SynchrophasorFramework.Modeling.InternalID"/> of the <see cref="MeasuredFromNode"/>.</param>
        /// <param name="measuredToNodeID">The <see cref="SynchrophasorFramework.Modeling.InternalID"/> of the <see cref="MeasuredToNode"/>.</param>
        public CurrentPhasorGroup(int internalID, int number, string name, string description, Phasor positiveSequence, Phasor phaseA, Phasor phaseB, Phasor phaseC, StatusWord statusWord, int measuredFromNodeID, int measuredToNodeID)
            : base(internalID, number, name, description, positiveSequence, phaseA, phaseB, phaseC, statusWord)
        {
            m_measuredFromNodeID = measuredFromNodeID;
            m_measuredToNodeID = measuredToNodeID;

            PositiveSequence.Measurement.Type = PhasorType.CurrentPhasor;
            PositiveSequence.Estimate.Type = PhasorType.CurrentPhasor;

            NegativeSequence.Measurement.Type = PhasorType.CurrentPhasor;
            NegativeSequence.Estimate.Type = PhasorType.CurrentPhasor;

            ZeroSequence.Measurement.Type = PhasorType.VoltagePhasor;
            ZeroSequence.Estimate.Type = PhasorType.VoltagePhasor;

            PhaseA.Measurement.Type = PhasorType.CurrentPhasor;
            PhaseA.Estimate.Type = PhasorType.CurrentPhasor;

            PhaseB.Measurement.Type = PhasorType.CurrentPhasor;
            PhaseB.Estimate.Type = PhasorType.CurrentPhasor;

            PhaseC.Measurement.Type = PhasorType.CurrentPhasor;
            PhaseC.Estimate.Type = PhasorType.CurrentPhasor;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="CurrentPhasorGroup"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="CurrentPhasorGroup"/> class.</returns>
        public override string ToString()
        {
            return "-IPHGRP- ID: " + InternalID.ToString() + " Acronym: " + Acronym + " Name: " + Name + " Description: " + Description + " FromNode: " + m_measuredFromNodeID.ToString() + " ToNode: " + m_measuredToNodeID.ToString();
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="CurrentPhasorGroup"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="CurrentPhasorGroup"/> class.</returns>
        public new string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Current Phasor Group -----------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("         Acronym: " + Acronym + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("       UseStatus: " + UseStatusFlagForRemovingMeasurements.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Status: " + Status.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PosSeqMKeys: " + PositiveSequence.Measurement.MagnitudeKey + " | " + PositiveSequence.Measurement.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PosSeqEKeys: " + PositiveSequence.Estimate.MagnitudeKey + " | " + PositiveSequence.Estimate.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseAMKeys: " + PhaseA.Measurement.MagnitudeKey + " | " + PhaseA.Measurement.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseAEKeys: " + PhaseA.Estimate.MagnitudeKey + " | " + PhaseA.Estimate.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseBMKeys: " + PhaseB.Measurement.MagnitudeKey + " | " + PhaseB.Measurement.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseBEKeys: " + PhaseB.Estimate.MagnitudeKey + " | " + PhaseB.Estimate.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseCMKeys: " + PhaseC.Measurement.MagnitudeKey + " | " + PhaseC.Measurement.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     PhaseCEKeys: " + PhaseC.Estimate.MagnitudeKey + " | " + PhaseC.Estimate.AngleKey + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("MeasuredFromNode: " + MeasuredFromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("  MeasuredToNode: " + MeasuredToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        ///// <summary>
        ///// Overridden to prevent compilation warnings.
        ///// </summary>
        ///// <returns>An integer hash code determined by the base class.</returns>
        //public override int GetHashCode()
        //{
        //    return base.GetHashCode();
        //}

        ///// <summary>
        ///// Determines equality between two <see cref="CurrentPhasorGroup"/> objects
        ///// </summary>
        ///// <param name="target">The target object for equality testing.</param>
        ///// <returns>A bool representing the result of the equality check.</returns>
        //public override bool Equals(object target)
        //{
        //    // If parameter is null return false.
        //    if (target == null)
        //    {
        //        return false;
        //    }

        //    // If parameter cannot be cast to CurrentPhasorGroup return false.
        //    CurrentPhasorGroup currentPhasorGroup = target as CurrentPhasorGroup;

        //    if ((object)currentPhasorGroup == null)
        //    {
        //        return false;
        //    }

        //    // Return true if the fields match:
        //    if (!base.Equals(currentPhasorGroup))
        //    {
        //        return false;
        //    }
        //    else if (this.m_measuredFromNodeID != currentPhasorGroup.MeasuredFromNodeID)
        //    {
        //        return false;
        //    }
        //    else if (this.m_measuredToNodeID != currentPhasorGroup.MeasuredToNodeID)
        //    {
        //        return false;
        //    }
        //    else
        //    {
        //        return true;
        //    }
        //}

        /// <summary>
        /// Performs a deep copy of the <see cref="CurrentPhasorGroup"/> object.
        /// </summary>
        /// <returns>A deep copy of the target <see cref="CurrentPhasorGroup"/>.</returns>
        public CurrentPhasorGroup Copy()
        {
            CurrentPhasorGroup copy = (CurrentPhasorGroup)this.MemberwiseClone();
            copy.Status = this.Status.Copy();
            copy.PositiveSequence = PositiveSequence.Copy();
            copy.PhaseA = PhaseA.Copy();
            copy.PhaseB = PhaseB.Copy();
            copy.PhaseC = PhaseC.Copy();
            return copy;
        }

        #endregion
    }
}
