﻿//******************************************************************************************************
//  NetworkMeasurements.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.IO;

namespace SynchrophasorAnalytics.Measurements
{
    [Serializable()]
    public class NetworkMeasurements : IClearable
    {
        #region [ Private Members ] 

        //private Dictionary<int, VoltagePhasorGroup> m_voltageMeasurements;
        //private Dictionary<int, CurrentPhasorGroup> m_currentMeasurements;
        //private Dictionary<int, BreakerStatus> m_breakerStatuses;
        //private Dictionary<int, StatusWord> m_statusWords;

        private List<VoltagePhasorGroup> m_voltages;
        private List<CurrentPhasorGroup> m_currents;
        private List<BreakerStatus> m_breakerStatuses;
        private List<StatusWord> m_statusWords;

        #endregion

        #region [ Properties ]
        
        //[XmlIgnore()]
        //public Dictionary<int, VoltagePhasorGroup> VoltageMeasurements
        //{
        //    get { return m_voltageMeasurements; }
        //    set { m_voltageMeasurements = value; }
        //}

        //[XmlIgnore()]
        //public Dictionary<int, CurrentPhasorGroup> CurrentMeasurements
        //{
        //    get { return m_currentMeasurements; }
        //    set { m_currentMeasurements = value; }
        //}

        //[XmlIgnore()]
        //public Dictionary<int, BreakerStatus> BreakerStatuses
        //{
        //    get { return m_breakerStatuses; }
        //    set { m_breakerStatuses = value; }
        //}

        //[XmlIgnore()]
        //public Dictionary<int, StatusWord> StatusWords
        //{
        //    get { return m_statusWords; }
        //    set { m_statusWords = value; }
        //}

        [XmlArray("Voltages")]
        public List<VoltagePhasorGroup> Voltages
        {
            get
            {
                return m_voltages;
            }
            set
            {
                m_voltages = value;
            }
        }

        [XmlArray("Currents")]
        public List<CurrentPhasorGroup> Currents
        {
            get
            {
                return m_currents;
            }
            set
            {
                m_currents = value;
            }
        }

        [XmlArray("BreakerStatuses")]
        public List<BreakerStatus> BreakerStatuses
        {
            get
            {
                return m_breakerStatuses;
            }
            set
            {
                m_breakerStatuses = value;
            }
        }

        [XmlArray("StatusWords")]
        public List<StatusWord> StatusWords
        {
            get
            {
                return m_statusWords;
            }
            set
            {
                m_statusWords = value;
            }
        }

        [XmlIgnore()]
        public List<string> InputMeasurementKeys
        {
            get
            {
                // Create an empty list to store the input measurement keys.
                List<string> inputMeasurementKeys = new List<string>();

                // Get all of the input measurements keys for the status words
                foreach (StatusWord statusWord in m_statusWords)
                {
                    inputMeasurementKeys.Add(statusWord.Key);
                }

                // Get all of the input measurement keys for the breaker statuses.
                // Because a single digital value can contain multiple statuses, don't add one twice.
                foreach (BreakerStatus breakerStatus in m_breakerStatuses)
                {
                    if (!inputMeasurementKeys.Contains(breakerStatus.Key))
                    {
                        inputMeasurementKeys.Add(breakerStatus.Key);
                    }
                }

                // Get all of the input measurement keys for the voltage phasors
                foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
                {
                    inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Measurement.AngleKey);

                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Measurement.AngleKey);

                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Measurement.AngleKey);

                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Measurement.AngleKey);
                }

                // Get all of the input measurement keys for the current phasors
                foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
                {
                    inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Measurement.AngleKey);

                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Measurement.AngleKey);

                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Measurement.AngleKey);

                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Measurement.MagnitudeKey);
                    inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Measurement.AngleKey);
                }

                return inputMeasurementKeys;
            }
        }

        [XmlIgnore()]
        public string InputMeasurementKeysForConnectionString
        {
            get
            {
                // Get all of the input measurement keys
                List<string> inputMeasurementKeys = InputMeasurementKeys;

                // Create a single string to hold the concatenation
                string inputMeasurementKeysForConnectionString = "";

                // Concatenate each of the input measurement keys for the connection string parameter
                foreach (string key in inputMeasurementKeys)
                {
                    inputMeasurementKeysForConnectionString += key + ";";
                }

                return inputMeasurementKeysForConnectionString;
            }
        }
        
        #endregion

        #region [ Constructors ]

        public NetworkMeasurements()
            :this(new List<VoltagePhasorGroup>(), new List<CurrentPhasorGroup>(), new List<BreakerStatus>(), new List<StatusWord>())
        {
        }

        public NetworkMeasurements(List<VoltagePhasorGroup> voltages, List<CurrentPhasorGroup> currents, List<BreakerStatus> breakerStatuses, List<StatusWord> statusWords)
        {
            m_voltages = voltages;
            m_currents = currents;
            m_breakerStatuses = breakerStatuses;
            m_statusWords = statusWords;
        }

        #endregion

        #region [ Methods ]

        public void OnNewMeasurements(Dictionary<string, double> rawMeasurements)  
        {
            InsertVoltageMeasurements(rawMeasurements);
            InsertCurrentMeasurements(rawMeasurements);
            InsertBreakerStatuses(rawMeasurements);
            InsertStatusWords(rawMeasurements);
        }

        public void Initialize()
        {
            MapStatusWordsToPhasorGroups();
        }

        /// <summary>
        /// Clears all the values of the network measurements by setting them to zero.
        /// </summary>
        public void ClearValues()
        {
            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                voltagePhasorGroup.ClearValues();
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                currentPhasorGroup.ClearValues();
            }

            //foreach (BreakerStatus breakerStatus in m_breakerStatuses.Values)
            //{
            //    breakerStatus.ClearValues();
            //}

            //foreach (StatusWord statusWord in m_statusWords.Values)
            //{
            //    statusWord.BinaryValue = 0;
            //}
        }

        public void SerializeData(bool serializeOption)
        {
            foreach (StatusWord statusWord in m_statusWords)
            {
                statusWord.ShouldSerializeData = serializeOption;
            }

            //foreach (BreakerStatus breakerStatus in m_breakers)
            //{
            //    breakerStatus
            //}

            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                voltagePhasorGroup.PositiveSequence.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PositiveSequence.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseA.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseA.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseB.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseB.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseC.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseC.Estimate.ShouldSerializeData = serializeOption;
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                currentPhasorGroup.PositiveSequence.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PositiveSequence.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseA.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseA.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseB.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseB.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseC.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseC.Estimate.ShouldSerializeData = serializeOption;
            }
        }

        public static NetworkMeasurements DeserializeFromXml(string pathName)
        {
            //try
            //{
                // Create an empy NetworkMeasurements object reference.
                NetworkMeasurements networkMeasurements = null;

                // Create an XmlSerializer with the type of NetworkMeasurements.
                XmlSerializer deserializer = new XmlSerializer(typeof(NetworkMeasurements));

                // Read the data in from the file.
                StreamReader reader = new StreamReader(pathName);

                // Cast the deserialized data as a NetworkMeasurements object.
                networkMeasurements = (NetworkMeasurements)deserializer.Deserialize(reader);

                // Close the connection.
                reader.Close();

                return networkMeasurements;
            //}
            //catch
            //{
            //    return null;
            //}

        }

        public void SerializeToXml(string pathName)
        {
            try
            {
                // Create an XmlSerializer with the type of NetworkMeasurements
                XmlSerializer serializer = new XmlSerializer(typeof(NetworkMeasurements));

                // Open a connection to the file and path.
                TextWriter writer = new StreamWriter(pathName);

                // Serialize this instance of NetworkMeasurements
                serializer.Serialize(writer, this);

                // Close the connection
                writer.Close();
            }
            catch (Exception exception)
            {
                Console.WriteLine(exception.InnerException);
            }
        }

        #endregion

        #region [ Private Methods ]

        private void MapStatusWordsToPhasorGroups()
        {
            Dictionary<int, StatusWord> statusWords = m_statusWords.ToDictionary(x => x.InternalID, x => x);

            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                StatusWord value = null;
                if (statusWords.TryGetValue(voltagePhasorGroup.StatusWordID, out value))
                {
                    voltagePhasorGroup.Status = value;
                    value = null;
                }
                else
                {
                    voltagePhasorGroup.Status = null;
                }
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                StatusWord value = null;
                if (statusWords.TryGetValue(currentPhasorGroup.StatusWordID, out value))
                {
                    currentPhasorGroup.Status = value;
                    value = null;
                }
                else
                {
                    currentPhasorGroup.Status = null;
                }
            }
        }

        /// <summary>
        /// Maps the raw voltage measurement values to their virtual counterpart
        /// </summary>
        /// <param name="rawMeasurements">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertVoltageMeasurements(Dictionary<string, double> rawMeasurements)
        {
            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                double value = 0;
                // Positive Sequence Magnitude
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PositiveSequence.Measurement.MagnitudeKey, out value))
                {
                    voltagePhasorGroup.PositiveSequence.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PositiveSequence.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Positive Sequence Angle
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PositiveSequence.Measurement.AngleKey, out value))
                {
                    voltagePhasorGroup.PositiveSequence.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PositiveSequence.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase A Magnitude
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseA.Measurement.MagnitudeKey, out value))
                {
                    voltagePhasorGroup.PhaseA.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseA.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase A Angle
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseA.Measurement.AngleKey, out value))
                {
                    voltagePhasorGroup.PhaseA.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseA.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase B Magnitude
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseB.Measurement.MagnitudeKey, out value))
                {
                    voltagePhasorGroup.PhaseB.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseB.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase B Angle
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseB.Measurement.AngleKey, out value))
                {
                    voltagePhasorGroup.PhaseB.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseB.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase C Magnitude
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseC.Measurement.MagnitudeKey, out value))
                {
                    voltagePhasorGroup.PhaseC.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseC.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase C Angle
                if (rawMeasurements.TryGetValue(voltagePhasorGroup.PhaseC.Measurement.AngleKey, out value))
                {
                    voltagePhasorGroup.PhaseC.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    voltagePhasorGroup.PhaseC.Measurement.AngleInDegrees = 0;
                    value = 0;
                }
            }
        }

        /// <summary>
        /// Maps the raw current measurement values to their virtual counterpart
        /// </summary>
        /// <param name="rawMeasurements">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertCurrentMeasurements(Dictionary<string, double> rawMeasurements)
        {
            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                double value = 0;
                // Positive Sequence Magnitude
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PositiveSequence.Measurement.MagnitudeKey, out value))
                {
                    currentPhasorGroup.PositiveSequence.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PositiveSequence.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Positive Sequence Angle
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PositiveSequence.Measurement.AngleKey, out value))
                {
                    currentPhasorGroup.PositiveSequence.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PositiveSequence.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase A Magnitude
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseA.Measurement.MagnitudeKey, out value))
                {
                    currentPhasorGroup.PhaseA.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseA.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase A Angle
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseA.Measurement.AngleKey, out value))
                {
                    currentPhasorGroup.PhaseA.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseA.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase B Magnitude
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseB.Measurement.MagnitudeKey, out value))
                {
                    currentPhasorGroup.PhaseB.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseB.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase B Angle
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseB.Measurement.AngleKey, out value))
                {
                    currentPhasorGroup.PhaseB.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseB.Measurement.AngleInDegrees = 0;
                    value = 0;
                }

                // Phase C Magnitude
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseC.Measurement.MagnitudeKey, out value))
                {
                    currentPhasorGroup.PhaseC.Measurement.Magnitude = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseC.Measurement.Magnitude = 0;
                    value = 0;
                }

                // Phase C Angle
                if (rawMeasurements.TryGetValue(currentPhasorGroup.PhaseC.Measurement.AngleKey, out value))
                {
                    currentPhasorGroup.PhaseC.Measurement.AngleInDegrees = value;
                    value = 0;
                }
                else
                {
                    currentPhasorGroup.PhaseC.Measurement.AngleInDegrees = 0;
                    value = 0;
                }
            }
        }

        /// <summary>
        /// Maps the raw digital values to their virtual counterpart
        /// </summary>
        /// <param name="rawMeasurements">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertBreakerStatuses(Dictionary<string, double> rawMeasurements)
        {
            foreach (BreakerStatus breakerStatus in m_breakerStatuses)
            {
                double value = 0;
                // Breaker Statuses
                if (rawMeasurements.TryGetValue(breakerStatus.Key, out value))
                {
                    breakerStatus.BinaryValue = Convert.ToInt32(value);
                    value = 0;
                }
                else
                {
                    //breakerStatus.BinaryValue = Convert.ToInt32(value);
                    value = 0;
                }
            }
        }

        /// <summary>
        /// Maps the raw status word values to their virtual counterpart
        /// </summary>
        /// <param name="rawMeasurements">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertStatusWords(Dictionary<string, double> rawMeasurements)
        {
            foreach (StatusWord statusWord in m_statusWords)
            {
                double value = 0;
                // Status Words
                if (rawMeasurements.TryGetValue(statusWord.Key, out value))
                {
                    statusWord.BinaryValue = value;
                    value = 0;
                }
                else
                {
                    //statusWord.BinaryValue = value;
                    value = 0;
                }
            }
        }

        #endregion

    }
}
