﻿//******************************************************************************************************
//  Phasor.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;

namespace SynchrophasorAnalytics.Measurements
{
    [Serializable()]
    public class Phasor
    {
        #region [ Private Members ]

        private PhasorMeasurement m_phasorMeasurement;
        private PhasorEstimate m_phasorEstimate;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// A <see cref="PhasorMeasurement"/> - a measured phasor value.
        /// </summary>
        [XmlElement("Measurement")]
        public PhasorMeasurement Measurement
        {
            get 
            { 
                return m_phasorMeasurement; 
            }
            set 
            { 
                m_phasorMeasurement = value; 
            }
        }

        /// <summary>
        /// A <see cref="PhasorEstimate"/>. - an estimated phasor value.
        /// </summary>
        [XmlElement("Estimate")]
        public PhasorEstimate Estimate
        {
            get 
            { 
                return m_phasorEstimate; 
            }
            set 
            { 
                m_phasorEstimate = value; 
            }
        }

        /// <summary>
        /// The residual of the measured and estimated phasor magnitude in line-to-neutral kV.
        /// </summary>
        [XmlIgnore]
        public double MagnitudeResidual
        {
            get 
            { 
                return m_phasorMeasurement.Magnitude - m_phasorEstimate.Magnitude; 
            }
        }

        /// <summary>
        /// The residual of the measured and estimated phasor magnitude in per unit.
        /// </summary>
        [XmlIgnore]
        public double PerUnitMagnitudeResidual
        {
            get 
            { 
                return m_phasorMeasurement.PerUnitMagnitude - m_phasorEstimate.PerUnitMagnitude; 
            }
        }

        /// <summary>
        /// The residual of the measured and estimated phasor angle in degrees.
        /// </summary>
        [XmlIgnore]
        public double AngleResidualInDegrees
        {
            get 
            { 
                return m_phasorMeasurement.AngleInDegrees - m_phasorEstimate.AngleInDegrees; 
            }
        }

        /// <summary>
        /// The residual of the measured and estimated phasor anglge in radians.
        /// </summary>
        [XmlIgnore]
        public double AngleResidualInRadians
        {
            get 
            { 
                return m_phasorMeasurement.AngleInRadians - m_phasorEstimate.AngleInRadians; 
            }
        }

        /// <summary>
        /// The statistical cost of this <see cref="Phasor"/>.
        /// </summary>
        [XmlIgnore]
        public double Cost
        {
            get
            {
                //if (m_phasorMeasurement.Type == PhasorMeasurementType.VoltagePhasor)
                //{
                //    double magnitudeCost = (PerUnitMagnitudeResidual * PerUnitMagnitudeResidual) / m_phasorMeasurement.MeasurementCovariance;
                //    double angleCost = (AngleResidualInDegrees * AngleResidualInDegrees) / 0.01;
                //    return magnitudeCost;// +angleCost;
                //}
                //else if (m_phasorMeasurement.Type == PhasorMeasurementType.CurrentPhasor)
                //{
                //    double magnitudeCost = ((PerUnitMagnitudeResidual / m_phasorMeasurement.PerUnitMagnitude) * (PerUnitMagnitudeResidual / m_phasorMeasurement.PerUnitMagnitude)) / m_phasorMeasurement.MeasurementCovariance;
                //    double angleCost = (AngleResidualInDegrees * AngleResidualInDegrees) / 0.01;
                //    return magnitudeCost;// +angleCost;
                //}
                //else
                //{
                //    return 99999;
                //}
                throw new NotImplementedException();
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Phasor()
            :this(new PhasorMeasurement(), new PhasorEstimate())
        {
        }

        /// <summary>
        /// A constructor for the <see cref="Phasor"/> class which only specifies the <see cref="PhasorMeasurement"/>.
        /// </summary>
        /// <param name="measurement"></param>
        public Phasor(PhasorMeasurement measurement)
            :this(measurement, new PhasorEstimate())
        {
        }

        /// <summary>
        /// The designated constructor for the class.
        /// </summary>
        /// <param name="measurement">A measured phasor - <see cref="PhasorMeasurement"/>.</param>
        /// <param name="estimate">An estimated phasor - <see cref="PhasorEstimate"/>.</param>
        public Phasor(PhasorMeasurement measurement, PhasorEstimate estimate)
        {
            m_phasorMeasurement = measurement;
            m_phasorEstimate = estimate;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A descriptive string representation of the <see cref="Phasor"/> object
        /// </summary>
        /// <returns>A descriptive string representation of the <see cref="Phasor"/> object</returns>
        public override string ToString()
        {
            return m_phasorMeasurement.BaseKV.Name + " " + m_phasorMeasurement.Type.ToString() + " with keys: " + m_phasorMeasurement.MagnitudeKey + " " + m_phasorMeasurement.AngleKey;
        }

        /// <summary>
        /// A verbose descriptive string representation of the <see cref="Phasor"/> object.
        /// </summary>
        /// <returns>A verbose descriptive string representation of the <see cref="Phasor"/> object.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Phasor -------------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("     Measurement: " + m_phasorMeasurement.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        Estimate: " + m_phasorEstimate.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Performs a deep copy of the <see cref="Phasor"/> object.
        /// </summary>
        /// <returns>A deep copy of the target <see cref="Phasor"/> object.</returns>
        public Phasor Copy()
        {
            Phasor copy = new Phasor();
            copy.Measurement = m_phasorMeasurement.DeepCopy();
            copy.Estimate = m_phasorEstimate.Copy();
            return copy;

        }

        #endregion
    }
}
