﻿//******************************************************************************************************
//  CircuitBreaker.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Modeling
{
    public class CircuitBreaker : SwitchingDeviceBase
    {
        #region [ Private Members ]

        private BreakerStatus m_breakerStatus;
        private int m_breakerStatusID;

        private Substation m_parentSubstation;
        private int m_parentSubstationID;

        #endregion

        #region [ Properties ]

        [XmlIgnore()]
        public BreakerStatus Status
        {
            get 
            { 
                return m_breakerStatus; 
            }
            set 
            { 
                m_breakerStatus = value;
                m_breakerStatusID = value.InternalID;
            }
        }

        [XmlAttribute("Status")]
        public int StatusID
        {
            get
            {
                return m_breakerStatusID;
            }
            set
            {
                m_breakerStatusID = value;
            }
        }

        [XmlIgnore()]
        public Substation ParentSubstation
        {
            get
            {
                return m_parentSubstation;
            }
            set
            {
                m_parentSubstation = value;
                m_parentSubstationID = value.InternalID;
            }
        }

        [XmlAttribute("ParentSubstation")]
        public int ParentSubstationID
        {
            get
            {
                return m_parentSubstationID;
            }
            set
            {
                m_parentSubstationID = value;
            }
        }

        [XmlIgnore()]
        public bool IsOpen
        {
            get
            {
                if (InManualOverrideMode)
                {
                    if (ActualState == SwitchingDeviceActualState.Open)
                    {
                        return true;
                    }
                    else if (ActualState == SwitchingDeviceActualState.Closed)
                    {
                        return false;
                    }
                    else
                    {
                        return false;
                    }
                }
                else if (StatusID != 0)
                {
                    return Status.Value;
                }
                else
                {
                    if (NormalState.Equals(SwitchingDeviceNormalState.Open))
                    {
                        return true;
                    }
                    else if (NormalState.Equals(SwitchingDeviceNormalState.Closed))
                    {
                        return false;
                    }
                    else
                    {
                        return false;
                    }
                }
            }
        }

        [XmlIgnore()]
        public bool IsClosed
        {
            get
            {
                return !IsOpen;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public CircuitBreaker()
            :this(0, 0, "Undefined", "Undefined", SwitchingDeviceNormalState.Open)
        {
        }

        public CircuitBreaker(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState)
            :this(internalID, number, name, description,  normalState, 0, 0)
        {
        }

        public CircuitBreaker(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState, BreakerStatus status)
            :this(internalID, number, name, description, normalState, 0, 0, status)
        {
        }

        public CircuitBreaker(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState, int fromNodeID, int toNodeID)
            :this(internalID, number, name, description, normalState, fromNodeID, toNodeID, new BreakerStatus())
        {
        }

        public CircuitBreaker(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState, int fromNodeID, int toNodeID, BreakerStatus status)
            :base(internalID, number, name, description, normalState, fromNodeID, toNodeID)
        {
            m_breakerStatus = status;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the <see cref="CircuitBreaker"/>.
        /// </summary>
        /// <returns>A string representation of the <see cref="CircuitBreaker"/>.</returns>
        public override string ToString()
        {
            return "-CB- ID: " + InternalID.ToString() + " Number: " + Number.ToString() + " Name: " + Name + " Description: " + Description;
        }

        /// <summary>
        /// A verbose string representation of the <see cref="CircuitBreaker"/>.
        /// </summary>
        /// <returns>A verbose string representation of the <see cref="CircuitBreaker"/>.</returns>
        public new string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Circuit Breaker ----------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        Normally: " + NormalState.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        Actually: " + ActualState.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Manual: " + InManualOverrideMode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("ParentSubstation: " + m_parentSubstation.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Status: {0}", Environment.NewLine);
            stringBuilder.AppendFormat(m_breakerStatus.ToStatusString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        #endregion
    }
}
