﻿//******************************************************************************************************
//  Company.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;

namespace SynchrophasorAnalytics.Modeling
{
    /// <summary>
    /// 
    /// </summary>
    [Serializable()]
    public class Company : INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_acronym;
        private string m_name;
        private string m_description;
        private List<Division> m_childrenDivisions;

        #endregion

        #region [ Properties ] 

        /// <summary>
        /// The unique integer identifier for each <see cref="Company"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            { 
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="Company"/>. 
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="Company"/>.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return m_acronym; 
            }
            set 
            { 
                m_acronym = value; 
            }
        }

        /// <summary>
        /// The name of the <see cref="Company"/>.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="Company"/>.
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// Returns the type of the object.
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// All of the <see cref="Division"/> objects owned by the <see cref="Company"/>.
        /// </summary>
        [XmlArray("Divisions")]
        public List<Division> Divisions
        {
            get 
            { 
                return m_childrenDivisions; 
            }
            set 
            { 
                m_childrenDivisions = value;
                foreach (Division division in m_childrenDivisions)
                {
                    division.ParentCompany = this;
                }
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Company()
            :this(0, 0, "IDK", "Undefined Name", "Undefined Description")
        {
        }

        /// <summary>
        /// A constructor for the <see cref="Company"/> class which requires all of the properties defined by the <see cref="INetworkDescribable"/> interface.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Company"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Company"/>.</param>
        /// <param name="acronym">A descriptive number for the <see cref="Company"/>.</param>
        /// <param name="name">The name of the <see cref="Company"/>.</param>
        /// <param name="description">A description of the <see cref="Company"/>.</param>
        public Company(int internalID, int number, string acronym, string name, string description)
            :this(internalID, number, acronym, name, description, new List<Division>())
        {
        }

        /// <summary>
        /// The designated initializer for the <see cref="Company"/> class. Requires all of the properties defined by the <see cref="INetworkDescribable"/> interface and a list of its children <see cref="Division"/> objects.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Company"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Company"/>.</param>
        /// <param name="acronym">A descriptive number for the <see cref="Company"/>.</param>
        /// <param name="name">The name of the <see cref="Company"/>.</param>
        /// <param name="description">A description of the <see cref="Company"/>.</param>
        /// <param name="divisions">A list of all of the children <see cref="Division"/> objects.</param>
        public Company(int internalID, int number, string acronym, string name, string description, List<Division> divisions)
        {
            m_internalID = internalID;
            m_number = number;
            m_acronym = acronym;
            m_name = name;
            m_description = description;
            m_childrenDivisions = divisions;
            foreach (Division division in m_childrenDivisions)
            {
                division.ParentCompany = this;
            }
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A descriptive string representation of the class instance.
        /// </summary>
        /// <returns>A descriptive string representation of the class instance.</returns>
        public override string ToString()
        {
            return "-CO- ID: " + m_internalID.ToString() + " Acronym: " + m_acronym + " Name: " + m_name;
        }

        /// <summary>
        /// A verbose descriptive string representation of the class instance.
        /// </summary>
        /// <returns>A verbose descriptive string representation of the class instance.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Company ------------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + m_internalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + m_number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("         Acronym: " + m_acronym + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + m_name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + m_description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("       Divisions:  {0}", Environment.NewLine);
            foreach (Division division in m_childrenDivisions)
            {
                stringBuilder.AppendFormat("        " + division.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Determines equality between to <see cref="Company"/> objects. Child <see cref="Division"/> objects
        /// are not included in the equality check.
        /// </summary>
        /// <param name="target">The target object for equality testing.</param>
        /// <returns>A bool representing the result of the equality check.</returns>
        public override bool Equals(object target)
        {
            // If parameter is null return false.
            if (target == null)
            {
                return false;
            }

            // If parameter cannot be cast to PhasorBase return false.
            Company company = target as Company;

            if ((object)company == null)
            {
                return false;
            }

            // Return true if the fields match:
            if (this.m_internalID != company.InternalID)
            {
                return false;
            }
            else if (this.m_number != company.Number)
            {
                return false;
            }
            else if (!this.m_acronym.Equals(company.Acronym))
            {
                return false;
            }
            else if (!this.m_name.Equals(company.Name))
            {
                return false;
            }
            else if (!this.m_description.Equals(company.Description))
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// <summary>
        /// Overridden to prevent compilation warnings
        /// </summary>
        /// <returns>An integer hash code of the base class</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Performs a deep copy of the target <see cref="Company"/> object.
        /// </summary>
        /// <returns>A deep copy of the target <see cref="Company"/> object.</returns>
        public Company DeepCopy()
        {
            Company copy = (Company)this.MemberwiseClone();

            List<Division> copyDivisions = new List<Division>();

            foreach (Division division in m_childrenDivisions)
            {
                copyDivisions.Add(division.DeepCopy());
            }

            foreach (Division division in copyDivisions)
            {
                division.ParentCompany = copy;
            }

            copy.Divisions = copyDivisions;

            return copy;
        }
        
        #endregion
    }
}
