﻿//******************************************************************************************************
//  Division.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class Division : INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_acronym;
        private string m_name;
        private string m_description;
        private Company m_parentCompany;
        private List<Substation> m_childrenSubstations;
        private List<TransmissionLine> m_childrenTransmissionLines;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The unique integer identifier for each <see cref="Division"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            {
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="Division"/>. 
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="Division"/>.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return m_acronym; 
            }
            set 
            { 
                m_acronym = value; 
            }
        }

        /// <summary>
        /// The name of the <see cref="Division"/>.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="Division"/>.
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// Returns the type of the object.
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The parent <see cref="Company"/> of the <see cref="Divsion"/>.
        /// </summary>
        [XmlIgnore()]
        public Company ParentCompany
        {
            get 
            { 
                return m_parentCompany; 
            }
            set 
            { 
                m_parentCompany = value; 
            }
        }

        /// <summary>
        /// All of the <see cref="Substation"/> objects owned by the <see cref="Division"/>.
        /// </summary>
        [XmlArray("Substations")]
        public List<Substation> Substations
        {
            get 
            { 
                return m_childrenSubstations; 
            }
            set 
            { 
                m_childrenSubstations = value;
                foreach (Substation substation in m_childrenSubstations)
                {
                    substation.ParentDivision = this;
                }
            }
        }

        /// <summary>
        /// All of the <see cref="TransmissionLine"/> objects owned by the <see cref="Division"/>.
        /// </summary>
        [XmlArray("TransmissionLines")]
        public List<TransmissionLine> TransmissionLines
        {
            get 
            { 
                return m_childrenTransmissionLines; 
            }
            set 
            { 
                m_childrenTransmissionLines = value;
                foreach (TransmissionLine transmissionLine in m_childrenTransmissionLines)
                {
                    transmissionLine.ParentDivision = this;
                }
            }
        }

        #endregion

        #region [ Constructors ]
        
        /// <summary>
        /// A blank constructor with default values
        /// </summary>
        public Division()
            :this(0, 0, "IDK", "Undefined Name", "Undefined Description")
        {
        }

        /// <summary>
        /// A constructor for the <see cref="Division"/> class which requires all of the properties defined by the <see cref="INetworkDescribable"/> interface.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Division"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Division"/>.</param>
        /// <param name="acronym">A descriptive number for the <see cref="Division"/>.</param>
        /// <param name="name">The name of the <see cref="Division"/>.</param>
        /// <param name="description">A description of the <see cref="Division"/>.</param>
        public Division(int internalID, int number, string acronym, string name, string description)
            :this(internalID, number, acronym, name, description, new List<Substation>(), new List<TransmissionLine>())
        {
        }

        /// <summary>
        /// The designated initializer for the <see cref="Division"/> class. Requires all of the properties defined by the <see cref="INetworkDescribable"/> interface and a list of its children <see cref="Substation"/> and <see cref="TransmissionLine"/> objects.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Division"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Division"/>.</param>
        /// <param name="acronym">A descriptive number for the <see cref="Division"/>.</param>
        /// <param name="name">The name of the <see cref="Division"/>.</param>
        /// <param name="description">A description of the <see cref="Division"/>.</param>
        /// <param name="substations">A list of all of the children <see cref="Substation"/> objects of the <see cref="Division"/>.</param>
        /// <param name="transmissionLines">A list of all of the children <see cref="TransmissionLine"/> objects of the <see cref="Division"/>.</param>
        public Division(int internalID, int number, string acronym, string name, string description, List<Substation> substations, List<TransmissionLine> transmissionLines)
        {
            m_internalID = internalID;
            m_number = number;
            m_acronym = acronym;
            m_name = name;
            m_description = description;

            m_childrenSubstations = substations;
            foreach (Substation substation in m_childrenSubstations)
            {
                substation.ParentDivision = this;
            }

            m_childrenTransmissionLines = transmissionLines;
            foreach (TransmissionLine transmissionLine in m_childrenTransmissionLines)
            {
                transmissionLine.ParentDivision = this;
            }
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="Division"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="Division"/> class.</returns>
        public override string ToString()
        {
            return "-DIV- ID: " + m_internalID.ToString() + " Acronym: " + m_acronym + " Name: " + m_name;
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="Division"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="Division"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Division -----------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + m_internalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + m_number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("         Acronym: " + m_acronym + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + m_name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + m_description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("   ParentCompany: " + m_parentCompany.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Substations:  {0}", Environment.NewLine);
            foreach (Substation substation in m_childrenSubstations)
            {
                stringBuilder.AppendFormat("        " + substation.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendFormat("TrnsmissionLines:  {0}", Environment.NewLine);
            foreach (TransmissionLine transmissionLine in m_childrenTransmissionLines)
            {
                stringBuilder.AppendFormat("        " + transmissionLine.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Determines equality between to <see cref="Division"/> objects. Child <see cref="Substation"/> and <see cref="TransmissionLine"/> objects
        /// are not included in the equality check.
        /// </summary>
        /// <param name="target">The target object for equality testing.</param>
        /// <returns>A bool representing the result of the equality check.</returns>
        public override bool Equals(object target)
        {
            // If parameter is null return false.
            if (target == null)
            {
                return false;
            }

            // If parameter cannot be cast to Division return false.
            Division division = target as Division;

            if ((object)division == null)
            {
                return false;
            }

            // Return true if the fields match:
            if (this.m_internalID != division.InternalID)
            {
                return false;
            }
            else if (this.m_number != division.Number)
            {
                return false;
            }
            else if (!this.m_acronym.Equals(division.Acronym))
            {
                return false;
            }
            else if (!this.m_name.Equals(division.Name))
            {
                return false;
            }
            else if (!this.m_description.Equals(division.Description))
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// <summary>
        /// OVerridden to prevent compilation warnings.
        /// </summary>
        /// <returns>An integer hash code determined by the base class.</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Performs a deep copy of the target <see cref="Division"/> object.
        /// </summary>
        /// <returns>A deep copy of the target <see cref="Division"/> object.</returns>
        public Division DeepCopy()
        {
            Division copy = (Division)this.MemberwiseClone();

            List<Substation> copySubstations = new List<Substation>();
            List<TransmissionLine> copyTransmissionLines = new List<TransmissionLine>();

            foreach (Substation substation in m_childrenSubstations)
            {
                copySubstations.Add(substation.DeepCopy());
            }

            foreach (Substation substation in copySubstations)
            {
                substation.ParentDivision = copy;
            }

            foreach (TransmissionLine transmissionLine in m_childrenTransmissionLines)
            {
                copyTransmissionLines.Add(transmissionLine.Copy());
            }

            foreach (TransmissionLine transmissionLine in copyTransmissionLines)
            {
                transmissionLine.ParentDivision = copy;
            }

            return copy;
        }

        #endregion
    }
}
