﻿//******************************************************************************************************
//  Impedance.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//  09/17/2013 - Kevin D. Jones
//       Added substration (-) operator for comparing Impedance for series compensation estimation.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using Extreme.Mathematics.LinearAlgebra.Providers;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class Impedance : IImpedance
    {
        #region [ Private Members ]

        #region [ Resistance ]
        private double m_r1;
        private double m_r2;
        private double m_r3;
        private double m_r4;
        private double m_r5;
        private double m_r6;
        #endregion
        #region [ Reactance ]
        private double m_x1;
        private double m_x2;
        private double m_x3;
        private double m_x4;
        private double m_x5;
        private double m_x6;
        #endregion
        #region [ Susceptance ]
        private double m_b1;
        private double m_b2;
        private double m_b3;
        private double m_b4;
        private double m_b5;
        private double m_b6;
        #endregion

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The complex (R + jX) series impedance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSeriesImpedance
        {
            get
            {
                DoubleComplex Zs = (ThreePhaseSeriesImpedance[0, 0] + ThreePhaseSeriesImpedance[1, 1] + ThreePhaseSeriesImpedance[2, 2]) / 3;
                DoubleComplex Zm = (ThreePhaseSeriesImpedance[0, 1] + ThreePhaseSeriesImpedance[0, 2] + ThreePhaseSeriesImpedance[1, 2]) / 3;
                return (Zs - Zm);
            }
        }

        /// <summary>
        /// The complex 1/(R + jX) series admittance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSeriesAdmittance
        {
            get 
            { 
                return (1 / PositiveSequenceSeriesImpedance); 
            }
        }

        /// <summary>
        /// The complex (G + jB) shunt susceptance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceShuntSusceptance
        {
            get
            {
                DoubleComplex Bs = (ThreePhaseShuntSusceptance[0, 0] + ThreePhaseShuntSusceptance[1, 1] + ThreePhaseShuntSusceptance[2, 2]) / 3;
                DoubleComplex Bm = (ThreePhaseShuntSusceptance[0, 1] + ThreePhaseShuntSusceptance[0, 2] + ThreePhaseShuntSusceptance[1, 2]) / 3;
                return (Bs - Bm);
            }
        }

        /// <summary>
        /// The complex (G + jB) shunt susceptance for a single side of the series branch (divided by 2)
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSingleSidedShuntSusceptance
        {
            get
            {
                DoubleComplex Bs = (ThreePhaseShuntSusceptance[0, 0] + ThreePhaseShuntSusceptance[1, 1] + ThreePhaseShuntSusceptance[2, 2]) / 3;
                DoubleComplex Bm = (ThreePhaseShuntSusceptance[0, 1] + ThreePhaseShuntSusceptance[0, 2] + ThreePhaseShuntSusceptance[1, 2]) / 3;
                return (Bs - Bm)/2;
            }
        }

        /// <summary>
        /// The three phase complex (R + jX) series impedance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSeriesImpedance
        {
            get
            {
                DoubleComplex a = new DoubleComplex(m_r1, m_x1);
                DoubleComplex b = new DoubleComplex(m_r2, m_x2);
                DoubleComplex c = new DoubleComplex(m_r3, m_x3);
                DoubleComplex d = new DoubleComplex(m_r4, m_x4);
                DoubleComplex e = new DoubleComplex(m_r5, m_x5);
                DoubleComplex f = new DoubleComplex(m_r6, m_x6);
                ComplexMatrix impedance = ComplexMatrix.Create(new DoubleComplex[,] 
                               {{(a),(b),(d)},
                                {(b),(c),(e)},
                                {(d),(e),(f)}});
                return impedance;
            }
        }

        /// <summary>
        /// The three phase complex 1/(R +jX) series admittance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSeriesAdmittance
        {
            get 
            {
                if (System.Environment.Is64BitOperatingSystem)
                {
                    Kernel.UseProvider(new ParallelX64LinearAlgebraProvider());
                }
                else
                {
                    //Kernel.UseProvider(new ParallelX86LinearAlgebraProvider());
                }
                return ThreePhaseSeriesImpedance.GetInverse();
            }
        }

        /// <summary>
        /// The three phase complex (G + JB) shunt susceptance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseShuntSusceptance
        {
            get
            {
                DoubleComplex a = new DoubleComplex(0, m_b1);
                DoubleComplex b = new DoubleComplex(0, m_b2);
                DoubleComplex c = new DoubleComplex(0, m_b3);
                DoubleComplex d = new DoubleComplex(0, m_b4);
                DoubleComplex e = new DoubleComplex(0, m_b5);
                DoubleComplex f = new DoubleComplex(0, m_b6);
                ComplexMatrix susceptance = ComplexMatrix.Create(new DoubleComplex[,] 
                               {{(a),(b),(d)},
                                {(b),(c),(e)},
                                {(d),(e),(f)}});
                return susceptance;
            }
        }

        /// <summary>
        /// The three phase complex (G + JB) shunt susceptance as a 3 x 3 complex matrix for a single side of the series branch. (divided by 2)
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSingleSidedShuntSusceptance
        {
            get
            {
                DoubleComplex a = new DoubleComplex(0, m_b1 / 2);
                DoubleComplex b = new DoubleComplex(0, m_b2 / 2);
                DoubleComplex c = new DoubleComplex(0, m_b3 / 2);
                DoubleComplex d = new DoubleComplex(0, m_b4 / 2);
                DoubleComplex e = new DoubleComplex(0, m_b5 / 2);
                DoubleComplex f = new DoubleComplex(0, m_b6 / 2);
                ComplexMatrix susceptance = ComplexMatrix.Create(new DoubleComplex[,] 
                               {{(a),(b),(d)},
                                {(b),(c),(e)},
                                {(d),(e),(f)}});
                return susceptance;
            }
        }

        /// <summary>
        /// The real part of the (1, 1) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R1")]
        public double R1
        {
            get
            {
                return m_r1;
            }
            set
            {
                m_r1 = value;
            }
        }

        /// <summary>
        /// The real part of the (1, 2) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R2")]
        public double R2
        {
            get
            {
                return m_r2;
            }
            set
            {
                m_r2 = value;
            }
        }

        /// <summary>
        /// The real part of the (2, 2) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R3")]
        public double R3
        {
            get
            {
                return m_r3;
            }
            set
            {
                m_r3 = value;
            }
        }

        /// <summary>
        /// The real part of the (1, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R4")]
        public double R4
        {
            get
            {
                return m_r4;
            }
            set
            {
                m_r4 = value;
            }
        }

        /// <summary>
        /// The real part of the (2, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R5")]
        public double R5
        {
            get
            {
                return m_r5;
            }
            set
            {
                m_r5 = value;
            }
        }

        /// <summary>
        /// The real part of the (3, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("R6")]
        public double R6
        {
            get
            {
                return m_r6;
            }
            set
            {
                m_r6 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 1) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X1")]
        public double X1
        {
            get
            {
                return m_x1;
            }
            set
            {
                m_x1 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 2) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X2")]
        public double X2
        {
            get
            {
                return m_x2;
            }
            set
            {
                m_x2 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (2, 2) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X3")]
        public double X3
        {
            get
            {
                return m_x3;
            }
            set
            {
                m_x3 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X4")]
        public double X4
        {
            get
            {
                return m_x4;
            }
            set
            {
                m_x4 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (2, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X5")]
        public double X5
        {
            get
            {
                return m_x5;
            }
            set
            {
                m_x5 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (3, 3) element of the 3x3 impedance matrix. 
        /// </summary>
        [XmlAttribute("X6")]
        public double X6
        {
            get
            {
                return m_x6;
            }
            set
            {
                m_x6 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 1) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B1")]
        public double B1
        {
            get
            {
                return m_b1;
            }
            set
            {
                m_b1 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 2) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B2")]
        public double B2
        {
            get
            {
                return m_b2;
            }
            set
            {
                m_b2 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (2, 2) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B3")]
        public double B3
        {
            get
            {
                return m_b3;
            }
            set
            {
                m_b3 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (1, 3) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B4")]
        public double B4
        {
            get
            {
                return m_b4;
            }
            set
            {
                m_b4 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (2, 3) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B5")]
        public double B5
        {
            get
            {
                return m_b5;
            }
            set
            {
                m_b5 = value;
            }
        }

        /// <summary>
        /// The imaginary part of the (3, 3) element of the 3x3 shunt susceptance matrix. 
        /// </summary>
        [XmlAttribute("B6")]
        public double B6
        {
            get
            {
                return m_b6;
            }
            set
            {
                m_b6 = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default (zero) values.
        /// </summary>
        public Impedance()
            :this(new double[6] {0, 0, 0, 0, 0, 0}, new double[6] {0, 0, 0, 0, 0, 0}, new double[6] {0, 0, 0, 0, 0, 0})
        {
        }

        /// <summary>
        /// The designated constructor for the <see cref="Impedance"/> class.
        /// </summary>
        /// <param name="resistance">A double array of the 6 unique resistance elements in the 3x3 impedance matrix.</param>
        /// <param name="reactance">A double array of the 6 unique reactance elements in the 3x3 impedance matrix.</param>
        /// <param name="susceptance">A double array of the 6 unique susceptance elements in the 3x3 shunt susceptance matrix.</param>
        public Impedance(double[] resistance, double[] reactance, double[] susceptance)
        {
            m_r1 = resistance[0];
            m_r2 = resistance[1];
            m_r3 = resistance[2];
            m_r4 = resistance[3];
            m_r5 = resistance[4];
            m_r6 = resistance[5];

            m_x1 = reactance[0];
            m_x2 = reactance[1];
            m_x3 = reactance[2];
            m_x4 = reactance[3];
            m_x5 = reactance[4];
            m_x6 = reactance[5];

            m_b1 = susceptance[0];
            m_b2 = susceptance[1];
            m_b3 = susceptance[2];
            m_b4 = susceptance[3];
            m_b5 = susceptance[4];
            m_b6 = susceptance[5];
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of an instance of the <see cref="Impedance"/> class.
        /// </summary>
        /// <returns>A string representation of an instance of the <see cref="Impedance"/> class.</returns>
        public override string ToString()
        {
            return "-IMP- " + m_r1.ToString() + "," + m_r2.ToString() + "," + m_r3.ToString() + "," + m_r4.ToString() + "," + m_r5.ToString() + "," + m_r6.ToString() + "," + m_x1.ToString() + "," + m_x2.ToString() + "," + m_x3.ToString() + "," + m_x4.ToString() + "," + m_x5.ToString() + "," + m_x6.ToString() + "," + m_b1.ToString() + "," + m_b2.ToString() + "," + m_b3.ToString() + "," + m_b4.ToString() + "," + m_b5.ToString() + "," + m_b6.ToString();
        }

        /// <summary>
        /// A verbose string representation of an instance of the <see cref="Impedance"/> class.
        /// </summary>
        /// <returns>A verbose string representation of an instance of the <see cref="Impedance"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Impedance ----------------------------------------------------------------");
            stringBuilder.AppendLine();
            if (m_x1 < 0)
            {
                stringBuilder.AppendFormat(" (1, 1) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r1.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x1).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (1, 1) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r1.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded( Math.Abs(m_x1).ToString()) + "{0}", Environment.NewLine);
            }
            if (m_x2 < 0)
            {
                stringBuilder.AppendFormat(" (1, 2) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r2.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded( Math.Abs(m_x2).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (1, 2) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r2.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x2).ToString()) + "{0}", Environment.NewLine);
            }
            if (m_x3 < 0)
            {
                stringBuilder.AppendFormat(" (2, 2) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r3.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x3).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (2, 2) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r3.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded( Math.Abs(m_x3).ToString()) + "{0}", Environment.NewLine);
            }
            if (m_x4 < 0)
            {
                stringBuilder.AppendFormat(" (1, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r4.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x4).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (1, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r4.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x4).ToString()) + "{0}", Environment.NewLine);
            }
            if (m_x5 < 0)
            {
                stringBuilder.AppendFormat(" (2, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r5.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded( Math.Abs(m_x5).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (2, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r5.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x5).ToString()) + "{0}", Environment.NewLine);
            }
            if (m_x6 < 0)
            {
                stringBuilder.AppendFormat(" (3, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r6.ToString()) + " - j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x6).ToString()) + "{0}", Environment.NewLine);
            }
            else
            {
                stringBuilder.AppendFormat(" (3, 3) | " + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(m_r6.ToString()) + " + j" + TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(Math.Abs(m_x6).ToString()) + "{0}", Environment.NewLine);
            }

            stringBuilder.AppendLine();

            return stringBuilder.ToString();
        }

        /// <summary>
        /// Determines equality between to <see cref="Impedance"/> objects.
        /// </summary>
        /// <param name="target">The target object for equality testing.</param>
        /// <returns>A bool representing the result of the equality check.</returns>
        public override bool Equals(object target)
        {
            // If parameter is null return false.
            if (target == null)
            {
                return false;
            }

            // If parameter cannot be cast to PhasorBase return false.
            Impedance impedance = target as Impedance;

            if ((object)impedance == null)
            {
                return false;
            }

            // Return true if the fields match:
            if (this.m_r1 != impedance.R1)
            {
                return false;
            }
            else if (this.m_r2 != impedance.R2)
            {
                return false;
            }
            else if (this.m_r3 != impedance.R3)
            {
                return false;
            }
            else if (this.m_r4 != impedance.R4)
            {
                return false;
            }
            else if (this.m_r5 != impedance.R5)
            {
                return false;
            }
            else if (this.m_r6 != impedance.R6)
            {
                return false;
            }
            else if (this.m_x1 != impedance.X1)
            {
                return false;
            }
            else if (this.m_x2 != impedance.X2)
            {
                return false;
            }
            else if (this.m_x3 != impedance.X3)
            {
                return false;
            }
            else if (this.m_x4 != impedance.X4)
            {
                return false;
            }
            else if (this.m_x5 != impedance.X5)
            {
                return false;
            }
            else if (this.m_x6 != impedance.X6)
            {
                return false;
            }
            else if (this.m_b1 != impedance.B1)
            {
                return false;
            }
            else if (this.m_b2 != impedance.B2)
            {
                return false;
            }
            else if (this.m_b3 != impedance.B3)
            {
                return false;
            }
            else if (this.m_b4 != impedance.B4)
            {
                return false;
            }
            else if (this.m_b5 != impedance.B5)
            {
                return false;
            }
            else if (this.m_b6 != impedance.B6)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// <summary>
        /// Overridden to prevent compilation warnings.
        /// </summary>
        /// <returns>An integer hash code determined by the base class.</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Performs a deep copy of the <see cref="Impedance"/> object.
        /// </summary>
        /// <returns>A deep copy of the <see cref="Impedance"/> object.</returns>
        public Impedance Copy()
        {
            return (Impedance)this.MemberwiseClone();
        }

        private string TrimImpedanceStringOrAddSuffixOfZerosIfNeeded(string value)
        {
            if (value.Count() > 9)
            {
                return value.Substring(0, 9);
            }
            else if (value.Count() == 9)
            {
                return value;
            }
            else if (value.Count() < 9)
            {
                int zerosToAdd = 9 - value.Count();
                for (int i = 0; i < zerosToAdd; i++)
                {
                    value += "0";
                }
                return value;
            }
            return value;
        }

        #endregion

        #region [ Operators ]

        public static Impedance operator + (Impedance z1, Impedance z2)
        {
            Impedance impedance = new Impedance();

            impedance.R1 = z1.R1 + z2.R1;
            impedance.R2 = z1.R2 + z2.R2;
            impedance.R3 = z1.R3 + z2.R3;
            impedance.R4 = z1.R4 + z2.R4;
            impedance.R5 = z1.R5 + z2.R5;
            impedance.R6 = z1.R6 + z2.R6;

            impedance.X1 = z1.X1 + z2.X1;
            impedance.X2 = z1.X2 + z2.X2;
            impedance.X3 = z1.X3 + z2.X3;
            impedance.X4 = z1.X4 + z2.X4;
            impedance.X5 = z1.X5 + z2.X5;
            impedance.X6 = z1.X6 + z2.X6;

            impedance.B1 = z1.B1 + z2.B1;
            impedance.B2 = z1.B2 + z2.B2;
            impedance.B3 = z1.B3 + z2.B3;
            impedance.B4 = z1.B4 + z2.B4;
            impedance.B5 = z1.B5 + z2.B5;
            impedance.B6 = z1.B6 + z2.B6;

            return impedance;
        }

        public static Impedance operator - (Impedance z1, Impedance z2)
        {
            Impedance impedance = new Impedance();

            impedance.R1 = z1.R1 - z2.R1;
            impedance.R2 = z1.R2 - z2.R2;
            impedance.R3 = z1.R3 - z2.R3;
            impedance.R4 = z1.R4 - z2.R4;
            impedance.R5 = z1.R5 - z2.R5;
            impedance.R6 = z1.R6 - z2.R6;

            impedance.X1 = z1.X1 - z2.X1;
            impedance.X2 = z1.X2 - z2.X2;
            impedance.X3 = z1.X3 - z2.X3;
            impedance.X4 = z1.X4 - z2.X4;
            impedance.X5 = z1.X5 - z2.X5;
            impedance.X6 = z1.X6 - z2.X6;

            impedance.B1 = z1.B1 - z2.B1;
            impedance.B2 = z1.B2 - z2.B2;
            impedance.B3 = z1.B3 - z2.B3;
            impedance.B4 = z1.B4 - z2.B4;
            impedance.B5 = z1.B5 - z2.B5;
            impedance.B6 = z1.B6 - z2.B6;

            return impedance;
        }

        public static Impedance operator / (Impedance z1, int scalar)
        {
            Impedance impedance = new Impedance();

            impedance.R1 = z1.R1 / scalar;
            impedance.R2 = z1.R2 / scalar;
            impedance.R3 = z1.R3 / scalar;
            impedance.R4 = z1.R4 / scalar;
            impedance.R5 = z1.R5 / scalar;
            impedance.R6 = z1.R6 / scalar;

            impedance.X1 = z1.X1 / scalar;
            impedance.X2 = z1.X2 / scalar;
            impedance.X3 = z1.X3 / scalar;
            impedance.X4 = z1.X4 / scalar;
            impedance.X5 = z1.X5 / scalar;
            impedance.X6 = z1.X6 / scalar;

            impedance.B1 = z1.B1 / scalar;
            impedance.B2 = z1.B2 / scalar;
            impedance.B3 = z1.B3 / scalar;
            impedance.B4 = z1.B4 / scalar;
            impedance.B5 = z1.B5 / scalar;
            impedance.B6 = z1.B6 / scalar;

            return impedance;
        }


        #endregion

    }
}
