﻿//******************************************************************************************************
//  LineSegment.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class LineSegment : SeriesBranchBase
    {
        #region [ Private Members ]

        private TransmissionLine m_parentTransmissionLine;
        private int m_parentTransmissionLineID;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The parent <see cref="TransmissionLine"/> of the <see cref="LineSegment"/>.
        /// </summary>
        [XmlIgnore()]
        public TransmissionLine ParentTransmissionLine
        {
            get
            {
                return m_parentTransmissionLine;
            }
            set
            {
                m_parentTransmissionLine = value;
                m_parentTransmissionLineID = m_parentTransmissionLine.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="TransmissionLine.InternalID"/> of the <see cref="ParentTransmissionLine"/>
        /// </summary>
        [XmlAttribute("ParentTransmissionLine")]
        public int ParentTransmissionLineID
        {
            get
            {
                return m_parentTransmissionLineID;
            }
            set
            {
                m_parentTransmissionLineID = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public LineSegment()
            :this(0, 0, "Undefined Name", "Undefined Description", new Impedance())
        {
        }

        public LineSegment(int internalID, int number, string name, string description, Impedance impedance)
            :this(internalID, number, name, description, impedance, 0)
        {
        }

        public LineSegment(int internalID, int number, string name, string description, Impedance impedance, int parentTransmissionLineID)
            :base(internalID, number, name, description, impedance)
        {
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="LineSegment"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="LineSegment"/> class.</returns>
        public override string ToString()
        {
            return "-LNSG- ID: " + InternalID.ToString() + " Number: " + Number.ToString() + " Line: " + m_parentTransmissionLine.Number.ToString();
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="LineSegment"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="LineSegment"/> class.</returns>
        public new string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Line Segment -------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("ParntTrnsmsnLine: " + m_parentTransmissionLine.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat(RawImpedanceParameters.ToVerboseString());
            stringBuilder.AppendLine();
            return stringBuilder.ToString();

        }

        #endregion
    }
}
