﻿//******************************************************************************************************
//  NetworkModel.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//  07/10/2013 - Kevin D. Jones
//       Fixed issue with substations modeled without either direct or indirect observability. Before 
//       would cause a matrix singularity which is wrong. Correct LSE should never have a singularity. 
//       The observability analysis now removes those nodes from the problem. 
//  07/20/2013 - Kevin D. Jones
//       Added InputMeasurementKeys property for validating input measurement keys in the LSE.
//       Added AcceptsEstimates and AcceptsMeasurements for measurement mapping optimizations and
//       InputMeasurementKeys property. Added optimization to measurement mapping in Insert() methods.
//  08/10/2013 - Kevin D. Jones
//       Added I2-I1 ratio measurement keys to output measurements property.
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using SynchrophasorAnalytics.Measurements;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Matrices;
using SynchrophasorAnalytics.Calibration;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class NetworkModel
    {
        #region [ Private Members ]

        private Dictionary<string, double> m_rawMeasurementKeyValuePairs;
        private PhaseSelection m_phaseSelection;
        private List<Company> m_childrenCompanies;
        private List<VoltageLevel> m_voltageLevels;
        private List<TapConfiguration> m_tapConfigurations;
        private List<VoltagePhasorGroup> m_voltages;
        private List<CurrentPhasorGroup> m_currents;
        private List<BreakerStatus> m_breakerStatuses;
        private List<StatusWord> m_statusWords;
        private List<ObservedBus> m_observedBusses;
        private List<CurrentPhasorGroup> m_activeCurrents;
        private List<SwitchingDeviceBase> m_switchingDevices;
        private bool m_acceptsMeasurements;
        private bool m_acceptsEstimates;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// A Dictionary with the raw measurements and their respective measurement keys.
        /// </summary>
        [XmlIgnore()]
        public Dictionary<string, double> RawMeasurementKeyValuePairs
        {
            get
            {
                return m_rawMeasurementKeyValuePairs;
            }
            set
            {
                m_rawMeasurementKeyValuePairs = value;
            }
        }

        /// <summary>
        /// A Dictionary with the raw estimates and their respective measurement keys.
        /// </summary>
        [XmlIgnore()]
        public Dictionary<string, double> RawEstimateKeyValuePairs
        {
            get
            {
                Dictionary<string, double> rawEstimateKeyValuePairs = new Dictionary<string, double>();

                foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
                {
                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.NegativeSequenceToPositiveSequenceRatioMeasurementKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.NegativeSequenceToPositiveSequenceRatioMeasurementKey, voltagePhasorGroup.EstimatedNegativeSequenceToPositiveSequenceRatio);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PositiveSequence.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PositiveSequence.Estimate.MagnitudeKey, voltagePhasorGroup.PositiveSequence.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PositiveSequence.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PositiveSequence.Estimate.AngleKey, voltagePhasorGroup.PositiveSequence.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseA.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseA.Estimate.MagnitudeKey, voltagePhasorGroup.PhaseA.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseA.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseA.Estimate.AngleKey, voltagePhasorGroup.PhaseA.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseB.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseB.Estimate.MagnitudeKey, voltagePhasorGroup.PhaseB.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseB.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseB.Estimate.AngleKey, voltagePhasorGroup.PhaseB.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseC.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseC.Estimate.MagnitudeKey, voltagePhasorGroup.PhaseC.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(voltagePhasorGroup.PhaseC.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(voltagePhasorGroup.PhaseC.Estimate.AngleKey, voltagePhasorGroup.PhaseC.Estimate.AngleInDegrees);
                    }
                }

                foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
                {
                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.NegativeSequenceToPositiveSequenceRatioMeasurementKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.NegativeSequenceToPositiveSequenceRatioMeasurementKey, currentPhasorGroup.EstimatedNegativeSequenceToPositiveSequenceRatio);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PositiveSequence.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PositiveSequence.Estimate.MagnitudeKey, currentPhasorGroup.PositiveSequence.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PositiveSequence.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PositiveSequence.Estimate.AngleKey, currentPhasorGroup.PositiveSequence.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseA.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseA.Estimate.MagnitudeKey, currentPhasorGroup.PhaseA.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseA.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseA.Estimate.AngleKey, currentPhasorGroup.PhaseA.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseB.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseB.Estimate.MagnitudeKey, currentPhasorGroup.PhaseB.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseB.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseB.Estimate.AngleKey, currentPhasorGroup.PhaseB.Estimate.AngleInDegrees);
                    }

                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseC.Estimate.MagnitudeKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseC.Estimate.MagnitudeKey, currentPhasorGroup.PhaseC.Estimate.Magnitude);
                    }
                    if (!rawEstimateKeyValuePairs.ContainsKey(currentPhasorGroup.PhaseC.Estimate.AngleKey))
                    {
                        rawEstimateKeyValuePairs.Add(currentPhasorGroup.PhaseC.Estimate.AngleKey, currentPhasorGroup.PhaseC.Estimate.AngleInDegrees);
                    }
                }

                foreach (SwitchingDeviceBase switchingDevice in m_switchingDevices)
                {
                    if (!rawEstimateKeyValuePairs.ContainsKey(switchingDevice.MeasurementKey))
                    {
                        rawEstimateKeyValuePairs.Add(switchingDevice.MeasurementKey, (int)switchingDevice.ActualState);
                    }
                }

                return rawEstimateKeyValuePairs;
            }
        }

        [XmlIgnore()]
        public List<string> InputMeasurementKeys
        {
            get
            {
                List<string> inputMeasurementKeys = new List<string>();

                foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
                {
                    if (m_acceptsMeasurements)
                    {
                        inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Measurement.MagnitudeKey);
                        inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Measurement.AngleKey);

                        if (m_phaseSelection == PhaseSelection.ThreePhase)
                        {
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Measurement.AngleKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Measurement.AngleKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Measurement.AngleKey);
                        }
                    }

                    if (m_acceptsEstimates)
                    {
                        inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Estimate.MagnitudeKey);
                        inputMeasurementKeys.Add(voltagePhasorGroup.PositiveSequence.Estimate.AngleKey);

                        if (m_phaseSelection == PhaseSelection.ThreePhase)
                        {
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseA.Estimate.AngleKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseB.Estimate.AngleKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(voltagePhasorGroup.PhaseC.Estimate.AngleKey);
                        }
                    }
                }

                foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
                {
                    if (m_acceptsMeasurements)
                    {
                        inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Measurement.MagnitudeKey);
                        inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Measurement.AngleKey);

                        if (m_phaseSelection == PhaseSelection.ThreePhase)
                        {
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Measurement.AngleKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Measurement.AngleKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Measurement.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Measurement.AngleKey);
                        }
                    }

                    if (m_acceptsEstimates)
                    {
                        inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Estimate.MagnitudeKey);
                        inputMeasurementKeys.Add(currentPhasorGroup.PositiveSequence.Estimate.AngleKey);

                        if (m_phaseSelection == PhaseSelection.ThreePhase)
                        {
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseA.Estimate.AngleKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseB.Estimate.AngleKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Estimate.MagnitudeKey);
                            inputMeasurementKeys.Add(currentPhasorGroup.PhaseC.Estimate.AngleKey);
                        }
                    }
                }

                foreach (BreakerStatus breakerStatus in m_breakerStatuses)
                {
                    inputMeasurementKeys.Add(breakerStatus.Key);
                }

                foreach (StatusWord statusWord in m_statusWords)
                {
                    inputMeasurementKeys.Add(statusWord.Key);
                }

                return inputMeasurementKeys;
            }
        }

        [XmlIgnore()]
        public string OutputMeasurementKeys
        {
            get
            {
                string outputMeasurementKeys = "outputMeasurements={";
                List<string> rawEstimateKeys = RawEstimateKeyValuePairs.Keys.ToList();

                foreach (string key in rawEstimateKeys)
                {
                    if (!key.Equals("Undefined"))
                    {
                        outputMeasurementKeys += key + ";";
                    }
                }
                outputMeasurementKeys += "}";

                return outputMeasurementKeys;
            }
        }

        /// <summary>
        /// Determines whether to treat the network as a <see cref="PhaseSelection.PositiveSequence"/> approximation or as a full <see cref="PhaseSelection.ThreePhase"/> representation.
        /// </summary>
        [XmlAttribute("PhaseConfiguration")]
        public PhaseSelection PhaseConfiguration
        {
            get
            {
                return m_phaseSelection;
            }
            set
            {
                m_phaseSelection = value;
            }
        }

        /// <summary>
        /// A flag that determines whether the model accepts measured values during OnNewMeasurements()
        /// </summary>
        [XmlAttribute("AcceptsMeasurements")]
        public bool AcceptsMeasurements
        {
            get
            {
                return m_acceptsMeasurements;
            }
            set
            {
                m_acceptsMeasurements = value;
            }
        }

        /// <summary>
        /// A flag that determines whether the model accepts estimated values during OnNewMeasurements()
        /// </summary>
        [XmlAttribute("AcceptsEstimates")]
        public bool AcceptsEstimates
        {
            get
            {
                return m_acceptsEstimates;
            }
            set
            {
                m_acceptsEstimates = value;
            }
        }

        /// <summary>
        /// A list of <see cref="Company"/> objects in the <see cref="NetworkModel"/>
        /// </summary>
        [XmlArray("Companies")]
        public List<Company> Companies
        {
            get 
            { 
                return m_childrenCompanies; 
            }
            set 
            { 
                m_childrenCompanies = value; 
            }
        }

        /// <summary>
        /// A list of <see cref="VoltageLevels"/> in the <see cref="NetworkModel"/>
        /// </summary>
        [XmlArray("VoltageLevels")]
        public List<VoltageLevel> VoltageLevels
        {
            get
            {
                return m_voltageLevels;
            }
            set
            {
                m_voltageLevels = value;
            }
        }

        [XmlArray("TapConfigurations")]
        public List<TapConfiguration> TapConfigurations
        {
            get
            {
                return m_tapConfigurations;
            }
            set
            {
                m_tapConfigurations = value;
            }
        }

        [XmlIgnore()]
        public List<VoltagePhasorGroup> Voltages
        {
            get
            {
                return m_voltages;
            }
            set
            {
                m_voltages = value;
            }
        }

        [XmlIgnore()]
        public List<CurrentPhasorGroup> Currents
        {
            get
            {
                return m_currents;
            }
            set
            {
                m_currents = value;
            }
        }

        [XmlArray("BreakerStatuses")]
        public List<BreakerStatus> BreakerStatuses
        {
            get
            {
                return m_breakerStatuses;
            }
            set
            {
                m_breakerStatuses = value;
            }
        }

        [XmlArray("StatusWords")]
        public List<StatusWord> StatusWords
        {
            get
            {
                return m_statusWords;
            }
            set
            {
                m_statusWords = value;
            }
        }

        [XmlIgnore()]
        public List<ObservedBus> ObservedBusses
        {
            get
            {
                return m_observedBusses;
            }
            set
            {
                m_observedBusses = value;
            }
        }

        [XmlIgnore()]
        public List<CurrentPhasorGroup> ActiveCurrentPhasors
        {
            get
            {
                return m_activeCurrents;
            }
            set
            {
                m_activeCurrents = value;
            }
        }

        #endregion

        #region [ Constructors ]

        public NetworkModel()
        {
            m_rawMeasurementKeyValuePairs = new Dictionary<string, double>();
            m_childrenCompanies = new List<Company>();
            m_voltageLevels = new List<VoltageLevel>();
            m_tapConfigurations = new List<TapConfiguration>();
            m_voltages = new List<VoltagePhasorGroup>();
            m_currents = new List<CurrentPhasorGroup>();
            m_breakerStatuses = new List<BreakerStatus>();
            m_statusWords = new List<StatusWord>();
            m_switchingDevices = new List<SwitchingDeviceBase>();
            m_activeCurrents = new List<CurrentPhasorGroup>();
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Reconstitues the references between parent, child, and sibling objects in 
        /// the <see cref="NetworkModel"/> as well as several other initialization functions.
        /// </summary>
        public void Initialize()
        {
            LinkTransmissionLineReferencesToSubstations();
            LinkHierarchicalReferences();
            LinkVoltageLevelReferences();
            LinkTapConfigurationReferences(); 
            ListVoltagePhasors(); 
            ListCurrentPhasors();
            ListSwitchingDevices();
            LinkBreakerStatusToCircuitBreakers();
            LinkStatusWordsToPhasorGroups();
            LinkVoltageLevelsToPhasorGroups();
            InitializeComplexPowerCalculations();
            m_observedBusses = new List<ObservedBus>();
            m_activeCurrents = new List<CurrentPhasorGroup>();
        }

        /// <summary>
        /// Clears the values before receiving a new set of measurements
        /// </summary>
        public void ClearValues()
        {
            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                voltagePhasorGroup.ClearValues();
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                currentPhasorGroup.ClearValues();
            }

            foreach (BreakerStatus breakerStatus in m_breakerStatuses)
            {
                breakerStatus.ClearValues();
            }

            foreach (StatusWord statusWord in m_statusWords)
            {
                statusWord.ClearValues();
            }
        }

        /// <summary>
        /// Called when a full set of measurements has been passed to the <see cref="NetworkModel"/>
        /// </summary>
        /// <param name="rawMeasurementKeyValuePairs"></param>
        public void OnNewMeasurements()
        {
            InsertVoltageMeasurements();
            InsertCurrentMeasurements();
            InsertBreakerStatuses();
            InsertStatusWords();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public void ResolveToObservedBusses()
        {
            m_observedBusses.Clear();

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        substation.InitializeSubstationGraph();
                        substation.Graph.ResolveDirectlyConnectedAdjacencies();
                        List<ObservedBus> substationObservedBusses = CheckObservability(substation.Graph.ResolveToObservedBusses());
                        if (substationObservedBusses != null)
                        {
                            foreach (ObservedBus observedBus in substationObservedBusses)
                            {
                                m_observedBusses.Add(observedBus);
                            }
                        }
                    }

                    //foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    //{
                    //    transmissionLine.InitializeTransmissionLineGraph();
                    //    List<ObservedBus> transmissionLineObservedBusses = transmissionLine.Graph.ResolveToObservedBusses();
                    //    foreach (ObservedBus observedBus in transmissionLineObservedBusses)
                    //    {
                    //        observedBusses.Add(observedBus);
                    //    }
                    //}
                }
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public void DetermineActiveCurrentPhasors()
        {
            m_activeCurrents.Clear();

            //if (m_currents != null && m_currents.Count != 0)
            //{
            if (m_phaseSelection == PhaseSelection.PositiveSequence)
            {
                foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
                {
                    if (currentPhasorGroup.IncludeInPositiveSequenceEstimator)
                    {
                        m_activeCurrents.Add(currentPhasorGroup);
                    }
                }
            }
            else if (m_phaseSelection == PhaseSelection.ThreePhase)
            {
                foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
                {
                    if (currentPhasorGroup.IncludeInEstimator)
                    {
                        m_activeCurrents.Add(currentPhasorGroup);
                    }
                }
            }
            //}
        }

        public void DetermineActiveCurrentPhasorsForCalibration()
        {

        }

        /// <summary>
        /// Computes the sequence component values for the measured and estimated phasors
        /// </summary>
        public void ComputeSequenceValues()
        {
            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                voltagePhasorGroup.ComputeSequenceComponents();
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                currentPhasorGroup.ComputeSequenceComponents();
            }
        }

        public bool ToggleSwitchingDeviceStatus(string switchingDeviceName, SwitchingDeviceActualState desiredActualState)
        {
            Dictionary<string, SwitchingDeviceBase> switchingDevices = m_switchingDevices.ToDictionary(x => x.Name, x => x);

            SwitchingDeviceBase switchingDevice = null;

            if (switchingDevices.TryGetValue(switchingDeviceName, out switchingDevice))
            {
                switchingDevice.ManuallySwitchTo(desiredActualState);

                return true;
            }
            else
            {
                return false;
            }

        }

        public bool RemoveSwitchingDeviceFromManualWithPreserveOrDefault(string switchingDeviceName, bool revertToDefaultState)
        {
            Dictionary<string, SwitchingDeviceBase> switchingDevices = m_switchingDevices.ToDictionary(x => x.Name, x => x);

            SwitchingDeviceBase switchingDevice = null;

            if (switchingDevices.TryGetValue(switchingDeviceName, out switchingDevice))
            {
                if (revertToDefaultState)
                {
                    switchingDevice.RemoveFromManualAndRevertToDefault();
                }
                else
                {
                    switchingDevice.RemoveFromManualAndPreserveStateUntilUpdated();
                }

                return true;
            }
            else
            {
                return false;
            }
        }

        #endregion

        #region [ Private Methods ]

        /// <summary>
        /// Determines the <see cref="ObservationState"/> of each of the <see cref="Node"/> objects in the <see cref="NetworkModel"/>
        /// </summary>
        /// <param name="observedBusses">The set of <see cref="ObservedBus"/> objects to check the observability of.</param>
        /// <returns>The set of <see cref="ObservedBus"/> objects passed to function updated with the present <see cref="ObservationState"/>.</returns>
        private List<ObservedBus> CheckObservability(List<ObservedBus> observedBusses)
        {
            int numberOfDirectlyObservedNodes = 0;
            int numberOfIndirectlyObservedNodes = 0;

            // Mark the nodes that are directly observed by a voltage phasor
            foreach (ObservedBus observedBus in observedBusses)
            {
                if (m_phaseSelection == PhaseSelection.PositiveSequence)
                {
                    foreach (Node node in observedBus.Nodes)
                    {
                        if (node.Voltage.IncludeInPositiveSequenceEstimator)
                        {
                            node.Observability = ObservationState.DirectlyObserved;
                            numberOfDirectlyObservedNodes++;
                        }
                        else
                        {
                            node.Observability = ObservationState.Unobserved;
                        }
                    }
                }
                else if (m_phaseSelection == PhaseSelection.ThreePhase)
                {
                    foreach (Node node in observedBus.Nodes)
                    {
                        if (node.Voltage.IncludeInEstimator)
                        {
                            node.Observability = ObservationState.DirectlyObserved;
                            numberOfDirectlyObservedNodes++;
                        }
                        else
                        {
                            node.Observability = ObservationState.Unobserved;
                        }
                    }
                }
            }

            // Mark the remaining nodes in the coherent group based on if they are observed with at least one current phasor
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.Unobserved)
                    {
                        foreach (CurrentPhasorGroup currentPhasorGroup in m_activeCurrents)
                        {
                            if (currentPhasorGroup.MeasuredToNodeID == node.InternalID)
                            {
                                node.Observability = ObservationState.IndirectlyObserved;
                                numberOfIndirectlyObservedNodes++;
                            }
                        }
                    }
                }
            }

            // Mark the remaining nodes in the coherent group based on if their is at least one voltage phasor in on the bus
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (numberOfDirectlyObservedNodes > 0 || numberOfIndirectlyObservedNodes > 0)
                    {
                        if (node.Observability != ObservationState.DirectlyObserved && node.Observability != ObservationState.IndirectlyObserved)
                        {
                            node.Observability = ObservationState.IndirectlyObserved;
                            numberOfIndirectlyObservedNodes++;
                        }
                    }
                }
            }

            if (numberOfDirectlyObservedNodes == 0 && numberOfIndirectlyObservedNodes == 0)
            {
                return null;
            }
            else
            {
                return observedBusses;
            }
        }

        /// <summary>
        /// Establishes the necessesary references for complex power calculations to occurr.
        /// </summary>
        private void InitializeComplexPowerCalculations()
        {
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Transformer transformer in substation.Transformers)
                        {
                            transformer.InitializeComplexPower();
                        }
                    }

                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        transmissionLine.InitializeComplexPower();
                    }
                }
            }
        }

        /// <summary>
        /// Links the <see cref="BreakerStatus"/> objects to their parent <see cref="CircuitBreaker"/>.
        /// </summary>
        private void LinkBreakerStatusToCircuitBreakers()
        {
            Dictionary<int, BreakerStatus> breakerStatuses = m_breakerStatuses.ToDictionary(x => x.InternalID, x => x);
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
                        {
                            BreakerStatus value = null;
                            if (circuitBreaker.StatusID != 0 && breakerStatuses.TryGetValue(circuitBreaker.StatusID, out value))
                            {
                                circuitBreaker.Status = value;
                                circuitBreaker.Status.ParentCircuitBreaker = circuitBreaker;
                                value = null;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Uses the <see cref="PhasorGroup.StatusWordID"/> and the <see cref="StatusWord.InternalID"/> to link
        /// the <see cref="StatusWord"/> measurements with their associated <see cref="PhasorGroup"/>.
        /// </summary>
        private void LinkStatusWordsToPhasorGroups()
        {
            Dictionary<int, StatusWord> statusWords = m_statusWords.ToDictionary(x => x.InternalID, x => x);

            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                StatusWord value = null;

                if (voltagePhasorGroup.StatusWordID != 0 && statusWords.TryGetValue(voltagePhasorGroup.StatusWordID, out value))
                {
                    voltagePhasorGroup.Status = value;
                    value = null;
                }
                else
                {
                    voltagePhasorGroup.Status = new StatusWord();
                }
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                StatusWord value = null;
                if (currentPhasorGroup.StatusWordID != 0 && statusWords.TryGetValue(currentPhasorGroup.StatusWordID, out value))
                {
                    currentPhasorGroup.Status = value;
                    value = null;
                }
                else
                {
                    currentPhasorGroup.Status = new StatusWord();
                }
            }
        }

        /// <summary>
        /// Takes the <see cref="Node.Voltage"/> from all of the <see cref="Node"/> objects in the model
        /// and collects them into one list to make updating values much easier.
        /// </summary>
        private void ListVoltagePhasors()
        {
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Node node in substation.Nodes)
                        {
                            m_voltages.Add(node.Voltage);
                        }
                    }
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        foreach (Node node in transmissionLine.Nodes)
                        {
                            if (!m_voltages.Contains(node.Voltage))
                            {
                                m_voltages.Add(node.Voltage);
                            }
                        }
                    }
                }
            }
        }

        private void ListSwitchingDevices()
        {
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
                        {
                            m_switchingDevices.Add(circuitBreaker);
                        }

                        foreach (Switch circuitSwitch in substation.Switches)
                        {
                            m_switchingDevices.Add(circuitSwitch);
                        }
                    }
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        foreach (Switch circuitSwitch in transmissionLine.Switches)
                        {
                            m_switchingDevices.Add(circuitSwitch);
                        }
                    }
                }
            }
        }

        public void ComputeEstimatedCurrentFlows()
        {
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Transformer transformer in substation.Transformers)
                        {
                            if (m_phaseSelection == PhaseSelection.PositiveSequence)
                            {
                                transformer.ComputePositiveSequenceEstimatedCurrentFlow();
                            }
                            else if (m_phaseSelection == PhaseSelection.ThreePhase)
                            {
                                transformer.ComputeThreePhaseEstimatedCurrentFlow();
                            }
                        }
                    }
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        if (m_phaseSelection == PhaseSelection.PositiveSequence)
                        {
                            transmissionLine.ComputePositiveSequenceEstimatedCurrentFlow();
                        }
                        else if (m_phaseSelection == PhaseSelection.ThreePhase)
                        {
                            transmissionLine.ComputeThreePhaseEstimatedCurrentFlow();
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Takes the <see cref="Transformer.FromNodeCurrent"/>, <see cref="Transformer.ToNodeCurrent"/>,
        /// <see cref="TransmissionLine.FromSubstationCurrent"/>, and <see cref="TransmissionLine.ToSubstationCurrent"/>
        /// from all of the <see cref="Transformer"/> and <see cref="TransmissionLine"/> objects in the model
        /// and collects them into one list to make updating values much easier.
        /// </summary>
        private void ListCurrentPhasors()
        {
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Transformer transformer in substation.Transformers)
                        {
                            // Add the current phasor groups to the list
                            m_currents.Add(transformer.FromNodeCurrent);
                            m_currents.Add(transformer.ToNodeCurrent);

                            // Set the measured branch of each current phasor group
                            transformer.FromNodeCurrent.MeasuredBranch = transformer;
                            transformer.ToNodeCurrent.MeasuredBranch = transformer;
                        }
                    }
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {

                        // Add the current phasor groups to the list
                        m_currents.Add(transmissionLine.FromSubstationCurrent);
                        m_currents.Add(transmissionLine.ToSubstationCurrent);

                        // Set the measured branch of each current phasor group
                        transmissionLine.FromSubstationCurrent.MeasuredBranch = transmissionLine;
                        transmissionLine.ToSubstationCurrent.MeasuredBranch = transmissionLine;
                    }
                }
            }
        }

        /// <summary>
        /// Establishes the parent/child relationship between entities in the <see cref="NetworkModel"/>.
        /// </summary>
        private void LinkHierarchicalReferences()
        {
            Dictionary<int, Node> substationNodes = new Dictionary<int, Node>();
            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    division.ParentCompany = company;
                    foreach (Substation substation in division.Substations)
                    {
                        substation.ParentDivision = division;
                        foreach (Node node in substation.Nodes)
                        {
                            node.ParentSubstation = substation;
                            node.Voltage.MeasuredNode = node;
                            substationNodes.Add(node.InternalID, node);
                        }

                        Dictionary<int, Node> localNodes = substation.Nodes.ToDictionary(x => x.InternalID, x => x);
                        foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
                        {
                            circuitBreaker.ParentSubstation = substation;

                            Node value = null;
                            if (localNodes.TryGetValue(circuitBreaker.FromNodeID, out value))
                            {
                                circuitBreaker.FromNode = value;
                                value = null;
                            }
                            if (localNodes.TryGetValue(circuitBreaker.ToNodeID, out value))
                            {
                                circuitBreaker.ToNode = value;
                                value = null;
                            }
                        }

                        foreach (Switch circuitSwitch in substation.Switches)
                        {
                            circuitSwitch.ParentSubstation = substation;

                            Node value = null;
                            if (localNodes.TryGetValue(circuitSwitch.FromNodeID, out value))
                            {
                                circuitSwitch.FromNode = value;
                                value = null;
                            }
                            if (localNodes.TryGetValue(circuitSwitch.ToNodeID, out value))
                            {
                                circuitSwitch.ToNode = value;
                                value = null;
                            }
                        }

                        foreach (Transformer transformer in substation.Transformers)
                        {
                            transformer.ParentSubstation = substation;

                            Node value = null;
                            if (localNodes.TryGetValue(transformer.FromNodeID, out value))
                            {
                                transformer.FromNode = value;
                                value = null;
                            }
                            if (localNodes.TryGetValue(transformer.ToNodeID, out value))
                            {
                                transformer.ToNode = value;
                                value = null;
                            }
                        }
                    }
                }
            }

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        transmissionLine.ParentDivision = division;
                        transmissionLine.SetImpedanceValues();
                        Dictionary<int, Node> transmissionLineNodes = transmissionLine.Nodes.ToDictionary(x => x.InternalID, x => x);
                        Node value = null;
                        if (substationNodes.TryGetValue(transmissionLine.FromNodeID, out value))
                        {
                            transmissionLine.FromNode = value;
                            transmissionLine.FromNode.ParentTransmissionLine = transmissionLine;
                            transmissionLine.FromSubstationCurrent.MeasuredFromNode = value;
                            transmissionLine.ToSubstationCurrent.MeasuredToNode = value;
                            transmissionLineNodes.Add(value.InternalID, value);
                            value = null;
                        }

                        if (substationNodes.TryGetValue(transmissionLine.ToNodeID, out value))
                        {
                            transmissionLine.ToNode = value;
                            transmissionLine.ToNode.ParentTransmissionLine = transmissionLine;
                            transmissionLine.ToSubstationCurrent.MeasuredFromNode = value;
                            transmissionLine.FromSubstationCurrent.MeasuredToNode = value;
                            transmissionLineNodes.Add(value.InternalID, value);
                            value = null;
                        }

                        foreach (Node node in transmissionLine.Nodes)
                        {
                            node.ParentTransmissionLine = transmissionLine;
                        }

                        
                        foreach (Switch lineSwitch in transmissionLine.Switches)
                        {
                            lineSwitch.ParentTransmissionLine = transmissionLine;

                            value = null;
                            if (transmissionLineNodes.TryGetValue(lineSwitch.FromNodeID, out value))
                            {
                                lineSwitch.FromNode = value;
                                value = null;
                            }
                            if (transmissionLineNodes.TryGetValue(lineSwitch.ToNodeID, out value))
                            {
                                lineSwitch.ToNode = value;
                                value = null;
                            }
                        }

                        foreach (LineSegment lineSegment in transmissionLine.LineSegments)
                        {
                            lineSegment.ParentTransmissionLine = transmissionLine;

                            value = null;
                            if (transmissionLineNodes.TryGetValue(lineSegment.FromNodeID, out value))
                            {
                                lineSegment.FromNode = value;
                                value = null;
                            }
                            if (transmissionLineNodes.TryGetValue(lineSegment.ToNodeID, out value))
                            {
                                lineSegment.ToNode = value;
                                value = null;
                            }
                        }
                    }    
                }
            }
        }

        /// <summary>
        /// Uses the <see cref="Transformer.TapConfigurationID"/> and the <see cref="TapConfiguration.InternalID"/> to link
        /// the <see cref="TapConfiguration"/> objects with their associated <see cref="Transformer.Tap"/>.
        /// </summary>
        private void LinkTapConfigurationReferences()
        {
            Dictionary<int, TapConfiguration> tapConfigurations = m_tapConfigurations.ToDictionary(x => x.InternalID, x => x);

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Transformer transformer in substation.Transformers)
                        {
                            TapConfiguration value = null;
                            if (tapConfigurations.TryGetValue(transformer.TapConfigurationID, out value))
                            {
                                transformer.Tap = value;
                                value = null;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Uses the <see cref="Node.VoltageLevelID"/> and the <see cref="VoltageLevel.InternalID"/> to link
        /// the <see cref="VoltageLevel"/> records with their associated <see cref="Node"/>.
        /// </summary>
        private void LinkVoltageLevelReferences()
        {
            Dictionary<int, VoltageLevel> voltageLevels = m_voltageLevels.ToDictionary(x => x.InternalID, x => x);

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        foreach (Node node in substation.Nodes)
                        {
                            VoltageLevel value = null;
                            if (voltageLevels.TryGetValue(node.VoltageLevelID, out value))
                            {
                                node.BaseKV = value;
                                value = null;
                            }
                        }
                    }

                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        foreach (Node node in transmissionLine.Nodes)
                        {
                            VoltageLevel value = null;
                            if (voltageLevels.TryGetValue(node.VoltageLevelID, out value))
                            {
                                node.BaseKV = value;
                                value = null;
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Uses the <see cref="PhasorBase.VoltageLevelID"/> and the <see cref="VoltageLevel.InternalID"/> to link
        /// the <see cref="VoltageLevel"/> records with their associated <see cref="PhasorBase"/>.
        /// </summary>
        private void LinkVoltageLevelsToPhasorGroups()
        {
            Dictionary<int, VoltageLevel> voltageLevels = m_voltageLevels.ToDictionary(x => x.InternalID, x => x);

            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                VoltageLevel value = null;
                // Positive Sequence
                if (voltageLevels.TryGetValue(voltagePhasorGroup.PositiveSequence.Measurement.VoltageLevelID, out value))
                {
                    voltagePhasorGroup.PositiveSequence.Measurement.BaseKV = value;
                    value = null;
                }
                if (voltageLevels.TryGetValue(voltagePhasorGroup.PositiveSequence.Estimate.VoltageLevelID, out value))
                {
                    voltagePhasorGroup.PositiveSequence.Estimate.BaseKV = value;
                    value = null;
                }

                if (m_phaseSelection == PhaseSelection.ThreePhase)
                {
                    // Phase A
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseA.Measurement.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseA.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseA.Estimate.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseA.Estimate.BaseKV = value;
                        value = null;
                    }

                    // Phase B
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseB.Measurement.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseB.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseB.Estimate.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseB.Estimate.BaseKV = value;
                        value = null;
                    }

                    // Phase C
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseC.Measurement.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseC.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(voltagePhasorGroup.PhaseC.Estimate.VoltageLevelID, out value))
                    {
                        voltagePhasorGroup.PhaseC.Estimate.BaseKV = value;
                        value = null;
                    }
                }

            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                VoltageLevel value = null;
                // Positive Sequence
                if (voltageLevels.TryGetValue(currentPhasorGroup.PositiveSequence.Measurement.VoltageLevelID, out value))
                {
                    currentPhasorGroup.PositiveSequence.Measurement.BaseKV = value;
                    value = null;
                }
                if (voltageLevels.TryGetValue(currentPhasorGroup.PositiveSequence.Estimate.VoltageLevelID, out value))
                {
                    currentPhasorGroup.PositiveSequence.Estimate.BaseKV = value;
                    value = null;
                }

                if (m_phaseSelection == PhaseSelection.ThreePhase)
                {
                    // Phase A
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseA.Measurement.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseA.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseA.Estimate.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseA.Estimate.BaseKV = value;
                        value = null;
                    }

                    // Phase B
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseB.Measurement.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseB.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseB.Estimate.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseB.Estimate.BaseKV = value;
                        value = null;
                    }

                    // Phase C
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseC.Measurement.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseC.Measurement.BaseKV = value;
                        value = null;
                    }
                    if (voltageLevels.TryGetValue(currentPhasorGroup.PhaseC.Estimate.VoltageLevelID, out value))
                    {
                        currentPhasorGroup.PhaseC.Estimate.BaseKV = value;
                        value = null;
                    }
                }
            }
        }

        /// <summary>
        /// Uses the <see cref="TransmissionLine.FromSubstationID"/> and <see cref="TransmissionLine.ToSubstationID"/>
        /// and the <see cref="Substation.InternalID"/> to link the <see cref="Substations"/> records with their associated <see cref="TransmissionLine"/>.
        /// </summary>
        private void LinkTransmissionLineReferencesToSubstations()
        {
            // Transmission Line From and To Substations
            Dictionary<int, Substation> substations = new Dictionary<int, Substation>();

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        substations.Add(substation.InternalID, substation);
                    }
                }
            }

            foreach (Company company in m_childrenCompanies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                        Substation value = null;
                        if (substations.TryGetValue(transmissionLine.FromSubstationID, out value))
                        {
                            transmissionLine.FromSubstation = value;
                            value = null;
                        }
                        if (substations.TryGetValue(transmissionLine.ToSubstationID, out value))
                        {
                            transmissionLine.ToSubstation = value;
                            value = null;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Maps the raw voltage measurement values to their virtual counterpart
        /// </summary>
        /// <param name="m_rawMeasurementKeyValuePairs">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertVoltageMeasurements()
        {
            foreach (VoltagePhasorGroup voltagePhasorGroup in m_voltages)
            {
                double value = 0;

                if (m_acceptsMeasurements)
                {
                    //                                    Positive Sequence Magnitude Measurement
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PositiveSequence.Measurement.MagnitudeKey, out value))
                    {
                        voltagePhasorGroup.PositiveSequence.Measurement.Magnitude = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    voltagePhasorGroup.PositiveSequence.Measurement.ClearValues();
                    //    value = 0;
                    //}

                    //                                    Positive Sequence Angle Measurement
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PositiveSequence.Measurement.AngleKey, out value))
                    {
                        voltagePhasorGroup.PositiveSequence.Measurement.AngleInDegrees = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    voltagePhasorGroup.PositiveSequence.Measurement.ClearValues();
                    //    value = 0;
                    //}

                    if (m_phaseSelection == PhaseSelection.ThreePhase)
                    {
                        //                                    Phase A Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseA.Measurement.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseA.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseA.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase A Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseA.Measurement.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseA.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseA.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase B Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseB.Measurement.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseB.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseB.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase B Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseB.Measurement.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseB.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseB.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase C Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseC.Measurement.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseC.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseC.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase C Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseC.Measurement.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseC.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseC.Measurement.ClearValues();
                        //    value = 0;
                        //}
                    }
                }

                if (m_acceptsEstimates)
                {
                    //                                    Positive Sequence Magnitude Estimate
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PositiveSequence.Estimate.MagnitudeKey, out value))
                    {
                        voltagePhasorGroup.PositiveSequence.Estimate.Magnitude = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    voltagePhasorGroup.PositiveSequence.Estimate.ClearValues();
                    //    value = 0;
                    //}

                    //                                    Positive Sequence Angle Estimate
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PositiveSequence.Estimate.AngleKey, out value))
                    {
                        voltagePhasorGroup.PositiveSequence.Estimate.AngleInDegrees = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    voltagePhasorGroup.PositiveSequence.Estimate.ClearValues();
                    //    value = 0;
                    //}

                    if (m_phaseSelection == PhaseSelection.ThreePhase)
                    {
                        //                                    Phase A Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseA.Estimate.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseA.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseA.Estimate.ClearValues();
                        //    value = 0;
                        //}


                        //                                    Phase A Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseA.Estimate.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseA.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseA.Estimate.ClearValues();
                        //    value = 0;
                        //}


                        //                                    Phase B Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseB.Estimate.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseB.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseB.Estimate.ClearValues();
                        //    value = 0;
                        //}



                        //                                    Phase B Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseB.Estimate.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseB.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseB.Estimate.ClearValues();
                        //    value = 0;
                        //}


                        //                                    Phase C Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseC.Estimate.MagnitudeKey, out value))
                        {
                            voltagePhasorGroup.PhaseC.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseC.Estimate.ClearValues();
                        //    value = 0;
                        //}


                        //                                    Phase C Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(voltagePhasorGroup.PhaseC.Estimate.AngleKey, out value))
                        {
                            voltagePhasorGroup.PhaseC.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    voltagePhasorGroup.PhaseC.Estimate.ClearValues();
                        //    value = 0;
                        //}
                    }
                }
            }
        }

        /// <summary>
        /// Maps the raw current measurement values to their virtual counterpart
        /// </summary>
        /// <param name="m_rawMeasurementKeyValuePairs">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertCurrentMeasurements()
        {
            foreach (CurrentPhasorGroup currentPhasorGroup in m_currents)
            {
                double value = 0;

                if (m_acceptsMeasurements)
                {
                    //                                    Positive Sequence Magnitude Measurement
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PositiveSequence.Measurement.MagnitudeKey, out value))
                    {
                        currentPhasorGroup.PositiveSequence.Measurement.Magnitude = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    currentPhasorGroup.PositiveSequence.Measurement.ClearValues();
                    //    value = 0;
                    //}

                    //                                    Positive Sequence Angle Measurement
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PositiveSequence.Measurement.AngleKey, out value))
                    {
                        currentPhasorGroup.PositiveSequence.Measurement.AngleInDegrees = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    currentPhasorGroup.PositiveSequence.Measurement.ClearValues();
                    //    value = 0;
                    //}

                    if (m_phaseSelection == PhaseSelection.ThreePhase)
                    {
                        //                                    Phase A Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseA.Measurement.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseA.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseA.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase A Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseA.Measurement.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseA.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseA.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase B Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseB.Measurement.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseB.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseB.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase B Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseB.Measurement.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseB.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseB.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase C Magnitude Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseC.Measurement.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseC.Measurement.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseC.Measurement.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase C Angle Measurement
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseC.Measurement.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseC.Measurement.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseC.Measurement.ClearValues();
                        //    value = 0;
                        //}
                    }
                }


                if (m_acceptsEstimates)
                {
                //                                    Positive Sequence Magnitude Estimate
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PositiveSequence.Estimate.MagnitudeKey, out value))
                    {
                        currentPhasorGroup.PositiveSequence.Estimate.Magnitude = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    currentPhasorGroup.PositiveSequence.Estimate.ClearValues();
                    //    value = 0;
                    //}

                    //                                    Positive Sequence Angle Estimate
                    if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PositiveSequence.Estimate.AngleKey, out value))
                    {
                        currentPhasorGroup.PositiveSequence.Estimate.AngleInDegrees = value;
                        value = 0;
                    }
                    //else
                    //{
                    //    currentPhasorGroup.PositiveSequence.Estimate.ClearValues();
                    //    value = 0;
                    //}


                    if (m_phaseSelection == PhaseSelection.ThreePhase)
                    {
                        //                                    Phase A Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseA.Estimate.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseA.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseA.Estimate.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase A Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseA.Estimate.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseA.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseA.Estimate.ClearValues();
                        //    value = 0;
                        //}

                        //                                    Phase B Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseB.Estimate.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseB.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseB.Estimate.ClearValues();
                        //    value = 0;
                        //}



                        //                                    Phase B Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseB.Estimate.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseB.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseB.Estimate.ClearValues();
                        //    value = 0;
                        //}



                        //                                    Phase C Magnitude Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseC.Estimate.MagnitudeKey, out value))
                        {
                            currentPhasorGroup.PhaseC.Estimate.Magnitude = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseC.Estimate.ClearValues();
                        //    value = 0;
                        //}


                        //                                    Phase C Angle Estimate
                        if (m_rawMeasurementKeyValuePairs.TryGetValue(currentPhasorGroup.PhaseC.Estimate.AngleKey, out value))
                        {
                            currentPhasorGroup.PhaseC.Estimate.AngleInDegrees = value;
                            value = 0;
                        }
                        //else
                        //{
                        //    currentPhasorGroup.PhaseC.Estimate.ClearValues();
                        //    value = 0;
                        //}
                    }
                }
            }
        }

        /// <summary>
        /// Maps the raw digital values to their virtual counterpart
        /// </summary>
        /// <param name="m_rawMeasurementKeyValuePairs">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertBreakerStatuses()
        {
            foreach (BreakerStatus breakerStatus in m_breakerStatuses)
            {
                double value = 0;
                // Breaker Statuses
                if (m_rawMeasurementKeyValuePairs.TryGetValue(breakerStatus.Key, out value))
                {
                    breakerStatus.BinaryValue = Convert.ToInt32(value);
                    value = 0;
                }
            }
        }

        /// <summary>
        /// Maps the raw status word values to their virtual counterpart
        /// </summary>
        /// <param name="m_rawMeasurementKeyValuePairs">A Dictionary containing the raw measurement values and keys.</param>
        private void InsertStatusWords()
        {
            foreach (StatusWord statusWord in m_statusWords)
            {
                double value = 0;
                // Status Words
                if (m_rawMeasurementKeyValuePairs.TryGetValue(statusWord.Key, out value))
                {
                    statusWord.BinaryValue = value;
                    value = 0;
                }
            }
        }

        #endregion

        #region [ Xml Serialization/Deserialization ]

        /// <summary>
        /// Creates a new <see cref="NetworkModel"/> by deserializing the configuration file from the specified location.
        /// </summary>
        /// <param name="pathName">The location of the configuration file including the file name.</param>
        /// <returns>A new <see cref="NetworkModel"/> based on the configuration file.</returns>
        public static NetworkModel DeserializeFromXml(string pathName)
        {
            try
            {
                // Create an empy NetworkModel object reference.
                NetworkModel networkModel = null;

                // Create an XmlSerializer with the type of NetworkModel.
                XmlSerializer deserializer = new XmlSerializer(typeof(NetworkModel));

                // Read the data in from the file.
                StreamReader reader = new StreamReader(pathName);

                // Cast the deserialized data as a NetworkModel object.
                networkModel = (NetworkModel)deserializer.Deserialize(reader);

                // Close the connection.
                reader.Close();

                return networkModel;
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Deserialize the NetworkModel from the Configuration File: " + exception.ToString());
            }
        }

        /// <summary>
        /// Serialized the <see cref="NetworkModel"/> to the specified file.
        /// </summary>
        /// <param name="pathName">The directory name included the file name of the desired location for Xml Serialization.</param>
        public void SerializeToXml(string pathName)
        {
            try
            {
                // Create an XmlSerializer with the type of NetworkModel
                XmlSerializer serializer = new XmlSerializer(typeof(NetworkModel));

                // Open a connection to the file and path.
                TextWriter writer = new StreamWriter(pathName);

                // Serialize this instance of NetworkModel
                serializer.Serialize(writer, this);

                // Close the connection
                writer.Close();
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Serialize the NetworkModel to the Configuration File: " + exception.ToString());
            }
        }

        #endregion
    }
}
