﻿//******************************************************************************************************
//  Node.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class Node : INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_acronym;
        private string m_name;
        private string m_description;

        private VoltageLevel m_voltageLevel;
        private int m_voltageLevelID;

        private Substation m_parentSubstation;
        private int m_parentSubstationID;

        private TransmissionLine m_parentTransmissionLine;
        private int m_parentTransmissionLineID;

        private VoltagePhasorGroup m_voltage;
        private int m_voltagePhasorGroupID;

        private ObservationState m_observationState;
        private List<Node> m_adjacencies;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The unique integer identifier for each <see cref="Node"/> in the <see cref="NetworkModel"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            { 
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="Node"/>.
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="Node"/>.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return m_acronym; 
            }
            set 
            { 
                m_acronym = value; 
            }
        }

        /// <summary>
        /// A name for the <see cref="Node"/>.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="Node"/>. 
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// Returns the type of the object.
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The <see cref="VoltageLevel"/> of the <see cref="Node"/>.
        /// </summary>
        [XmlIgnore()]
        public VoltageLevel BaseKV
        {
            get 
            { 
                return m_voltageLevel; 
            }
            set 
            { 
                m_voltageLevel = value;
                m_voltageLevelID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="VoltageLevel.InternalID"/> of the baseKV <see cref="VoltageLevel"/>
        /// </summary>
        [XmlAttribute("BaseKV")]
        public int VoltageLevelID
        {
            get 
            { 
                return m_voltageLevelID; 
            }
            set 
            { 
                m_voltageLevelID = value; 
            }
        }

        /// <summary>
        /// The parent/owner <see cref="Substation"/>
        /// </summary>
        [XmlIgnore()]
        public Substation ParentSubstation
        {
            get 
            { 
                return m_parentSubstation; 
            }
            set 
            { 
                m_parentSubstation = value;
                m_parentSubstationID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="Substation.InternalID"/> of the parent <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("ParentSubstation")]
        public int ParentSubstationID
        {
            get 
            { 
                return m_parentSubstationID; 
            }
            set 
            { 
                m_parentSubstationID = value; 
            }
        }

        /// <summary>
        /// The parent/owner <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlIgnore()]
        public TransmissionLine ParentTransmissionLine
        {
            get 
            { 
                return m_parentTransmissionLine; 
            }
            set 
            { 
                m_parentTransmissionLine = value;
                m_parentTransmissionLineID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="TransmissionLine.InternalID"/> of the parent <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("ParentTransmissionLine")]
        public int ParentTransmissionLineID
        {
            get 
            { 
                return m_parentTransmissionLineID; 
            }
            set 
            { 
                m_parentTransmissionLineID = value; 
            }
        }

        /// <summary>
        /// The <see cref="VoltagePhasorGroup"/> which measures the <see cref="Node"/>.
        /// </summary>
        [XmlElement("Voltage")]
        public VoltagePhasorGroup Voltage
        {
            get 
            { 
                return m_voltage; 
            }
            set 
            { 
                m_voltage = value;
                m_voltagePhasorGroupID = value.InternalID;
                m_voltage.MeasuredNode = this;
            }
        }

        /// <summary>
        /// The <see cref="VoltagePhasorGroup.InternalID"/> of the <see cref="Voltage"/>.
        /// </summary>
        [XmlIgnore()]
        public int VoltagePhasorGroupID
        {
            get
            {
                return m_voltagePhasorGroupID;
            }
            set
            {
                m_voltagePhasorGroupID = value;
            }
        }

        /// <summary>
        /// The present status of the observability of the <see cref="Node"/>.
        /// </summary>
        [XmlIgnore()]
        public ObservationState Observability
        {
            get
            {
                return m_observationState;
            }
            set
            {
                m_observationState = value;
            }
        }

        [XmlIgnore()]
        public List<Node> AdjacencyList
        {
            get
            {
                return m_adjacencies;
            }
            set
            {
                m_adjacencies = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Node()
            :this(0, 0, "Undefined Name", "Undefined Description", 0)
        {
        }

        /// <summary>
        /// A constructor for the <see cref="Node"/> class which requires the <see cref="VoltageLevel"/> and the required properties defined by the <see cref="INetworkDescribable"/> interface.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Node"/> in the <see cref="NetworkModel"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Node"/>.</param>
        /// <param name="name">A name for the <see cref="Node"/>.</param>
        /// <param name="description">A description of the <see cref="Node"/>.</param>
        /// <param name="baseKV">The <see cref="VoltageLevel"/> of the <see cref="Node"/>.</param>
        public Node(int internalID, int number, string name, string description, int voltageLevelID)
            :this(internalID, number, name, description, voltageLevelID, null, null)
        {
        }

        /// <summary>
        /// The designated initializer for the <see cref="Node"/> class. It requires specification of the <see cref="VoltageLevel"/> and the required properties defined by the <see cref="INetworkDescribable"/> interface as well as references to its parent <see cref="Substation"/> and/or <see cref="TransmissionLine"/>.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each <see cref="Node"/> in the <see cref="NetworkModel"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Node"/>.</param>
        /// <param name="name">A name for the <see cref="Node"/>.</param>
        /// <param name="description">A description of the <see cref="Node"/>.</param>
        /// <param name="baseKV">The <see cref="VoltageLevel"/> of the <see cref="Node"/>.</param>
        /// <param name="parentSubstation"></param>
        /// <param name="parentTransmissionLine"></param>
        public Node(int internalID, int number, string name, string description, int voltageLevelID, Substation parentSubstation, TransmissionLine parentTransmissionLine)
        {
            m_internalID = internalID;
            m_number = number;
            m_name = name;
            m_description = description;
            m_voltageLevelID = voltageLevelID;
            m_parentSubstation = parentSubstation;
            m_parentTransmissionLine = parentTransmissionLine;
            //m_voltage = new VoltagePhasorGroup();
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="Division"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="Division"/> class.</returns>
        public override string ToString()
        {
            return "-NODE- ID: " + InternalID.ToString() + " KV: " + m_voltageLevel.Name + " Number: " + Number.ToString() + " Name: " + m_name;
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="Division"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="Division"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Node ---------------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          BaseKV: " + m_voltageLevel.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("ObservationState: " + m_observationState.ToString() + "{0}", Environment.NewLine);
            if (m_parentSubstation != null)
            {
                stringBuilder.AppendFormat("ParentSubstation: " + m_parentSubstation.ToString() + "{0}", Environment.NewLine);
            }
            if (m_parentTransmissionLine != null)
            {
                stringBuilder.AppendFormat("ParntTrnsmsnLine: " + m_parentTransmissionLine.ToString() + "{0}", Environment.NewLine);
            }

            stringBuilder.AppendFormat(m_voltage.ToVerboseString());
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        #endregion
    }
}
