﻿//******************************************************************************************************
//  ObservedBus.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Modeling
{
    public class ObservedBus : INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private List<Node> m_observedNodes;

        #endregion

        #region [ Properties ]
        public int InternalID
        {
            get
            {
                return m_internalID;
            }
            set
            {
                m_internalID = value;
            }
        }

        public int Number
        {
            get
            {
                return m_internalID;
            }
            set
            {
            }
        }

        public string Acronym
        {
            get
            {
                StringBuilder stringBuilder = new StringBuilder();
                foreach (Node node in m_observedNodes)
                {
                    stringBuilder.AppendFormat(node.Acronym + "{0}", Environment.NewLine);
                }
                return stringBuilder.ToString();
            }
            set
            {
            }
        }

        public string Name
        {
            get
            {
                StringBuilder stringBuilder = new StringBuilder();
                foreach (Node node in m_observedNodes)
                {
                    stringBuilder.AppendFormat(node.Name + "{0}", Environment.NewLine);
                }
                return stringBuilder.ToString();
            }
            set
            {
            }
        }

        public string Description
        {
            get
            {
                StringBuilder stringBuilder = new StringBuilder();
                foreach (Node node in m_observedNodes)
                {
                    stringBuilder.AppendFormat(node.Description + "{0}", Environment.NewLine);
                }
                return stringBuilder.ToString();
            }
            set
            {
            }
        }

        /// <summary>
        /// Gets the type of the object.
        /// </summary>
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        public List<Node> Nodes
        {
            get
            {
                return m_observedNodes;
            }
            set
            {
                m_observedNodes = value;
            }
        }

        public VoltageLevel BaseKV
        {
            get
            {
                if (m_observedNodes.Count < 1)
                {
                    throw new Exception("Observed Bus cannot have a voltage level without having children nodes");
                }
                else
                {
                    return m_observedNodes[0].BaseKV.DeepCopy();
                }
            }
        }

        public VoltagePhasorGroup Value
        {
            set
            {
                foreach (Node node in m_observedNodes)
                {
                    node.Voltage.PositiveSequence.Estimate.ComplexPhasor = value.PositiveSequence.Estimate.ComplexPhasor;
                    node.Voltage.PhaseA.Estimate.ComplexPhasor = value.PhaseA.Estimate.ComplexPhasor;
                    node.Voltage.PhaseB.Estimate.ComplexPhasor = value.PhaseB.Estimate.ComplexPhasor;
                    node.Voltage.PhaseC.Estimate.ComplexPhasor = value.PhaseC.Estimate.ComplexPhasor;
                }
            }
        }

        #endregion

        #region [ Constructors ]

        public ObservedBus()
            :this(0, new List<Node>())
        {
        }

        public ObservedBus(int internalID, Node observedNode)
        {
            m_internalID = internalID;
            m_observedNodes = new List<Node>();
            m_observedNodes.Add(observedNode);
        }

        public ObservedBus(int internalID, List<Node> observedNodes)
        {
            m_internalID = internalID;
            m_observedNodes = observedNodes;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="ObservedBus"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="ObservedBus"/> class.</returns>
        public override string ToString()
        {
            string observedBusAsString = "";
            foreach (Node node in m_observedNodes)
            {
                observedBusAsString += node.InternalID + "|" + node.Name + ",";
            }
            observedBusAsString.Remove(observedBusAsString.Length - 1);
            return "-OBS- ID: " + m_internalID.ToString() + "  " + observedBusAsString;
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="ObservedBus"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="Observed"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Observed Bus -------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            foreach (Node node in m_observedNodes)
            {
                stringBuilder.AppendFormat(node.ToVerboseString() + "{0}", Environment.NewLine);
            }
            
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        #endregion

        
    }
}
