﻿//******************************************************************************************************
//  SeriesCompensation.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  09/17/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class SeriesCompensation : LineSegment
    {
        #region [ Private Members ]

        private double m_lowerBounds;
        private double m_upperBounds;
        private string m_measurementKey;
        private SeriesCompensationStatus m_status;
        private SeriesCompensationType m_type;

        #endregion

        #region [ Properties ]

        [XmlAttribute("LowerBounds")]
        public double LowerBounds
        {
            get
            {
                return m_lowerBounds;
            }
            set
            {
                m_lowerBounds = value;
            }
        }

        [XmlAttribute("UpperBounds")]
        public double UpperBounds
        {
            get
            {
                return m_upperBounds;
            }
            set
            {
                m_upperBounds = value;
            }
        }

        [XmlAttribute("Key")]
        public string MeasurementKey
        {
            get
            {
                return m_measurementKey;
            }
            set
            {
                m_measurementKey = value;
            }
        }

        [XmlAttribute("Type")]
        public SeriesCompensationType Type
        {
            get
            {
                return m_type;
            }
            set
            {
                m_type = value;
            }
        }

        [XmlAttribute("Status")]
        public SeriesCompensationStatus Status
        {
            get
            {
                return m_status;
            }
            set
            {
                m_status = value;
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public SeriesCompensation()
            : this(0, 0, "Undefined Name", "Undefined Description", new Impedance(), "Undefined", 0.1, 0.1)
        {
        }

        public SeriesCompensation(int internalID, int number, string name, string description, Impedance impedance, string measurementKey, double lowerBounds, double upperBounds)
            : this(internalID, number, name, description, impedance, 0, measurementKey, lowerBounds, upperBounds)
        {
            m_lowerBounds = lowerBounds;
            m_upperBounds = upperBounds;
            m_status = SeriesCompensationStatus.Bypassed;
        }

        public SeriesCompensation(int internalID, int number, string name, string description, Impedance impedance, int parentTransmissionLineID, string measurementKey, double lowerBounds, double upperBounds)
            : base(internalID, number, name, description, impedance)
        {
            m_measurementKey = measurementKey;
            m_lowerBounds = lowerBounds;
            m_upperBounds = upperBounds;
            m_status = SeriesCompensationStatus.Bypassed;
            m_type = SeriesCompensationType.Capacitor;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="SeriesCompensation"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="SeriesCompensation"/> class.</returns>
        public override string ToString()
        {
            return "-SRSCMP- ID: " + InternalID.ToString() + " Number: " + Number.ToString() + " Line: " + ParentTransmissionLine.Number.ToString();
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="SeriesCompensation"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="SeriesCompensation"/> class.</returns>
        public new string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Series Compensation ------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("ParntTrnsmsnLine: " + ParentTransmissionLine.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     LowerBounds: " + LowerBounds.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     UpperBounds: " + UpperBounds.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Type: " + Type.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Status: " + Status.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat(RawImpedanceParameters.ToVerboseString());
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        #endregion
    }
}
