﻿//******************************************************************************************************
//  ShuntBranchBase.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;

namespace SynchrophasorAnalytics.Modeling
{
    public class ShuntBranchBase : ISingleTerminal, IImpedance, INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_name;
        private string m_description;
        private Impedance m_impedance;

        private Node m_connectedNode;
        private int m_connectedNodeID;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The unique integer identifier for each instance of a <see cref="ShuntBranchBase"/> object.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            { 
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="ShuntBranchBase"/> object.
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="ShuntBranchBase"/> object. Will always return 'G + jB'.
        /// </summary>
        [XmlIgnore()]
        public string Acronym
        {
            get 
            { 
                return "G + jB"; 
            }
            set 
            { 
            }
        }

        /// <summary>
        /// A name for the <see cref="ShuntBranchBase"/> object.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="ShuntBranchBase"/> object. 
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// Returns the type of the object.
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The <see cref="Node"/> which the <see cref="ShuntBranchBase"/> connects to.
        /// </summary>
        [XmlIgnore()]
        public Node ConnectedNode
        {
            get
            {
                return m_connectedNode;
            }
            set
            {
                m_connectedNode = value;
                m_connectedNodeID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="Node.InternalID"/> of the <see cref="ConnectedNode"/>.
        /// </summary>
        [XmlAttribute()]
        public int ConnectedNodeID
        {
            get
            {
                return m_connectedNodeID;
            }
            set
            {
                m_connectedNodeID = value;
            }
        }

        /// <summary>
        /// The complex (R + jX) shunt impedance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSeriesImpedance
        {
            get 
            { 
                return m_impedance.PositiveSequenceSeriesImpedance; 
            }
        }

        /// <summary>
        /// The complex 1/(R + jX) shunt admittance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSeriesAdmittance
        {
            get 
            { 
                return m_impedance.PositiveSequenceSeriesAdmittance; 
            }
        }

        /// <summary>
        /// The complex (G + jB) shunt susceptance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceShuntSusceptance
        {
            get 
            { 
                return m_impedance.PositiveSequenceShuntSusceptance; 
            }
        }

        /// <summary>
        /// The complex (G + jB) shunt susceptance for a single side of the shunt impedance.
        /// </summary>
        [XmlIgnore()]
        public DoubleComplex PositiveSequenceSingleSidedShuntSusceptance
        {
            get
            {
                return m_impedance.PositiveSequenceSingleSidedShuntSusceptance;
            }
        }

        /// <summary>
        /// The three phase complex (R + jX) shunt impedance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSeriesImpedance
        {
            get 
            { 
                return m_impedance.ThreePhaseSeriesImpedance; 
            }
        }

        /// <summary>
        /// The three phase complex 1/(R +jX) shunt admittance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSeriesAdmittance
        {
            get 
            { 
                return m_impedance.ThreePhaseSeriesAdmittance; 
            }
        }

        /// <summary>
        /// The three phase complex (G + JB) shunt susceptance as a 3 x 3 complex matrix.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseShuntSusceptance
        {
            get 
            { 
                return m_impedance.ThreePhaseShuntSusceptance; 
            }
        }


        /// <summary>
        /// The three phase complex (G + JB) shunt susceptance as a 3 x 3 complex matrix for a single side of the shunt branch.
        /// </summary>
        [XmlIgnore()]
        public ComplexMatrix ThreePhaseSingleSidedShuntSusceptance
        {
            get
            {
                return m_impedance.ThreePhaseSingleSidedShuntSusceptance;
            }
        }

        [XmlElement("Impedance")]
        public Impedance RawImpedanceParamters
        {
            get 
            { 
                return m_impedance; 
            }
            set 
            { 
                m_impedance = value; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public ShuntBranchBase()
            :this(0, 0, "Undefined Name","Undefined Description", new Impedance(), 0)
        {
        }

        /// <summary>
        /// The designated constructor for the <see cref="ShuntBranchBase"/> class.
        /// </summary>
        /// <param name="internalID">The unique integer identifier for each instance of a <see cref="ShuntBranchBase"/> object.</param>
        /// <param name="number">A descriptive number for the <see cref="ShuntBranchBase"/> object.</param>
        /// <param name="name">A name for the <see cref="ShuntBranchBase"/> object.</param>
        /// <param name="description">A description of the <see cref="ShuntBranchBase"/> object. </param>
        /// <param name="impedance">The <see cref="Impedance"/> of the <see cref="ShuntBranchBase"/>.</param>
        /// <param name="connectedNodeID">The <see cref="Node.InternalID"/> of the <see cref="ConnectedNode"/>.</param>
        public ShuntBranchBase(int internalID, int number, string name, string description, Impedance impedance, int connectedNodeID)
        {
            m_internalID = internalID;
            m_number = number;
            m_name = name;
            m_description = description;
            m_impedance = impedance;
            m_connectedNodeID = connectedNodeID;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of the instance of the <see cref="ShuntBranchBase"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="ShuntBranchBase"/> class.</returns>
        public override string ToString()
        {
            return "-SHUNT- ID: " + InternalID.ToString() + " Number: " + Number.ToString();
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="ShuntBranchBase"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="ShuntBranchBase"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Shunt Branch Base --------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + m_internalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + m_number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + m_name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + m_description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("   ConnectedNode: " + m_connectedNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat(m_impedance.ToVerboseString());
            stringBuilder.AppendLine();
            return stringBuilder.ToString();

        }

        #endregion

    }
}
