﻿//******************************************************************************************************
//  Substation.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Graph;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class Substation : INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_acronym;
        private string m_name;
        private string m_description;
        private Division m_parentDivision;
        private List<Node> m_childrenNodes;
        private List<Transformer> m_childrenTransformers;
        private List<CircuitBreaker> m_childrenCircuitBreakers;
        private List<Switch> m_childrenSwitches;
        private SubstationGraph m_graph;

        #endregion

        #region [ Properties ] 

        /// <summary>
        /// A unique integer identifier for the <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            { 
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return m_acronym; 
            }
            set 
            { 
                m_acronym = value; 
            }
        }

        /// <summary>
        /// A descriptive name for the <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="Substation"/>.
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// Gets the type of the object
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The parent <see cref="Division"/> of the <see cref="Substation"/>.
        /// </summary>
        [XmlIgnore()]
        public Division ParentDivision
        {
            get 
            { 
                return m_parentDivision; 
            }
            set 
            {
                m_parentDivision = value; 
            }
        }

        /// <summary>
        /// All of the children <see cref="Node"/> objects of the <see cref="Substation"/>.
        /// </summary>
        [XmlArray("Nodes")]
        public List<Node> Nodes
        {
            get 
            { 
                return m_childrenNodes; 
            }
            set 
            { 
                m_childrenNodes = value;
                foreach (Node node in m_childrenNodes)
                {
                    node.ParentSubstation = this;
                }
            }
        }

        /// <summary>
        /// All of the children <see cref="Transformer"/> objects of the <see cref="Substation"/>.
        /// </summary>
        [XmlArray("Transformers")]
        public List<Transformer> Transformers
        {
            get 
            { 
                return m_childrenTransformers; 
            }
            set 
            { 
                m_childrenTransformers = value;
                foreach (Transformer transformer in m_childrenTransformers)
                {
                    transformer.ParentSubstation = this;
                }
            }
        }

        /// <summary>
        /// All of the children <see cref="CircuitBreakers"/> objects of the <see cref="Substation"/>.
        /// </summary>
        [XmlArray("CircuitBreakers")]
        public List<CircuitBreaker> CircuitBreakers
        {
            get 
            { 
                return m_childrenCircuitBreakers; 
            }
            set 
            { 
                m_childrenCircuitBreakers = value;
                foreach (CircuitBreaker circuitBreaker in m_childrenCircuitBreakers)
                {
                    circuitBreaker.ParentSubstation = this;
                }
            }
        }

        /// <summary>
        /// All of the children <see cref="Switch"/> objects of the <see cref="Substation"/>.
        /// </summary>
        [XmlArray("Switches")]
        public List<Switch> Switches
        {
            get 
            { 
                return m_childrenSwitches; 
            }
            set 
            { 
                m_childrenSwitches = value;
                foreach (Switch switchingDevice in m_childrenSwitches)
                {
                    switchingDevice.ParentSubstation = this;
                }
            }
        }

        [XmlIgnore()]
        public SubstationGraph Graph
        {
            get 
            { 
                return m_graph; 
            }
            set 
            { 
                m_graph = value; 
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Substation()
            :this (0, 0, "IDK", "Undefined Name", "Undefined Description")
        {
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="internalID">A unique integer identifier for the <see cref="Substation"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Substation"/>.</param>
        /// <param name="acronym">A descriptive acronym for the <see cref="Substation"/>.</param>
        /// <param name="name">A descriptive name for the <see cref="Substation"/>.</param>
        /// <param name="description">A description of the <see cref="Substation"/>.</param>
        public Substation(int internalID, int number, string acronym, string name, string description)
            :this(internalID, number, acronym, name, description, new List<Node>(), new List<Transformer>(), new List<CircuitBreaker>(), new List<Switch>())
        {
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="internalID">A unique integer identifier for the <see cref="Substation"/>.</param>
        /// <param name="number">A descriptive number for the <see cref="Substation"/>.</param>
        /// <param name="acronym">A descriptive acronym for the <see cref="Substation"/>.</param>
        /// <param name="name">A descriptive name for the <see cref="Substation"/>.</param>
        /// <param name="description">A description of the <see cref="Substation"/>.</param>
        /// <param name="nodes"></param>
        /// <param name="transformers"></param>
        /// <param name="circuitBreakers"></param>
        /// <param name="switches"></param>
        public Substation(int internalID, int number, string acronym, string name, string description, List<Node> nodes, List<Transformer> transformers, List<CircuitBreaker> circuitBreakers, List<Switch> switches)
        {
            m_internalID = internalID;
            m_number = number;
            m_acronym = acronym;
            m_name = name;
            m_description = description;
            m_childrenNodes = nodes;
            m_childrenTransformers = transformers;
            m_childrenCircuitBreakers = circuitBreakers;
            m_childrenSwitches = switches;
        }

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// A string representation of an instance of the <see cref="Substation"/> class.
        /// </summary>
        /// <returns>A string representation of an instance of the <see cref="Substation"/> class.</returns>
        public override string ToString()
        {
            return "-SUB- ID: " + m_internalID.ToString() + " Name: " + m_name;
        }

        /// <summary>
        /// A verbose string representation of an instance of the <see cref="Substation"/> class.
        /// </summary>
        /// <returns>A verbose string representation of an instance of the <see cref="Substation"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Substation ---------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + m_internalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + m_number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("         Acronym: " + m_acronym + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + m_name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + m_description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("  ParentDivision: " + m_parentDivision.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("           Nodes:  {0}", Environment.NewLine);
            foreach (Node node in m_childrenNodes)
            {
                stringBuilder.AppendFormat(node.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("    Transformers:  {0}", Environment.NewLine);
            foreach (Transformer transformer in m_childrenTransformers)
            {
                stringBuilder.AppendFormat(transformer.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat(" CircuitBreakers:  {0}", Environment.NewLine);
            foreach (CircuitBreaker circuitBreaker in m_childrenCircuitBreakers)
            {
                stringBuilder.AppendFormat(circuitBreaker.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("        Switches:  {0}", Environment.NewLine);
            foreach (Switch circuitSwitch in m_childrenSwitches)
            {
                stringBuilder.AppendFormat(circuitSwitch.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Determines equality between to <see cref="Substation"/> objects. Children objects are not included in the equality check.
        /// </summary>
        /// <param name="target">The target object for equality testing.</param>
        /// <returns>A bool representing the result of the equality check.</returns>
        public override bool Equals(object target)
        {
            // If parameter is null return false.
            if (target == null)
            {
                return false;
            }

            // If parameter cannot be cast to PhasorBase return false.
            Substation substation = target as Substation;

            if ((object)substation == null)
            {
                return false;
            }

            // Return true if the fields match:
            if (this.m_internalID != substation.InternalID)
            {
                return false;
            }
            else if (this.m_number != substation.Number)
            {
                return false;
            }
            else if (!this.m_acronym.Equals(substation.Acronym))
            {
                return false;
            }
            else if (this.m_name != substation.Name)
            {
                return false;
            }
            else if (this.m_description != substation.Description)
            {
                return false;
            }
            else
            {
                return true;
            }
        }

        /// <summary>
        /// Overridden to prevent compilation warnings.
        /// </summary>
        /// <returns>An integer hash code determined by the base class.</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Performs a deep copy of the target <see cref="Substation"/> object.
        /// </summary>
        /// <returns>A deep copy of the target <see cref="Substation"/> object.</returns>
        public Substation DeepCopy()
        {
            // Incomplete
            Substation copy = (Substation)this.MemberwiseClone();

            List<Node> copyNodes = new List<Node>();
            List<Transformer> copyTransformers = new List<Transformer>();
            List<CircuitBreaker> copyCircuitBreakers = new List<CircuitBreaker>();

            foreach (Node node in m_childrenNodes)
            {
                //copyNodes.Add(node.Copy());
            }

            foreach (Transformer transformer in m_childrenTransformers)
            {
                //copyTransformers.Add(transformer.Copy());
            }

            foreach (CircuitBreaker circuitBreaker in m_childrenCircuitBreakers)
            {
                //copyCircuitBreakers.Add(circuitBreaker.Copy());
            }

            copy.Nodes = copyNodes;
            copy.Transformers = copyTransformers;
            copy.CircuitBreakers = copyCircuitBreakers;

            return copy; // Incomplete
        }

        #endregion

        #region [ Private Methods ]

        public void InitializeSubstationGraph()
        {
            m_graph = new SubstationGraph(this);
        }

        #endregion
    }
}
