﻿//******************************************************************************************************
//  SwitchingDeviceBase.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;

namespace SynchrophasorAnalytics.Modeling
{
    public class SwitchingDeviceBase : ITwoTerminal, INetworkDescribable
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_name;
        private string m_description;
        private string m_measurementKey;
        private SwitchingDeviceNormalState m_normalState;
        private SwitchingDeviceActualState m_actualState;

        private bool m_inManualOverrideMode;

        private Node m_fromNode;
        private int m_fromNodeID;

        private Node m_toNode;
        private int m_toNodeID;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// A unique integer identifier for the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set
            {
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get 
            { 
                return m_number; 
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="SwitchingDeviceBase"/>. Will always return 'SWDV'.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return "SWDV"; 
            }
            set 
            { 
            }
        }

        /// <summary>
        /// A descriptive name for the <see cref="SwitchingDeviceBase"/>
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="SwitchingDeviceBase"/>
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            { 
                m_description = value; 
            }
        }

        /// <summary>
        /// The measurement key for the status of the switchable device.
        /// </summary>
        [XmlAttribute("Key")]
        public string MeasurementKey
        {
            get
            {
                return m_measurementKey;
            }
            set
            {
                m_measurementKey = value;
            }
        }

        /// <summary>
        /// Gets the type of the object.
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The <see cref="Node"/> at the originating end of the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        [XmlIgnore()]
        public Node FromNode
        {
            get 
            { 
                return m_fromNode; 
            }
            set 
            { 
                m_fromNode = value;
                m_fromNodeID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="Node.InternalID"/> of the <see cref="FromNode"/>.
        /// </summary>
        [XmlAttribute("FromNode")]
        public int FromNodeID
        {
            get
            {
                return m_fromNodeID;
            }
            set
            {
                m_fromNodeID = value;
            }
        }

        /// <summary>
        /// The <see cref="Node"/> at the terminating end of the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        [XmlIgnore()]
        public Node ToNode
        {
            get 
            {
                return m_toNode; 
            }
            set 
            { 
                m_toNode = value;
                m_toNodeID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="Node.InternalID"/> of the <see cref="ToNode"/>.
        /// </summary>
        [XmlAttribute("ToNode")]
        public int ToNodeID
        {
            get
            {
                return m_toNodeID;
            }
            set
            {
                m_toNodeID = value;
            }
        }

        /// <summary>
        /// The normal or default state of the switch. Either <see cref="SwitchingDeviceNormalState.Open"/> or <see cref="SwitchingDeviceNormalState.Closed"/>
        /// </summary>
        [XmlAttribute("Normally")]
        public SwitchingDeviceNormalState NormalState
        {
            get 
            { 
                return m_normalState; 
            }
            set 
            { 
                m_normalState = value; 
            }
        }

        /// <summary>
        /// The actual current state of the switch. Either <see cref="SwitchingDeviceActualState.Open"/> or <see cref="SwitchingDeviceActualState.Closed"/>
        /// </summary>
        [XmlAttribute("Actually")]
        public SwitchingDeviceActualState ActualState
        {
            get
            {
                return m_actualState;
            }
            set
            {
                m_actualState = value;
            }
        }

        [XmlAttribute("ManualOverride")]
        public bool InManualOverrideMode
        {
            get
            {
                return m_inManualOverrideMode;
            }
        }

        #endregion

        #region [ Constructors ]

        public SwitchingDeviceBase()
            :this(0, 0, "Undefined", "Undefined", SwitchingDeviceNormalState.Open)
        {
        }

        public SwitchingDeviceBase(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState)
            :this(internalID, number, name, description,  normalState, 0, 0)
        {
        }

        public SwitchingDeviceBase(int internalID, int number, string name, string description, SwitchingDeviceNormalState normalState, int fromNodeID, int toNodeID)
        {
            m_internalID = internalID;
            m_number = number;
            m_name = name;
            m_description = description;
            m_normalState = normalState;
            m_fromNodeID = fromNodeID;
            m_toNodeID = toNodeID;

            RemoveFromManualAndRevertToDefault();
        }

        #endregion

        #region [ Public Methods ]

        public void ManuallySwitchTo(SwitchingDeviceActualState switchingDeviceActualState)
        {
            m_inManualOverrideMode = true;
            m_actualState = switchingDeviceActualState;
        }

        public void RemoveFromManualAndPreserveStateUntilUpdated()
        {
            m_inManualOverrideMode = false;
        }

        public void RemoveFromManualAndRevertToDefault()
        {
            m_inManualOverrideMode = false;

            if (m_normalState == SwitchingDeviceNormalState.Closed)
            {
                m_actualState = SwitchingDeviceActualState.Closed;
            }
            else if (m_normalState == SwitchingDeviceNormalState.Open)
            {
                m_actualState = SwitchingDeviceActualState.Open;
            }
        }

        /// <summary>
        /// A string representation of the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        /// <returns>A string representation of the <see cref="SwitchingDeviceBase"/>.</returns>
        public override string ToString()
        {
            return "-SWDV- ID: " + InternalID.ToString() + " Name: " + Name;
        }

        /// <summary>
        /// A verbose string representation of the <see cref="SwitchingDeviceBase"/>.
        /// </summary>
        /// <returns>A verbose string representation of the <see cref="SwitchingDeviceBase"/>.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Switching Device Base ----------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        Normally: " + NormalState.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        Actually: " + ActualState.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Manual: " + InManualOverrideMode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        #endregion
    }
}
