﻿//******************************************************************************************************
//  Transformer.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.Runtime.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Measurements;
using SynchrophasorAnalytics.GIC;

namespace SynchrophasorAnalytics.Modeling
{
    [Serializable()]
    public class Transformer : SeriesBranchBase, IGic
    {
        #region [ Private Members ]

        private CurrentPhasorGroup m_fromNodeCurrent;
        private int m_fromNodeCurrentPhasorGroupID;

        private CurrentPhasorGroup m_toNodeCurrent;
        private int m_toNodeCurrentPhasorGroupID;

        private Substation m_parentSubstation;
        private int m_parentSubstationID;

        private Impedance m_presentImpedance;

        private TapConfiguration m_tapConfiguration;
        private int m_tapConfigurationID;

        private TransformerType m_type;

        private bool m_complexPowerHasBeenInitialized;
        private ComplexPowerGroup m_fromNodeComplexPower;
        private ComplexPowerGroup m_toNodeComplexPower;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// The <see cref="TransformerType"/> of the <see cref="Transformer"/>.
        /// </summary>
        public TransformerType Type
        {
            get
            {
                return m_type;
            }
            set
            {
                m_type = value;
            }
        }

        /// <summary>
        /// The <see cref="CurrentPhasorGroup"/> that originates at the <see cref="FromNode"/>
        /// </summary>
        [XmlElement("FromNodeCurrent")]
        public CurrentPhasorGroup FromNodeCurrent
        {
            get 
            { 
                return m_fromNodeCurrent; 
            }
            set 
            { 
                m_fromNodeCurrent = value;
                m_fromNodeCurrentPhasorGroupID = value.InternalID;
                m_fromNodeCurrent.MeasuredFromNode = this.FromNode;
                m_fromNodeCurrent.MeasuredToNode = this.ToNode;
            }
        }

        /// <summary>
        /// The <see cref="CurrentPhasorGroup.InternalID"/> of the <see cref="FromNodeCurrent"/>.
        /// </summary>
        [XmlAttribute("FromNodeCurrentPhasorGroup")]
        public int FromNodeCurrentPhasorGroupID
        {
            get 
            { 
                return m_fromNodeCurrentPhasorGroupID; 
            }
            set 
            {
                m_fromNodeCurrentPhasorGroupID = value; 
            }
        }

        /// <summary>
        /// The <see cref="CurrentPhasorGroup"/> that originates at the <see cref="ToNode"/> of the <see cref="Transformer"/>
        /// </summary>
        [XmlElement("ToNodeCurrent")]
        public CurrentPhasorGroup ToNodeCurrent
        {
            get 
            { 
                return m_toNodeCurrent; 
            }
            set 
            { 
                m_toNodeCurrent = value;
                m_toNodeCurrentPhasorGroupID = value.InternalID;
                m_toNodeCurrent.MeasuredFromNode = this.ToNode;
                m_toNodeCurrent.MeasuredToNode = this.FromNode;
            }
        }

        /// <summary>
        /// The <see cref="CurrentPhasorGroup.InternalID"/> of the <see cref="ToNodeCurrent"/>.
        /// </summary>
        [XmlAttribute("ToNodeCurrentPhasorGroup")]
        public int ToNodeCurrentPhasorGroupID
        {
            get 
            { 
                return m_toNodeCurrentPhasorGroupID; 
            }
            set 
            { 
                m_toNodeCurrentPhasorGroupID = value; 
            }
        }

        /// <summary>
        /// The <see cref="Substation"/> that contains this <see cref="Transformer"/>.
        /// </summary>
        [XmlIgnore()]
        public Substation ParentSubstation
        {
            get
            {
                return m_parentSubstation;
            }
            set
            {
                m_parentSubstation = value;
                m_parentSubstationID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="Substation.InternalID"/> of the <see cref="ParentSubstation"/>.
        /// </summary>
        [XmlAttribute("ParentSubstation")]
        public int ParentSubstationID
        {
            get
            {
                return m_parentSubstationID;
            }
            set
            {
                m_parentSubstationID = value;
            }
        }

        /// <summary>
        /// The <see cref="Impedance"/> determined by the <see cref="Type"/>, <see cref="Tap"/>, and <see cref="RawImpedanceParamters"/>
        /// </summary>
        public Impedance PresentImpedance
        {
            get
            {
                return m_presentImpedance;
            }
            set
            {
                m_presentImpedance = value; 
            }
        }

        /// <summary>
        /// The <see cref="TapConfiguration"/> of the <see cref="Transformer"/>.
        /// </summary>
        [XmlIgnore()]
        public TapConfiguration Tap
        {
            get
            {
                return m_tapConfiguration;
            }
            set
            {
                m_tapConfiguration = value;
                m_tapConfigurationID = value.InternalID;
            }
        }

        /// <summary>
        /// The <see cref="TapConfiguration.InternalID"/> of the <see cref="Tap"/>.
        /// </summary>
        [XmlAttribute("TapConfiguration")]
        public int TapConfigurationID
        {
            get
            {
                return m_tapConfigurationID;
            }
            set
            {
                m_tapConfigurationID = value;
            }
        }

        [XmlIgnore()]
        public ComplexPowerGroup FromNodeComplexPower
        {
            get
            {
                if (m_complexPowerHasBeenInitialized)
                {
                    return m_fromNodeComplexPower;
                }
                else
                {
                    return null;
                }
            }
            set
            {
                if (m_complexPowerHasBeenInitialized)
                {
                    m_fromNodeComplexPower = value;
                }
            }
        }

        [XmlIgnore()]
        public ComplexPowerGroup ToNodeComplexPower
        {
            get
            {
                return m_toNodeComplexPower;
            }
            set
            {
                m_toNodeComplexPower = value;
            }
        }

        [XmlIgnore()]
        public double PositiveSequenceGicQFlowByFromSide
        {
            get 
            {
                // Use only measured values for GIC calculation as impedance skews results
                m_fromNodeComplexPower.UseEstimatedValues = false;
                m_fromNodeComplexPower.Compute();

                m_toNodeComplexPower.UseEstimatedValues = false;
                m_toNodeComplexPower.Compute();

                double measuredQLoss = Math.Abs(m_fromNodeComplexPower.PositiveSequenceReactivePower - m_toNodeComplexPower.PositiveSequenceReactivePower);
                double qLossByImpedanceFromSide = 100 * m_fromNodeCurrent.PositiveSequence.Measurement.PerUnitMagnitude * m_fromNodeCurrent.PositiveSequence.Measurement.PerUnitMagnitude * PositiveSequenceSeriesImpedance.Im;

                m_fromNodeComplexPower.UseEstimatedValues = true;
                m_toNodeComplexPower.UseEstimatedValues = true;

                return Math.Abs(measuredQLoss - qLossByImpedanceFromSide);
            }
        }

        [XmlIgnore()]
        public double PositiveSequenceGicFlowByToSide
        {
            get
            {
                // Use only measured values for GIC calculation as impedance skews results
                m_fromNodeComplexPower.UseEstimatedValues = false;
                m_fromNodeComplexPower.Compute();

                m_toNodeComplexPower.UseEstimatedValues = false;
                m_toNodeComplexPower.Compute();

                double measuredQLoss = Math.Abs(m_fromNodeComplexPower.PositiveSequenceReactivePower - m_toNodeComplexPower.PositiveSequenceReactivePower);
                double qLossByImpedanceToSide = 100 * m_toNodeCurrent.PositiveSequence.Measurement.PerUnitMagnitude * m_toNodeCurrent.PositiveSequence.Measurement.PerUnitMagnitude * PositiveSequenceSeriesImpedance.Im;

                m_fromNodeComplexPower.UseEstimatedValues = true;
                m_toNodeComplexPower.UseEstimatedValues = true;

                return Math.Abs(measuredQLoss - qLossByImpedanceToSide);
            }
        }

        [XmlIgnore()]
        public double ThreePhaseGicQFlowByFromSide
        {
            get
            {
                // Use only measured values for GIC calculation as impedance skews results
                m_fromNodeComplexPower.UseEstimatedValues = false;
                m_fromNodeComplexPower.Compute();

                m_toNodeComplexPower.UseEstimatedValues = false;
                m_toNodeComplexPower.Compute();

                double measuredQLoss = Math.Abs(m_fromNodeComplexPower.ThreePhaseReactivePower - m_toNodeComplexPower.ThreePhaseReactivePower);

                ComplexMatrix fromSideCurrentFlows = ComplexMatrix.Create(3, 1, 0);
                fromSideCurrentFlows[0, 0] = m_fromNodeCurrent.PhaseA.Measurement.PerUnitComplexPhasor;
                fromSideCurrentFlows[1, 0] = m_fromNodeCurrent.PhaseB.Measurement.PerUnitComplexPhasor;
                fromSideCurrentFlows[2, 0] = m_fromNodeCurrent.PhaseC.Measurement.PerUnitComplexPhasor;

                ComplexMatrix perUnitZLossByImpedanceFromSide = fromSideCurrentFlows.Transpose() * ThreePhaseSeriesImpedance * fromSideCurrentFlows;

                m_fromNodeComplexPower.UseEstimatedValues = true;
                m_toNodeComplexPower.UseEstimatedValues = true;

                return Math.Abs(measuredQLoss - (100 * perUnitZLossByImpedanceFromSide[0, 0].Im));
            }
        }

        [XmlIgnore()]
        public double ThreePhaseGiCFlowByToSide
        {
            get
            {
                // Use only measured values for GIC calculation as impedance skews results
                m_fromNodeComplexPower.UseEstimatedValues = false;
                m_fromNodeComplexPower.Compute();

                m_toNodeComplexPower.UseEstimatedValues = false;
                m_toNodeComplexPower.Compute();

                double measuredQLoss = Math.Abs(m_fromNodeComplexPower.ThreePhaseReactivePower - m_toNodeComplexPower.ThreePhaseReactivePower);

                ComplexMatrix toSideCurrentFlows = ComplexMatrix.Create(3, 1, 0);
                toSideCurrentFlows[0, 0] = m_toNodeCurrent.PhaseA.Measurement.PerUnitComplexPhasor;
                toSideCurrentFlows[1, 0] = m_toNodeCurrent.PhaseB.Measurement.PerUnitComplexPhasor;
                toSideCurrentFlows[2, 0] = m_toNodeCurrent.PhaseC.Measurement.PerUnitComplexPhasor;

                ComplexMatrix perUnitZLossByImpedanceToSide = toSideCurrentFlows.Transpose() * ThreePhaseSeriesImpedance * toSideCurrentFlows;

                m_fromNodeComplexPower.UseEstimatedValues = true;
                m_toNodeComplexPower.UseEstimatedValues = true;

                return Math.Abs(measuredQLoss - (100 * perUnitZLossByImpedanceToSide[0, 0].Im));
            }
        }
        
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Transformer()
            :this(0, 0, "Undefined Name", "Undefined Description", new Impedance())
        {
        }

        public Transformer(int internalID, int number, string name, string description, Impedance impedance)
            :this(internalID, number, name, description, impedance, 0, 0)
        {
        }

        public Transformer(int internalID, int number, string name, string description, Impedance impedance, int fromNodeCurrentPhasorGroupID, int toNodeCurrentPhasorGroupID)
            :this(internalID, number, name, description, impedance, fromNodeCurrentPhasorGroupID, toNodeCurrentPhasorGroupID, 0)
        {
        }

        public Transformer(int internalID, int number, string name, string description, Impedance impedance, int fromNodeCurrentPhasorGroupID, int toNodeCurrentPhasorGroupID, int tapConfigurationID)
            :base(internalID, number, name, description, impedance)
        {
            m_fromNodeCurrentPhasorGroupID = fromNodeCurrentPhasorGroupID;
            m_toNodeCurrentPhasorGroupID = toNodeCurrentPhasorGroupID;
            m_tapConfigurationID = tapConfigurationID;
            m_complexPowerHasBeenInitialized = false;
        }

        #endregion

        #region [ Public Methods ]

        public void InitializeComplexPower()
        {
            m_fromNodeComplexPower = new ComplexPowerGroup(FromNode.Voltage, FromNodeCurrent);
            m_fromNodeComplexPower.UseEstimatedValues = true;

            m_toNodeComplexPower = new ComplexPowerGroup(ToNode.Voltage, ToNodeCurrent);
            m_toNodeComplexPower.UseEstimatedValues = true;

            m_complexPowerHasBeenInitialized = true;
        }

        public void ComputePositiveSequenceEstimatedCurrentFlow()
        {
            DoubleComplex Vi = FromNode.Voltage.PositiveSequence.Estimate.PerUnitComplexPhasor;
            DoubleComplex Vj = ToNode.Voltage.PositiveSequence.Estimate.PerUnitComplexPhasor;

            DoubleComplex Iij = (PositiveSequenceSeriesAdmittance + PositiveSequenceSingleSidedShuntSusceptance) * Vi - PositiveSequenceSeriesAdmittance * Vj;
            DoubleComplex Iji = (PositiveSequenceSeriesAdmittance + PositiveSequenceSingleSidedShuntSusceptance) * Vi - PositiveSequenceSeriesAdmittance * Vj;

            FromNodeCurrent.PositiveSequence.Estimate.PerUnitComplexPhasor = Iij;
            ToNodeCurrent.PositiveSequence.Estimate.PerUnitComplexPhasor = Iji;
        }

        public void ComputeThreePhaseEstimatedCurrentFlow()
        {
            ComplexMatrix Vi = ComplexMatrix.Create(3, 1, 0);
            ComplexMatrix Vj = ComplexMatrix.Create(3, 1, 0);

            Vi[0, 0] = FromNode.Voltage.PhaseA.Estimate.PerUnitComplexPhasor;
            Vi[1, 0] = FromNode.Voltage.PhaseB.Estimate.PerUnitComplexPhasor;
            Vi[2, 0] = FromNode.Voltage.PhaseC.Estimate.PerUnitComplexPhasor;

            Vj[0, 0] = ToNode.Voltage.PhaseA.Estimate.PerUnitComplexPhasor;
            Vj[1, 0] = ToNode.Voltage.PhaseB.Estimate.PerUnitComplexPhasor;
            Vj[2, 0] = ToNode.Voltage.PhaseC.Estimate.PerUnitComplexPhasor;

            ComplexMatrix Iij = (ThreePhaseSeriesAdmittance + ThreePhaseSingleSidedShuntSusceptance) * Vi - ThreePhaseSeriesAdmittance * Vj;
            ComplexMatrix Iji = (ThreePhaseSeriesAdmittance + ThreePhaseSingleSidedShuntSusceptance) * Vj - ThreePhaseSeriesAdmittance * Vi;

            FromNodeCurrent.PhaseA.Estimate.PerUnitComplexPhasor = Iij[0, 0];
            FromNodeCurrent.PhaseB.Estimate.PerUnitComplexPhasor = Iij[1, 0];
            FromNodeCurrent.PhaseC.Estimate.PerUnitComplexPhasor = Iij[2, 0];

            ToNodeCurrent.PhaseA.Estimate.PerUnitComplexPhasor = Iji[0, 0];
            ToNodeCurrent.PhaseB.Estimate.PerUnitComplexPhasor = Iji[1, 0];
            ToNodeCurrent.PhaseC.Estimate.PerUnitComplexPhasor = Iji[2, 0];
        }

        /// <summary>
        /// A string representation of the instance of the <see cref="SeriesBranchBase"/> class.
        /// </summary>
        /// <returns>A string representation of the instance of the <see cref="SeriesBranchBase"/> class.</returns>
        public override string ToString()
        {
            return "-TX- ID: " + InternalID.ToString() + " Number: " + Number.ToString();
        }

        /// <summary>
        /// A verbose string representation of the instance of the <see cref="SeriesBranchBase"/> class.
        /// </summary>
        /// <returns>A verbose string representation of the instance of the <see cref="SeriesBranchBase"/> class.</returns>
        public new string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Transformer --------------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + InternalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + Number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + Name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + Description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("ParentSubstation: " + m_parentSubstation.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat(" FromNodeCurrent: " + m_fromNodeCurrent.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("   ToNodeCurrent: " + m_toNodeCurrent.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat(RawImpedanceParameters.ToVerboseString());
            stringBuilder.AppendLine();
            return stringBuilder.ToString();

        }

        #endregion
    }
}
