﻿//******************************************************************************************************
//  TransmissionLine.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//  09/17/2013 - Kevin D. Jones
//       Added m_possibleImpedanceValues field.
//       Added m_childrenSeriesCompensationDevices field & associated property (SeriesCompensationDevices)
//       Added new declaration for m_childrenSeriesCompensationDevices in designated constructor.
//       Added public method InitializePossibleImpedanceValues()
//       Added private method ComputeBaseCaseImpedance()
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using SynchrophasorAnalytics.Measurements;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;

namespace SynchrophasorAnalytics.Modeling
{
    public class TransmissionLine : INetworkDescribable, ITwoTerminal
    {
        #region [ Private Members ]

        private int m_internalID;
        private int m_number;
        private string m_acronym;
        private string m_name;
        private string m_description;
        private bool m_complexPowerHasBeenInitialized;
        private Division m_parentDivision;

        private Substation m_fromSubstation;
        private int m_fromSubstationID;

        private Node m_fromNode;
        private int m_fromNodeID;

        private Substation m_toSubstation;
        private int m_toSubstationID;

        private Node m_toNode;
        private int m_toNodeID;

        private List<Node> m_childrenNodes;
        private List<LineSegment> m_childrenLineSegments;
        private List<SeriesCompensation> m_childrenSeriesCompensationDevices;
        private List<Switch> m_childrenSwitches;

        private CurrentPhasorGroup m_fromSubstationCurrent;
        private int m_fromSubstationCurrentID;

        private CurrentPhasorGroup m_toSubstationCurrent;
        private int m_toSubstationCurrentID;

        private ComplexPowerGroup m_fromSubstationComplexPower;
        private ComplexPowerGroup m_toSubstationComplexPower;

        private Impedance m_fromSideImpedanceToDeepestObservability;
        private Impedance m_toSideImpedanceToDeepestObservability;

        private Impedance m_realTimeCalculatedImpedance;

        private List<Impedance> m_possibleImpedanceValues;

        #endregion

        #region [ Properties ]

        /// <summary>
        /// A unique integer identifier for the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("ID")]
        public int InternalID
        {
            get 
            { 
                return m_internalID; 
            }
            set 
            { 
                m_internalID = value; 
            }
        }

        /// <summary>
        /// A descriptive number for the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("Number")]
        public int Number
        {
            get
            {
                return m_number;
            }
            set 
            { 
                m_number = value; 
            }
        }

        /// <summary>
        /// A descriptive acronym for the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("Acronym")]
        public string Acronym
        {
            get 
            { 
                return m_acronym; 
            }
            set 
            { 
                m_acronym = value; 
            }
        }

        /// <summary>
        /// A descriptive name for the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("Name")]
        public string Name
        {
            get 
            { 
                return m_name; 
            }
            set 
            { 
                m_name = value; 
            }
        }

        /// <summary>
        /// A description of the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlAttribute("Description")]
        public string Description
        {
            get 
            { 
                return m_description; 
            }
            set 
            {
                m_description = value; 
            }
        }

        /// <summary>
        /// Gets the type of the object
        /// </summary>
        [XmlIgnore()]
        public string ElementType
        {
            get 
            { 
                return this.GetType().ToString(); 
            }
        }

        /// <summary>
        /// The parent <see cref="Division"/> of the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlIgnore()]
        public Division ParentDivision
        {
            get 
            { 
                return m_parentDivision; 
            }
            set 
            { 
                m_parentDivision = value; 
            }
        }

        /// <summary>
        /// All of the children <see cref="Node"/> objects of the <see cref="TransmissionLine"/>.
        /// </summary>
        [XmlArray("Nodes")]
        public List<Node> Nodes
        {
            get
            {
                return m_childrenNodes;
            }
            set
            {
                m_childrenNodes = value;
                foreach (Node node in m_childrenNodes)
                {
                    node.ParentTransmissionLine = this;
                }
            }
        }

        [XmlArray("LineSegments")]
        public List<LineSegment> LineSegments
        {
            get
            {
                return m_childrenLineSegments;
            }
            set
            {
                m_childrenLineSegments = value;
                foreach (LineSegment lineSegment in m_childrenLineSegments)
                {
                    lineSegment.ParentTransmissionLine = this;
                }
            }
        }

        [XmlArray("SeriesCompensationDevices")]
        public List<SeriesCompensation> SeriesCompensationDevices
        {

            get
            {
                return m_childrenSeriesCompensationDevices;
            }
            set
            {
                m_childrenSeriesCompensationDevices = value;
                foreach (SeriesCompensation seriesCompensationDevice in m_childrenSeriesCompensationDevices)
                {
                    seriesCompensationDevice.ParentTransmissionLine = this;
                }
            }
        }

        [XmlArray("Switches")]
        public List<Switch> Switches
        {
            get
            {
                return m_childrenSwitches;
            }
            set
            {
                m_childrenSwitches = value;
                foreach (Switch lineSwitch in m_childrenSwitches)
                {
                    lineSwitch.ParentTransmissionLine = this;
                }
            }
        }

        [XmlIgnore()]
        public Node FromNode
        {
            get 
            {
                return m_fromNode;
            }
            set
            {
                m_fromNode = value;
            }
        }

        [XmlAttribute("FromNode")]
        public int FromNodeID
        {
            get
            {
                return m_fromNodeID;
            }
            set
            {
                m_fromNodeID = value;
            }
        }

        [XmlIgnore()]
        public Node ToNode
        {
            get
            {
                return m_toNode;
            }
            set
            {
                m_toNode = value;
            }
        }

        [XmlAttribute("ToNode")]
        public int ToNodeID
        {
            get
            {
                return m_toNodeID;
            }
            set
            {
                m_toNodeID = value;
            }
        }

        [XmlIgnore()]
        public Substation FromSubstation
        {
            get
            {
                return m_fromSubstation;
            }
            set
            {
                m_fromSubstation = value;
                m_fromSubstationID = value.InternalID;
            }
        }

        [XmlAttribute("FromSubstation")]
        public int FromSubstationID
        {
            get
            {
                return m_fromSubstationID;
            }
            set
            {
                m_fromSubstationID = value;
            }
        }

        [XmlIgnore()]
        public Substation ToSubstation
        {
            get
            {
                return m_toSubstation;
            }
            set
            {
                m_toSubstation = value;
                m_toSubstationID = value.InternalID;
            }
        }

        [XmlAttribute("ToSubstation")]
        public int ToSubstationID
        {
            get
            {
                return m_toSubstationID;
            }
            set
            {
                m_toSubstationID = value;
            }
        }

        [XmlElement("FromSubstationCurrent")]
        public CurrentPhasorGroup FromSubstationCurrent
        {
            get 
            { 
                return m_fromSubstationCurrent; 
            }
            set 
            { 
                m_fromSubstationCurrent = value;
                m_fromSubstationCurrentID = value.InternalID;
                if (FromNode != null)
                {
                    m_fromSubstationCurrent.MeasuredFromNode = FromNode;
                }
            }
        }

        [XmlAttribute("FromSubstationCurrentPhasorGroup")]
        public int FromSubstationCurrentID
        {
            get
            {
                return m_fromSubstationCurrentID;
            }
            set
            {
                m_fromSubstationCurrentID = value;
            }
        }

        [XmlElement("ToSubstationCurrent")]
        public CurrentPhasorGroup ToSubstationCurrent
        {
            get 
            { 
                return m_toSubstationCurrent; 
            }
            set 
            { 
                m_toSubstationCurrent = value;
                m_toSubstationCurrentID = value.InternalID;
                if (this.ToNode != null)
                {
                    m_toSubstationCurrent.MeasuredToNode = this.ToNode;
                }
            }
        }

        [XmlAttribute("ToSubstationCurrentPhasorGroup")]
        public int ToSubstationCurrentID
        {
            get
            {
                return m_toSubstationCurrentID;
            }
            set
            {
                m_toSubstationCurrentID = value;
            }
        }

        [XmlIgnore()]
        public Impedance FromSideImpedanceToDeepestObservability
        {
            get
            {
                return m_fromSideImpedanceToDeepestObservability;
            }
            set
            {
                m_fromSideImpedanceToDeepestObservability = value;
            }
        }

        [XmlIgnore()]
        public Impedance ToSideImpedanceToDeepestObservability
        {
            get
            {
                return m_toSideImpedanceToDeepestObservability;
            }
            set
            {
                m_toSideImpedanceToDeepestObservability = value;
            }
        }

        [XmlIgnore()]
        public ComplexPowerGroup FromSubstationComplexPower
        {
            get
            {
                if (m_complexPowerHasBeenInitialized)
                {
                    return m_fromSubstationComplexPower;
                }
                else
                {
                    return null;
                }
            }
            set
            {
                if (m_complexPowerHasBeenInitialized)
                {
                    m_fromSubstationComplexPower = value;
                }
            }
        }

        [XmlIgnore()]
        public ComplexPowerGroup ToSubstationComplexPower
        {
            get
            {
                return m_toSubstationComplexPower;
            }
            set
            {
                m_toSubstationComplexPower = value;
            }
        }

        [XmlIgnore()]
        public double PositiveSequenceRealPowerLosses
        {
            get
            {
                return Math.Abs(FromSubstationComplexPower.PositiveSequenceRealPower - ToSubstationComplexPower.PositiveSequenceRealPower);
            }
        }

        [XmlIgnore()]
        public double PositiveSequenceReactivePowerLosses
        {
            get
            {
                return Math.Abs(FromSubstationComplexPower.PositiveSequenceReactivePower - ToSubstationComplexPower.PositiveSequenceReactivePower);
            }
        }

        [XmlIgnore()]
        public double ThreePhaseRealPowerLosses
        {
            get
            {
                return Math.Abs(FromSubstationComplexPower.ThreePhaseRealPower - ToSubstationComplexPower.ThreePhaseRealPower);
            }
        }

        [XmlIgnore()]
        public double ThreePhaseReactivePowerLosses
        {
            get
            {
                return Math.Abs(FromSubstationComplexPower.ThreePhaseReactivePower - ToSubstationComplexPower.ThreePhaseReactivePower);
            }
        }

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default parameters.
        /// </summary>
        public TransmissionLine()
            :this(0, 0, "IDK","Undefined", "Undefined")
        {
        }

        public TransmissionLine(int internalID, int number, string acronym, string name, string description)
            :this(internalID, number, acronym, name, description, null, null)
        {
        }

        public TransmissionLine(int internalID, int number, string acronym, string name, string description, Substation fromSubstation, Substation toSubstation)
        {
            m_internalID = internalID;
            m_number = number;
            m_acronym = acronym;
            m_name = name;
            m_description = description;
            m_fromSubstation = fromSubstation;
            m_toSubstation = toSubstation;
            m_childrenLineSegments = new List<LineSegment>();
            m_childrenSeriesCompensationDevices = new List<SeriesCompensation>();
            m_childrenNodes = new List<Node>();
            m_childrenSwitches = new List<Switch>();
            m_complexPowerHasBeenInitialized = false;
        }

        #endregion

        #region [ Public Methods ]

        public void InitializePossibleImpedanceValues()
        {
            m_possibleImpedanceValues = new List<Impedance>();

            m_possibleImpedanceValues.Add(ComputeBaseCaseImpedance());
           
        }

        public void ComputeRealTimeImpedance()
        {
            PhasorMeasurement Vs = m_fromNode.Voltage.PositiveSequence.Measurement;
            PhasorMeasurement Vr = m_toNode.Voltage.PositiveSequence.Measurement;
            PhasorMeasurement Is = m_fromSubstationCurrent.PositiveSequence.Measurement;
            PhasorMeasurement Ir = m_toSubstationCurrent.PositiveSequence.Measurement;

            if (Vs.IncludeInEstimator && Vr.IncludeInEstimator && Ir.IncludeInEstimator && Is.IncludeInEstimator)
            {
                // Series Impedance
                double resistance = ((Vs.PerUnitComplexPhasor * Vs.PerUnitComplexPhasor - Vr.PerUnitComplexPhasor * Vr.PerUnitComplexPhasor) / (Vs.PerUnitComplexPhasor * Is.PerUnitComplexPhasor - Vr.PerUnitComplexPhasor * Ir.PerUnitComplexPhasor)).Re;
                double reactance = ((Vs.PerUnitComplexPhasor * Vs.PerUnitComplexPhasor - Vr.PerUnitComplexPhasor * Vr.PerUnitComplexPhasor) / (Vs.PerUnitComplexPhasor * Is.PerUnitComplexPhasor - Vr.PerUnitComplexPhasor * Ir.PerUnitComplexPhasor)).Im;
                // Shunt Susceptance
                double susceptance = 2 * ((Ir.PerUnitComplexPhasor + Is.PerUnitComplexPhasor)/(Vr.PerUnitComplexPhasor + Vs.PerUnitComplexPhasor)).Im;

                Impedance impedance = new Impedance()
                {
                    R1 = resistance,
                    R3 = resistance,
                    R6 = resistance,
                    X1 = reactance,
                    X3 = reactance,
                    X6 = reactance,
                    B1 = susceptance,
                    B3 = susceptance,
                    B6 = susceptance
                };

                m_realTimeCalculatedImpedance = impedance;
            }
        }

        public void InitializeComplexPower()
        {
            m_fromSubstationComplexPower = new ComplexPowerGroup(FromNode.Voltage, FromSubstationCurrent);
            m_fromSubstationComplexPower.UseEstimatedValues = true;

            m_toSubstationComplexPower = new ComplexPowerGroup(ToNode.Voltage, ToSubstationCurrent);
            m_toSubstationComplexPower.UseEstimatedValues = true;

            m_complexPowerHasBeenInitialized = true;
        }

        /// <summary>
        /// THIS METHOD IS A HACK THAT ONLY SUPPORTS ONE LINE SEGMENT PER TRANSMISSION LINE
        /// THIS ASSUMPTION IS VALID FOR THE DEMONSTRATION BUT NOT FOR THE FINAL PRODUCT
        /// </summary>
        public void ComputePositiveSequenceEstimatedCurrentFlow()
        {
            DoubleComplex Vi = FromNode.Voltage.PositiveSequence.Estimate.PerUnitComplexPhasor;
            DoubleComplex Vj = ToNode.Voltage.PositiveSequence.Estimate.PerUnitComplexPhasor;

            DoubleComplex Iij = (m_childrenLineSegments[0].PositiveSequenceSeriesAdmittance + m_childrenLineSegments[0].PositiveSequenceSingleSidedShuntSusceptance) * Vi - m_childrenLineSegments[0].PositiveSequenceSeriesAdmittance * Vj;
            DoubleComplex Iji = (m_childrenLineSegments[0].PositiveSequenceSeriesAdmittance + m_childrenLineSegments[0].PositiveSequenceSingleSidedShuntSusceptance) * Vj - m_childrenLineSegments[0].PositiveSequenceSeriesAdmittance * Vi;

            FromSubstationCurrent.PositiveSequence.Estimate.PerUnitComplexPhasor = Iij;
            ToSubstationCurrent.PositiveSequence.Estimate.PerUnitComplexPhasor = Iji;
        }

        /// <summary>
        /// THIS METHOD IS A HACK THAT ONLY SUPPORTS ONE LINE SEGMENT PER TRANSMISSION LINE
        /// THIS ASSUMPTION IS VALID FOR THE DEMONSTRATION BUT NOT FOR THE FINAL PRODUCT
        /// </summary>
        public void ComputeThreePhaseEstimatedCurrentFlow()
        {
            ComplexMatrix Vi = ComplexMatrix.Create(3, 1, 0);
            ComplexMatrix Vj = ComplexMatrix.Create(3, 1, 0);

            Vi[0, 0] = FromNode.Voltage.PhaseA.Estimate.PerUnitComplexPhasor;
            Vi[1, 0] = FromNode.Voltage.PhaseB.Estimate.PerUnitComplexPhasor;
            Vi[2, 0] = FromNode.Voltage.PhaseC.Estimate.PerUnitComplexPhasor;

            Vj[0, 0] = ToNode.Voltage.PhaseA.Estimate.PerUnitComplexPhasor;
            Vj[1, 0] = ToNode.Voltage.PhaseB.Estimate.PerUnitComplexPhasor;
            Vj[2, 0] = ToNode.Voltage.PhaseC.Estimate.PerUnitComplexPhasor;

            ComplexMatrix Iij = (m_childrenLineSegments[0].ThreePhaseSeriesAdmittance + m_childrenLineSegments[0].ThreePhaseSingleSidedShuntSusceptance) * Vi - m_childrenLineSegments[0].ThreePhaseSeriesAdmittance * Vj;
            ComplexMatrix Iji = (m_childrenLineSegments[0].ThreePhaseSeriesAdmittance + m_childrenLineSegments[0].ThreePhaseSingleSidedShuntSusceptance) * Vj - m_childrenLineSegments[0].ThreePhaseSeriesAdmittance * Vi;

            FromSubstationCurrent.PhaseA.Estimate.PerUnitComplexPhasor = Iij[0, 0];
            FromSubstationCurrent.PhaseB.Estimate.PerUnitComplexPhasor = Iij[1, 0];
            FromSubstationCurrent.PhaseC.Estimate.PerUnitComplexPhasor = Iij[2, 0];

            ToSubstationCurrent.PhaseA.Estimate.PerUnitComplexPhasor = Iji[0, 0];
            ToSubstationCurrent.PhaseB.Estimate.PerUnitComplexPhasor = Iji[1, 0];
            ToSubstationCurrent.PhaseC.Estimate.PerUnitComplexPhasor = Iji[2, 0];
        }

        /// <summary>
        /// THIS METHOD IS A HACK THAT ONLY SUPPORTS ONE LINE SEGMENT PER TRANSMISSION LINE
        /// THIS ASSUMPTION IS VALID FOR THE DEMONSTRATION BUT NOT FOR THE FINAL PRODUCT
        /// </summary>
        public void SetImpedanceValues()
        {
            m_fromSideImpedanceToDeepestObservability = m_childrenLineSegments[0].RawImpedanceParameters;
            m_toSideImpedanceToDeepestObservability = m_childrenLineSegments[0].RawImpedanceParameters;
        }

        /// <summary>
        /// A string representation of an instance of the <see cref="TransmissionLine"/> class.
        /// </summary>
        /// <returns>A string representation of an instance of the <see cref="TransmissionLine"/> class.</returns>
        public override string ToString()
        {
            return "-TL- ID: " + m_internalID.ToString() + " Number: " + m_number.ToString();
        }

        /// <summary>
        /// A verbose string representation of an instance of the <see cref="TransmissionLine"/> class.
        /// </summary>
        /// <returns>A verbose string representation of an instance of the <see cref="TransmissionLine"/> class.</returns>
        public string ToVerboseString()
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("----- Transmission Line --------------------------------------------------------");
            stringBuilder.AppendLine();
            stringBuilder.AppendFormat("      InternalID: " + m_internalID.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          Number: " + m_number.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("         Acronym: " + m_acronym + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("            Name: " + m_name + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("     Description: " + m_description + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("  FromSubstation: " + m_fromSubstation.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("    ToSubstation: " + m_toSubstation.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("        FromNode: " + FromNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("          ToNode: " + ToNode.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("  ParentDivision: " + m_parentDivision.ToString() + "{0}", Environment.NewLine);
            stringBuilder.AppendFormat("           Nodes:  {0}", Environment.NewLine);
            stringBuilder.AppendLine();
            foreach (Node node in m_childrenNodes)
            {
                stringBuilder.AppendFormat(node.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendFormat("    Transformers:  {0}", Environment.NewLine);
            stringBuilder.AppendLine();
            foreach (LineSegment lineSegments in m_childrenLineSegments)
            {
                stringBuilder.AppendFormat(lineSegments.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendFormat("        Switches:  {0}", Environment.NewLine);
            stringBuilder.AppendLine();
            foreach (Switch circuitSwitch in m_childrenSwitches)
            {
                stringBuilder.AppendFormat(circuitSwitch.ToString() + "{0}", Environment.NewLine);
            }
            stringBuilder.AppendLine();
            return stringBuilder.ToString();
        }

        public TransmissionLine Copy()
        {
            TransmissionLine copy = (TransmissionLine)this.MemberwiseClone();
            // Incomplete
            return copy;
        }

        #endregion

        #region [ Private Methods ]

        private Impedance ComputeBaseCaseImpedance()
        {
            // Calculate the base case with no series compensation
            Impedance impedance = new Impedance();

            foreach (LineSegment lineSegment in m_childrenLineSegments)
            {
                impedance += lineSegment.RawImpedanceParameters;
            }

            return impedance;
        }

        #endregion
    }
}
