﻿//******************************************************************************************************
//  Network.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/01/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;
using Extreme.Mathematics;
using Extreme.Mathematics.LinearAlgebra;
using Extreme.Mathematics.LinearAlgebra.Complex;
using SynchrophasorAnalytics.Graph;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Measurements;
using SynchrophasorAnalytics.Matrices;
using SynchrophasorAnalytics.Calibration;

namespace SynchrophasorAnalytics.Networks
{
    [Serializable()]
    public class Network
    {
        #region [ Private Members ]

        private SystemMatrix m_systemMatrix;
        private NetworkModel m_networkModel;
        private bool m_hasChangedSinceLastIteration;
        private bool[] m_pastDiscreteVoltagePhasorState;
        private bool[] m_pastDiscreteCurrentPhasorState;
        private int[] m_pastBreakerStatusState;
        private double[] m_pastStatusWordState;

        //private NetworkMeasurements m_networkMeasurements;
        //private Dictionary<string, double> m_rawMeasurementKeyValuePairs;

        #endregion

        #region [ Properties ]

        ///// <summary>
        ///// A Dictionary with the raw measurements and their respective measurement keys.
        ///// </summary>
        //[XmlIgnore()]
        //public Dictionary<string, double> RawMeasurementKeyValuePairs
        //{
        //    get 
        //    { 
        //        return m_rawMeasurementKeyValuePairs; 
        //    }
        //    set 
        //    { 
        //        m_rawMeasurementKeyValuePairs = value; 
        //    }
        //}

        /// <summary>
        /// Determines whether to treat the network as a <see cref="PhaseSelection.PositiveSequence"/> approximation or as a full <see cref="PhaseSelection.ThreePhase"/> representation.
        /// </summary>
        [XmlIgnore()]
        public PhaseSelection PhaseConfiguration
        {
            get 
            {
                return m_networkModel.PhaseConfiguration; 
            }
            set 
            {
                m_networkModel.PhaseConfiguration = value; 
            }
        }

        /// <summary>
        /// The physical representation of the network.
        /// </summary>
        [XmlElement("Model")]
        public NetworkModel Model
        {
            get 
            { 
                return m_networkModel; 
            }
            set 
            { 
                m_networkModel = value; 
            }
        }

        /// <summary>
        /// 
        /// </summary>
        [XmlIgnore()]
        public SystemMatrix Matrix
        {
            get
            {
                return m_systemMatrix;
            }
        }

        [XmlIgnore()]
        public bool HasChangedSinceLastIteration
        {
            get
            {
                return m_hasChangedSinceLastIteration;
            }
        }

        [XmlIgnore()]
        public ObservedBus PerfectPtBusForPrimaryPhasorCalibration
        {
            get
            {
                foreach (ObservedBus observedBus in m_networkModel.ObservedBusses)
                {
                    foreach (Node node in observedBus.Nodes)
                    {
                        if (node.Voltage.PhaseA.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect &&
                            node.Voltage.PhaseB.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect &&
                            node.Voltage.PhaseC.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect)
                        {
                            return observedBus;
                        }
                    }
                }
                return null;
            }
        }

        #region [ State Vector ]

        [XmlIgnore()]
        public ComplexMatrix StateVector
        {
            get
            {
                bool usePerUnit = false;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    return GetPositiveSequenceStateVectorFromModel(usePerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    return GetThreePhaseStateVectorFromModel(usePerUnit);
                }
                else
                {
                    return null;
                }
            }
            set
            {
                bool isPerUnit = false;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    SendPositiveSequenceStateVectorToModel(value, isPerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    SendThreePhaseStateVectorToModel(value, isPerUnit);
                }
            }
        }

        [XmlIgnore()]
        public ComplexMatrix PerUnitStateVector
        {
            get
            {
                bool usePerUnit = true;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    return GetPositiveSequenceStateVectorFromModel(usePerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    return GetThreePhaseStateVectorFromModel(usePerUnit);
                }
                else
                {
                    return null;
                }
            }
            set
            {
                bool isPerUnit = true;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    SendPositiveSequenceStateVectorToModel(value, isPerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    SendThreePhaseStateVectorToModel(value, isPerUnit);
                }
            }
        }

        [XmlIgnore()]
        public double[] StateVectorMagnitude
        {
            get
            {
                ComplexMatrix stateVector = StateVector;
                double[] stateVectorMagnitude = new double[stateVector.RowCount];
                for (int i = 0; i < stateVector.RowCount; i++)
                {
                    stateVectorMagnitude[i] = Math.Sqrt(stateVector[i, 0].Re * stateVector[i, 0].Re + stateVector[i, 0].Im * stateVector[i, 0].Im);
                }
                return stateVectorMagnitude;
            }
        }

        [XmlIgnore()]
        public double[] PerUnitStateVectorMagnitude
        {
            get
            {
                ComplexMatrix perUnitStateVector = PerUnitStateVector;
                double[] stateVectorMagnitude = new double[perUnitStateVector.RowCount];
                for (int i = 0; i < perUnitStateVector.RowCount; i++)
                {
                    stateVectorMagnitude[i] = Math.Sqrt(perUnitStateVector[i, 0].Re * perUnitStateVector[i, 0].Re + perUnitStateVector[i, 0].Im * perUnitStateVector[i, 0].Im);
                }
                return stateVectorMagnitude;
            }
        }

        [XmlIgnore()]
        public double[] StateVectorAngleInDegrees
        {
            get
            {
                ComplexMatrix stateVector = StateVector;
                double[] stateVectorAngleInDegrees = new double[stateVector.RowCount];
                for (int i = 0; i < stateVector.RowCount; i++)
                {
                    PhasorBase phasorPlaceHolder = new PhasorBase();
                    phasorPlaceHolder.ComplexPhasor = stateVector[i, 0];
                    stateVectorAngleInDegrees[i] = phasorPlaceHolder.AngleInDegrees;
                }
                return stateVectorAngleInDegrees;
            }
        }

        [XmlIgnore()]
        public double[] StateVectorAngleInRadians
        {
            get
            {
                ComplexMatrix stateVector = StateVector;
                double[] stateVectorAngleInRadians = new double[stateVector.RowCount];
                List<PhasorBase> phasorContainer = new List<PhasorBase>();
                for (int i = 0; i < stateVector.RowCount; i++)
                {
                    PhasorBase phasorBase = new PhasorBase();
                    phasorBase.ComplexPhasor = stateVector[i, 0];
                    stateVectorAngleInRadians[i] = phasorBase.AngleInRadians;
                }
                return stateVectorAngleInRadians;
            }
        }

        #endregion

        
        #region [ Measurement Vector ]

        [XmlIgnore()]
        public ComplexMatrix MeasurementVector
        {
            get
            {
                bool usePerUnit = false;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    return GetPositiveSequenceMeasurementVectorFromModel(usePerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    return GetThreePhaseMeasurementVectorFromModel(usePerUnit);
                }
                else
                {
                    return null;
                }
            }
        }

        [XmlIgnore()]
        public ComplexMatrix PerUnitMeasurementVector
        {
            get
            {
                bool usePerUnit = true;
                if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    return GetPositiveSequenceMeasurementVectorFromModel(usePerUnit);
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    return GetThreePhaseMeasurementVectorFromModel(usePerUnit);
                }
                else
                {
                    return null;
                }
            }
        }

        [XmlIgnore()]
        public double[] MeasurementVectorMagnitude
        {
            get
            {
                ComplexMatrix measurementVector = MeasurementVector;
                double[] measurementVectorMagnitude = new double[measurementVector.RowCount];
                for (int i = 0; i < measurementVector.RowCount; i++)
                {
                    measurementVectorMagnitude[i] = Math.Sqrt(measurementVector[i, 0].Re * measurementVector[i, 0].Re + measurementVector[i, 0].Im * measurementVector[i, 0].Im);
                }
                return measurementVectorMagnitude;
            }
        }

        [XmlIgnore()]
        public double[] PerUnitMeasurementVectorMagnitude
        {
            get
            {
                ComplexMatrix perUnitMeasurementVector = PerUnitMeasurementVector;
                double[] measurementVectorMagnitude = new double[perUnitMeasurementVector.RowCount];
                for (int i = 0; i < perUnitMeasurementVector.RowCount; i++)
                {
                    measurementVectorMagnitude[i] = Math.Sqrt(perUnitMeasurementVector[i, 0].Re * perUnitMeasurementVector[i, 0].Re + perUnitMeasurementVector[i, 0].Im * perUnitMeasurementVector[i, 0].Im);
                }
                return measurementVectorMagnitude;
            }
        }

        [XmlIgnore()]
        public double[] MeasurementVectorAngleInDegrees
        {
            get
            {
                ComplexMatrix measurementVector = MeasurementVector;
                double[] measurementVectorAngleInDegrees = new double[measurementVector.RowCount];
                for (int i = 0; i < measurementVector.RowCount; i++)
                {
                    PhasorBase phasorPlaceHolder = new PhasorBase();
                    phasorPlaceHolder.ComplexPhasor = measurementVector[i, 0];
                    measurementVectorAngleInDegrees[i] = phasorPlaceHolder.AngleInDegrees;
                }
                return measurementVectorAngleInDegrees;
            }
        }

        [XmlIgnore()]
        public double[] MeasurementVectorAngleInRadians
        {
            get
            {
                ComplexMatrix measurementVector = MeasurementVector;
                double[] measurementVectorAngleInRadians = new double[measurementVector.RowCount];
                for (int i = 0; i < measurementVector.RowCount; i++)
                {
                    PhasorBase phasorPlaceHolder = new PhasorBase();
                    phasorPlaceHolder.ComplexPhasor = measurementVector[i, 0];
                    measurementVectorAngleInRadians[i] = phasorPlaceHolder.AngleInRadians;
                }
                return measurementVectorAngleInRadians;
            }
        }

        [XmlIgnore()]
        public ComplexMatrix CurrentMeasurementVectorForPrimaryPhasorCalibration
        {
            get
            {
                List<CurrentPhasorGroup> currentPhasorsForPrimaryPhasorCalibration = new List<CurrentPhasorGroup>();

                foreach (CurrentPhasorGroup currentPhasorGroup in m_networkModel.ActiveCurrentPhasors)
                {
                    if (currentPhasorGroup.PhaseA.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                        currentPhasorGroup.PhaseB.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                        currentPhasorGroup.PhaseC.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active)
                    {
                        currentPhasorsForPrimaryPhasorCalibration.Add(currentPhasorGroup);
                    }
                }

                ComplexMatrix currentMeasurementVectorForPrimaryPhasorCalibration = ComplexMatrix.Create(3 * currentPhasorsForPrimaryPhasorCalibration.Count(), 1, 0);

                for (int i = 0; i < currentMeasurementVectorForPrimaryPhasorCalibration.RowCount / 3; i++)
                {
                    currentMeasurementVectorForPrimaryPhasorCalibration[3 * i, 0] = currentPhasorsForPrimaryPhasorCalibration[i].PhaseA.Measurement.PerUnitComplexPhasor;
                    currentMeasurementVectorForPrimaryPhasorCalibration[3 * i + 1, 0] = currentPhasorsForPrimaryPhasorCalibration[i].PhaseB.Measurement.PerUnitComplexPhasor;
                    currentMeasurementVectorForPrimaryPhasorCalibration[3 * i + 2, 0] = currentPhasorsForPrimaryPhasorCalibration[i].PhaseC.Measurement.PerUnitComplexPhasor;
                }

                return currentMeasurementVectorForPrimaryPhasorCalibration;
            }
        }

        [XmlIgnore()]
        public ComplexMatrix VoltageMeasurementVectorForPrimaryPhasorCalibration
        {
            get
            {
                List<VoltagePhasorGroup> voltagePhasorForPrimaryPhasorCalibration = new List<VoltagePhasorGroup>();

                foreach (ObservedBus observedBus in m_networkModel.ObservedBusses)
                {
                    foreach (Node node in observedBus.Nodes)
                    {
                        if ((node.Voltage.PhaseA.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                             node.Voltage.PhaseA.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect) &&
                            (node.Voltage.PhaseB.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                             node.Voltage.PhaseB.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect) &&
                            (node.Voltage.PhaseC.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Active ||
                             node.Voltage.PhaseC.Measurement.InstrumentTransformerCalibrationSetting == CalibrationSetting.Perfect))
                        {
                            voltagePhasorForPrimaryPhasorCalibration.Add(node.Voltage);
                        }
                    }
                }

                ComplexMatrix voltageMeasurementVectorForPrimaryPhasorCalibration = ComplexMatrix.Create(3 * voltagePhasorForPrimaryPhasorCalibration.Count(), 1, 0);

                for (int i = 0; i < voltageMeasurementVectorForPrimaryPhasorCalibration.RowCount / 3; i++)
                {
                    voltageMeasurementVectorForPrimaryPhasorCalibration[3 * i, 0] = voltagePhasorForPrimaryPhasorCalibration[i].PhaseA.Measurement.PerUnitComplexPhasor;
                    voltageMeasurementVectorForPrimaryPhasorCalibration[3 * i + 1, 0] = voltagePhasorForPrimaryPhasorCalibration[i].PhaseB.Measurement.PerUnitComplexPhasor;
                    voltageMeasurementVectorForPrimaryPhasorCalibration[3 * i + 2, 0] = voltagePhasorForPrimaryPhasorCalibration[i].PhaseC.Measurement.PerUnitComplexPhasor;
                }

                return voltageMeasurementVectorForPrimaryPhasorCalibration;
            }
        }

        #endregion

        #endregion

        #region [ Constructors ]

        /// <summary>
        /// A blank constructor with default values.
        /// </summary>
        public Network()
            :this(new NetworkModel())
        {
        }

        /// <summary>
        /// The designated constructor for the <see cref="Network"/> class.
        /// </summary>
        /// <param name="networkModel">The virtualized <see cref="NetworkModel"/>.</param>
        public Network(NetworkModel networkModel)
        {
            m_networkModel = networkModel;
            m_hasChangedSinceLastIteration = true;
        }

        #endregion

        #region [ Private Methods ]

        private ComplexMatrix GetPositiveSequenceStateVectorFromModel(bool usePerUnit)
        {
            List<VoltagePhasorGroup> nodeVoltages = m_networkModel.Voltages;

            ComplexMatrix stateVector = ComplexMatrix.Create(nodeVoltages.Count, 1);

            if (usePerUnit)
            {
                for (int i = 0; i < nodeVoltages.Count; i++)
                {
                    stateVector[i, 0] = nodeVoltages[i].PositiveSequence.Estimate.PerUnitComplexPhasor;
                }
            }
            else
            {
                for (int i = 0; i < nodeVoltages.Count; i++)
                {
                    stateVector[i, 0] = nodeVoltages[i].PositiveSequence.Estimate.ComplexPhasor;
                }
            }

            return stateVector;
        }

        private ComplexMatrix GetThreePhaseStateVectorFromModel(bool usePerUnit)
        {
            List<VoltagePhasorGroup> nodeVoltages = m_networkModel.Voltages;

            ComplexMatrix stateVector = ComplexMatrix.Create(3 * nodeVoltages.Count, 1);

            if (usePerUnit)
            {
                for (int i = 0; i < nodeVoltages.Count; i++)
                {
                    stateVector[3 * i, 0] = nodeVoltages[i].PhaseA.Estimate.PerUnitComplexPhasor;
                    stateVector[3 * i + 1, 0] = nodeVoltages[i].PhaseB.Estimate.PerUnitComplexPhasor;
                    stateVector[3 * i + 2, 0] = nodeVoltages[i].PhaseC.Estimate.PerUnitComplexPhasor;
                }
            }
            else
            {
                for (int i = 0; i < nodeVoltages.Count; i++)
                {
                    stateVector[3 * i, 0] = nodeVoltages[i].PhaseA.Estimate.ComplexPhasor;
                    stateVector[3 * i + 1, 0] = nodeVoltages[i].PhaseB.Estimate.ComplexPhasor;
                    stateVector[3 * i + 2, 0] = nodeVoltages[i].PhaseC.Estimate.ComplexPhasor;
                }
            }

            return stateVector;
        }

        private ComplexMatrix GetPositiveSequenceMeasurementVectorFromModel(bool usePerUnit)
        {
            List<CurrentPhasorGroup> measuredCurrents = m_networkModel.ActiveCurrentPhasors;

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = m_networkModel.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                    }
                }
            }

            ComplexMatrix voltageMeasurementVector = ComplexMatrix.Create(measuredNodes.Count, 1);
            ComplexMatrix currentMeasurementVector = ComplexMatrix.Create(measuredCurrents.Count, 1);

            if (usePerUnit)
            {
                for (int i = 0; i < measuredNodes.Count; i++)
                {
                    voltageMeasurementVector[i, 0] = measuredNodes[i].Voltage.PositiveSequence.Measurement.PerUnitComplexPhasor;
                }

                for (int i = 0; i < measuredCurrents.Count; i++)
                {
                    currentMeasurementVector[i, 0] = measuredCurrents[i].PositiveSequence.Measurement.PerUnitComplexPhasor;
                }
            }
            else
            {
                for (int i = 0; i < measuredNodes.Count; i++)
                {
                    voltageMeasurementVector[i, 0] = measuredNodes[i].Voltage.PositiveSequence.Measurement.ComplexPhasor;
                }

                for (int i = 0; i < measuredCurrents.Count; i++)
                {
                    currentMeasurementVector[i, 0] = measuredCurrents[i].PositiveSequence.Measurement.ComplexPhasor;
                }
            }

            return MatrixCalculationExtensions.VerticallyConcatenate(voltageMeasurementVector, currentMeasurementVector);
        }

        private ComplexMatrix GetThreePhaseMeasurementVectorFromModel(bool usePerUnit) 
        {
            List<CurrentPhasorGroup> measuredCurrents = m_networkModel.ActiveCurrentPhasors;

            // Resolve the Network into a list of ObservedBusses
            List<ObservedBus> observedBusses = m_networkModel.ObservedBusses;

            // Create a list of directly measured nodes from the set of ObservedBusses
            List<Node> measuredNodes = new List<Node>();
            foreach (ObservedBus observedBus in observedBusses)
            {
                foreach (Node node in observedBus.Nodes)
                {
                    if (node.Observability == ObservationState.DirectlyObserved)
                    {
                        measuredNodes.Add(node);
                    }
                }
            }

            ComplexMatrix voltageMeasurementVector = ComplexMatrix.Create(3 * measuredNodes.Count, 1);
            ComplexMatrix currentMeasurementVector = ComplexMatrix.Create(3 * measuredCurrents.Count, 1);

            if (usePerUnit)
            {
                for (int i = 0; i < measuredNodes.Count; i++)
                {
                    voltageMeasurementVector[3 * i, 0] = measuredNodes[i].Voltage.PhaseA.Measurement.PerUnitComplexPhasor;
                    voltageMeasurementVector[3 * i + 1, 0] = measuredNodes[i].Voltage.PhaseB.Measurement.PerUnitComplexPhasor;
                    voltageMeasurementVector[3 * i + 2, 0] = measuredNodes[i].Voltage.PhaseC.Measurement.PerUnitComplexPhasor;
                }

                for (int i = 0; i < measuredCurrents.Count; i++)
                {
                    currentMeasurementVector[3 * i, 0] = measuredCurrents[i].PhaseA.Measurement.PerUnitComplexPhasor;
                    currentMeasurementVector[3 * i + 1, 0] = measuredCurrents[i].PhaseB.Measurement.PerUnitComplexPhasor;
                    currentMeasurementVector[3 * i + 2, 0] = measuredCurrents[i].PhaseC.Measurement.PerUnitComplexPhasor;
                }
            }
            else
            {
                for (int i = 0; i < measuredNodes.Count; i++)
                {
                    voltageMeasurementVector[3 * i, 0] = measuredNodes[i].Voltage.PhaseA.Measurement.ComplexPhasor;
                    voltageMeasurementVector[3 * i + 1, 0] = measuredNodes[i].Voltage.PhaseB.Measurement.ComplexPhasor;
                    voltageMeasurementVector[3 * i + 2, 0] = measuredNodes[i].Voltage.PhaseC.Measurement.ComplexPhasor;
                }

                for (int i = 0; i < measuredCurrents.Count; i++)
                {
                    currentMeasurementVector[3 * i, 0] = measuredCurrents[i].PhaseA.Measurement.ComplexPhasor;
                    currentMeasurementVector[3 * i + 1, 0] = measuredCurrents[i].PhaseB.Measurement.ComplexPhasor;
                    currentMeasurementVector[3 * i + 2, 0] = measuredCurrents[i].PhaseC.Measurement.ComplexPhasor;
                }
            }

            return MatrixCalculationExtensions.VerticallyConcatenate(voltageMeasurementVector, currentMeasurementVector);
        }

        private void SendPositiveSequenceStateVectorToModel(ComplexMatrix stateVector, bool isPerUnit)
        {
            List<ObservedBus> observedBusses = m_networkModel.ObservedBusses;

            for (int i = 0; i < observedBusses.Count; i++)
            {
                VoltagePhasorGroup voltagePhasorGroup = new VoltagePhasorGroup();
                voltagePhasorGroup.PositiveSequence.Estimate.BaseKV = observedBusses[i].BaseKV;
                voltagePhasorGroup.PositiveSequence.Estimate.PerUnitComplexPhasor = stateVector[i, 0];
                observedBusses[i].Value = voltagePhasorGroup;
            }
        }

        private void SendThreePhaseStateVectorToModel(ComplexMatrix stateVector, bool isPerUnit)
        {
            List<ObservedBus> observedBusses = m_networkModel.ObservedBusses;

            for (int i = 0; i < observedBusses.Count; i++)
            {
                VoltagePhasorGroup voltagePhasorGroup = new VoltagePhasorGroup();
                voltagePhasorGroup.PhaseA.Estimate.BaseKV = observedBusses[i].BaseKV;
                voltagePhasorGroup.PhaseB.Estimate.BaseKV = observedBusses[i].BaseKV;
                voltagePhasorGroup.PhaseC.Estimate.BaseKV = observedBusses[i].BaseKV;
                voltagePhasorGroup.PhaseA.Estimate.PerUnitComplexPhasor = stateVector[3 * i, 0];
                voltagePhasorGroup.PhaseB.Estimate.PerUnitComplexPhasor = stateVector[3 * i + 1, 0];
                voltagePhasorGroup.PhaseC.Estimate.PerUnitComplexPhasor = stateVector[3 * i + 2, 0];
                voltagePhasorGroup.ComputeSequenceComponents();
                observedBusses[i].Value = voltagePhasorGroup;
            }
        }

        //public List<VoltagePhasorGroup> NodeVoltages()
        //{
        //    List<VoltagePhasorGroup> nodeVoltages = new List<VoltagePhasorGroup>();

        //    foreach (Company company in m_networkModel.Companies)
        //    {
        //        foreach (Division division in company.Divisions)
        //        {
        //            foreach (Substation substation in division.Substations)
        //            {
        //                foreach (Node node in substation.Nodes)
        //                {
        //                    if (node.Voltage != null)
        //                    {
        //                        nodeVoltages.Add(node.Voltage);
        //                    }
        //                }
        //            }
        //        }
        //    }

        //    return nodeVoltages;
        //}

        //public List<CurrentPhasorGroup> CurrentFlows()
        //{
        //    List<CurrentPhasorGroup> currentFlows = new List<SynchrophasorAnalytics.Measurements.CurrentPhasorGroup>();

        //    foreach (Company company in m_networkModel.Companies)
        //    {
        //        foreach (Division division in company.Divisions)
        //        {
        //            foreach (Substation substation in division.Substations)
        //            {
        //                foreach (Transformer transformer in substation.Transformers)
        //                {
        //                    if (transformer.FromNodeCurrent != null)
        //                    {
        //                        currentFlows.Add(transformer.FromNodeCurrent);
        //                    }

        //                    if (transformer.ToNodeCurrent != null)
        //                    {
        //                        currentFlows.Add(transformer.ToNodeCurrent);
        //                    }
        //                }
        //            }

        //            foreach (TransmissionLine transmissionLine in division.TransmissionLines)
        //            {
        //                if (transmissionLine.FromSubstationCurrent != null)
        //                {
        //                    currentFlows.Add(transmissionLine.FromSubstationCurrent);
        //                }

        //                if (transmissionLine.ToSubstationCurrent != null)
        //                {
        //                    currentFlows.Add(transmissionLine.ToSubstationCurrent);
        //                }
        //            }
        //        }
        //    }

        //    return currentFlows;
        //}

        #endregion

        #region [ Public Methods ]

        /// <summary>
        /// Initializes the <see cref="Network"/> by re-establishing references that are not preserved hierarchically in the configuration file.
        /// </summary>
        public void Initialize()
        {
            // Initialize children
            m_networkModel.Initialize();

            InitializePastDiscreteStates();
            //UpdatePastDiscreteStates();
        }

        /// <summary>
        /// Computes the current state of the network
        /// </summary>
        public void ComputeSystemState()
        {
            if (m_systemMatrix == null || m_hasChangedSinceLastIteration)
            {
                m_systemMatrix = new SystemMatrix(this);
            }
            
            PerUnitStateVector = m_systemMatrix.PsuedoInverseOfMatrix * PerUnitMeasurementVector;
        }

        private void UpdatePastDiscreteStates()
        {

            for (int i = 0; i < m_networkModel.Voltages.Count(); i++)
            {
                if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    m_pastDiscreteVoltagePhasorState[i] = m_networkModel.Voltages[i].IncludeInEstimator;
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    m_pastDiscreteVoltagePhasorState[i] = m_networkModel.Voltages[i].IncludeInPositiveSequenceEstimator;
                }
            }

            for (int i = 0; i < m_networkModel.Currents.Count(); i++)
            {
                if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    m_pastDiscreteCurrentPhasorState[i] = m_networkModel.Currents[i].IncludeInEstimator;
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    m_pastDiscreteCurrentPhasorState[i] = m_networkModel.Currents[i].IncludeInPositiveSequenceEstimator;
                }
            }

            for (int i = 0; i < m_networkModel.BreakerStatuses.Count(); i++)
            {
                m_pastBreakerStatusState[i] = m_networkModel.BreakerStatuses[i].BinaryValue;
            }

            for (int i = 0; i < m_networkModel.StatusWords.Count(); i++)
            {
                m_pastStatusWordState[i] = m_networkModel.StatusWords[i].BinaryValue;
            }

        }

        private void InitializePastDiscreteStates()
        {

            m_pastDiscreteVoltagePhasorState = new bool[m_networkModel.Voltages.Count()];
            m_pastDiscreteCurrentPhasorState = new bool[m_networkModel.Currents.Count()];
            m_pastBreakerStatusState = new int[m_networkModel.BreakerStatuses.Count()];
            m_pastStatusWordState = new double[m_networkModel.StatusWords.Count()];

            for (int i = 0; i < m_pastDiscreteVoltagePhasorState.Length; i++)
            {
                m_pastDiscreteVoltagePhasorState[i] = true;
            }

            if (m_pastDiscreteCurrentPhasorState.Length > 0)
            {
                for (int i = 0; i < m_pastDiscreteCurrentPhasorState.Length; i++)
                {
                    m_pastDiscreteCurrentPhasorState[i] = true;
                }
            }
            if (m_pastBreakerStatusState.Length > 0)
            {
                for (int i = 0; i < m_pastBreakerStatusState.Length; i++)
                {
                    m_pastBreakerStatusState[i] = 0;
                }
            }

            if (m_pastStatusWordState.Length > 0)
            {
                for (int i = 0; i < m_pastStatusWordState.Length; i++)
                {
                    m_pastStatusWordState[i] = 0;
                }
            }
        }

        private bool ComparePresentAndPastDiscreteStates()
        {
            for (int i = 0; i < m_networkModel.Voltages.Count(); i++)
            {
                if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    if (m_pastDiscreteVoltagePhasorState[i] != m_networkModel.Voltages[i].IncludeInEstimator)
                    {
                        return true;
                    }
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    if (m_pastDiscreteVoltagePhasorState[i] != m_networkModel.Voltages[i].IncludeInPositiveSequenceEstimator)
                    {
                        return true;
                    }
                }
            }

            for (int i = 0; i < m_networkModel.Currents.Count(); i++)
            {
                if (m_networkModel.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    if (m_pastDiscreteCurrentPhasorState[i] != m_networkModel.Currents[i].IncludeInEstimator)
                    {
                        return true;
                    }
                }
                else if (m_networkModel.PhaseConfiguration == PhaseSelection.PositiveSequence)
                {
                    if (m_pastDiscreteCurrentPhasorState[i] != m_networkModel.Currents[i].IncludeInPositiveSequenceEstimator)
                    {
                        return true;
                    }
                }
            }

            for (int i = 0; i < m_networkModel.BreakerStatuses.Count(); i++)
            {
                if (m_pastBreakerStatusState[i] != m_networkModel.BreakerStatuses[i].BinaryValue)
                {
                    return true;
                }
            }

            for (int i = 0; i < m_networkModel.StatusWords.Count(); i++)
            {
                if (m_pastStatusWordState[i] != m_networkModel.StatusWords[i].BinaryValue)
                {
                    return true;
                }
            }

            return false;
        }

        public void RunNetworkReconstructionCheck()
        {
            if (ComparePresentAndPastDiscreteStates())
            {
                m_hasChangedSinceLastIteration = true;
                UpdatePastDiscreteStates();
            }
            else
            {
                m_hasChangedSinceLastIteration = false;
            }
        }

        #endregion

        #region [ Xml Serialization/Deserialization methods ]

        /// <summary>
        /// Creates a new <see cref="Network"/> by deserializing the configuration file from the specified location.
        /// </summary>
        /// <param name="pathName">The location of the configuration file including the file name.</param>
        /// <returns>A new (but uninitialized) <see cref="Network"/> based on the configuration file. Must call <see cref="NetworkModel.Initialize"/> method 
        /// to reconstitute parent, child, and sibling references in the <see cref="NetworkModel"/>.</returns>
        public static Network DeserializeFromXml(string pathName)
        {
            try
            {
                // Create an empy NetworkMeasurements object reference.
                Network network = null;

                // Create an XmlSerializer with the type of NetworkMeasurements.
                XmlSerializer deserializer = new XmlSerializer(typeof(Network));

                // Read the data in from the file.
                StreamReader reader = new StreamReader(pathName);

                // Cast the deserialized data as a NetworkMeasurements object.
                network = (Network)deserializer.Deserialize(reader);

                // Close the connection.
                reader.Close();

                return network;
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Deserialize the Network from the Configuration File: " + exception.ToString());
            }
        }

        /// <summary>
        /// Serialized the <see cref="Networ"/> to the specified file.
        /// </summary>
        /// <param name="pathName">The directory name included the file name of the desired location for Xml Serialization.</param>
        public void SerializeToXml(string pathName)
        {
            try
            {
                // Create an XmlSerializer with the type of Network
                XmlSerializer serializer = new XmlSerializer(typeof(Network));

                // Open a connection to the file and path.
                TextWriter writer = new StreamWriter(pathName);

                // Serialize this instance of NetworkMeasurements
                serializer.Serialize(writer, this);

                // Close the connection
                writer.Close();
            }
            catch (Exception exception)
            {
                throw new Exception("Failed to Serialize the Network to the Configuration File: " + exception.ToString());
            }
        }

        public void SerializeData(bool serializeOption)
        {
            foreach (StatusWord statusWord in m_networkModel.StatusWords)
            {
                statusWord.ShouldSerializeData = serializeOption;
            }

            //foreach (BreakerStatus breakerStatus in m_breakers)
            //{
            //    breakerStatus
            //}

            foreach (VoltagePhasorGroup voltagePhasorGroup in m_networkModel.Voltages)
            {
                voltagePhasorGroup.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PositiveSequence.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PositiveSequence.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseA.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseA.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseB.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseB.Estimate.ShouldSerializeData = serializeOption;

                voltagePhasorGroup.PhaseC.Measurement.ShouldSerializeData = serializeOption;
                voltagePhasorGroup.PhaseC.Estimate.ShouldSerializeData = serializeOption;
            }

            foreach (CurrentPhasorGroup currentPhasorGroup in m_networkModel.Currents)
            {
                currentPhasorGroup.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PositiveSequence.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PositiveSequence.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseA.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseA.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseB.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseB.Estimate.ShouldSerializeData = serializeOption;

                currentPhasorGroup.PhaseC.Measurement.ShouldSerializeData = serializeOption;
                currentPhasorGroup.PhaseC.Estimate.ShouldSerializeData = serializeOption;
            }
        }
        
        #endregion

    }
}
