﻿//******************************************************************************************************
//  CircuitBreakerDataGridUserControl.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  07/15/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using NetworkModelBuilder.DataModels;
using FirstFloor.ModernUI.Windows.Controls;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Measurements;

namespace NetworkModelBuilder.DataTables
{
    /// <summary>
    /// Interaction logic for CircuitBreakerDataGridUserControl.xaml
    /// </summary>
    public partial class CircuitBreakerDataGridUserControl : UserControl
    {
        private ObservableCollection<CircuitBreakerDataModel> m_circuitBreakerData;
        private ObservableCollection<string> m_breakerStatuses = new ObservableCollection<string>();
        private ObservableCollection<string> m_parentSubstations = new ObservableCollection<string>();
        private CircuitBreakerDataModel m_editedCircuitBreaker;

        public ObservableCollection<string> BreakerStatuses
        {
            get
            {
                return m_breakerStatuses;
            }
            set
            {
                m_breakerStatuses = value;
            }
        }

        public ObservableCollection<string> ParentSubstations
        {
            get
            {
                return m_parentSubstations;
            }
            set
            {
                m_parentSubstations = value;
            }
        }


        public CircuitBreakerDataGridUserControl()
        {
            InitializeComponent();

            // Initialize the circuit breaker data models
            m_circuitBreakerData = new ObservableCollection<CircuitBreakerDataModel>();

            // Set the data context for the DataGrid.
            CircuitBreakerDataGrid.DataContext = m_circuitBreakerData;

            // Set the Event Handlers for editing
            CircuitBreakerDataGrid.CellEditEnding += CircuitBreakerDataGrid_CellEditEnding;
            CircuitBreakerDataGrid.CurrentCellChanged += CircuitBreakerDataGrid_CurrentCellChanged;

            // Set the ItemsSource for the Combo Boxes in the Data Grid
            ParentSubstationComboBox.ItemsSource = ParentSubstations;
            BreakerStatusComboBox.ItemsSource = BreakerStatuses;

            // Load the view for the first time.
            LoadCurrentCircuitBreakerView(LoadNetworkModel());
        }

        private void CircuitBreakerDataGrid_CurrentCellChanged(object sender, EventArgs e)
        {
        }

        private void CircuitBreakerDataGrid_CellEditEnding(object sender, DataGridCellEditEndingEventArgs e)
        {
            if (e.EditAction == DataGridEditAction.Commit)
            {
                m_editedCircuitBreaker = e.Row.Item as CircuitBreakerDataModel;
            }
        }

        private void AddButton_Click(object sender, RoutedEventArgs e)
        {
            m_circuitBreakerData.Add(new CircuitBreakerDataModel());
        }

        private void RefreshButton_Click(object sender, RoutedEventArgs e)
        {
            LoadCurrentCircuitBreakerView(LoadNetworkModel());
        }

        private void DeleteButton_Click(object sender, RoutedEventArgs e)
        {
            CircuitBreakerDataModel selectedItem = (CircuitBreakerDataModel)CircuitBreakerDataGrid.SelectedItem;

            if (selectedItem != null)
            {
                StringBuilder stringBuilder = new StringBuilder();
                stringBuilder.AppendFormat("Are you sure you want to delete the Circuit Breaker?{0}", Environment.NewLine);
                stringBuilder.AppendLine();
                stringBuilder.AppendFormat("InternalID: " + selectedItem.InternalID.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Number: " + selectedItem.Number.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Name: " + selectedItem.Name + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Description: " + selectedItem.Description + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Key: " + selectedItem.Key + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Breaker Status: " + selectedItem.BreakerStatusID.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Normal State: " + selectedItem.NormalState.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Parent Substation: " + selectedItem.ParentSubstation + "{0}", Environment.NewLine);

                MessageBoxButton confirmDeletionButtons = MessageBoxButton.YesNo;
                var confirmDeletionResult = ModernDialog.ShowMessage(stringBuilder.ToString(), "Confirm Deletion", confirmDeletionButtons);

                if (confirmDeletionResult.ToString().Equals("Yes"))
                {
                    Network network = LoadNetworkModel();

                    int companyIndex = 0;
                    int divisionIndex = 0;
                    int substationIndex = 0;

                    foreach (Company company in network.Model.Companies)
                    {
                        foreach (Division division in company.Divisions)
                        {
                            foreach (Substation substation in division.Substations)
                            {
                                foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
                                {
                                    if (circuitBreaker.InternalID == selectedItem.InternalID)
                                    {
                                        companyIndex = network.Model.Companies.IndexOf(company);
                                        divisionIndex = company.Divisions.IndexOf(division);
                                        substationIndex = division.Substations.IndexOf(substation);
                                    }
                                }
                            }
                        }
                    }

                    network.Model.Companies[companyIndex].Divisions[divisionIndex].Substations[substationIndex].CircuitBreakers.RemoveAll(x => x.InternalID == selectedItem.InternalID);

                    network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

                    LoadCurrentCircuitBreakerView(LoadNetworkModel());
                }
            }
        }

        private void LoadCurrentCircuitBreakerView(Network network)
        {
            m_circuitBreakerData.Clear();
            m_parentSubstations.Clear();
            m_breakerStatuses.Clear();

            m_breakerStatuses.Add("Undefined");
            foreach (BreakerStatus breakerStatus in network.Model.BreakerStatuses)
            {
                m_breakerStatuses.Add(breakerStatus.InternalID.ToString());
            }

            foreach (Company company in network.Model.Companies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        m_parentSubstations.Add(substation.ParentDivision.ParentCompany.Acronym + " - " + substation.Name);
                        foreach (CircuitBreaker circuitBreaker in substation.CircuitBreakers)
                        {
                            if (!(circuitBreaker.StatusID > 0))
                            {
                                m_circuitBreakerData.Add(new CircuitBreakerDataModel()
                                {
                                    InternalID = circuitBreaker.InternalID,
                                    Number = circuitBreaker.Number,
                                    Name = circuitBreaker.Name,
                                    Description = circuitBreaker.Description,
                                    NormalState = circuitBreaker.NormalState,
                                    Key = circuitBreaker.MeasurementKey,
                                    ParentSubstation = substation.ParentDivision.ParentCompany.Acronym + " - " + substation.Name,
                                    BreakerStatusID = "Undefined"

                                });
                            }
                            else
                            {
                                m_circuitBreakerData.Add(new CircuitBreakerDataModel()
                                {
                                    InternalID = circuitBreaker.InternalID,
                                    Number = circuitBreaker.Number,
                                    Name = circuitBreaker.Name,
                                    Description = circuitBreaker.Description,
                                    NormalState = circuitBreaker.NormalState,
                                    Key = circuitBreaker.MeasurementKey,
                                    ParentSubstation = substation.ParentDivision.ParentCompany.Acronym + " - " + substation.Name,
                                    BreakerStatusID = circuitBreaker.StatusID.ToString()

                                });
                            }
                           
                        }
                    }
                }
            }
        }

        private Network LoadNetworkModel()
        {
            Network network = Network.DeserializeFromXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);
            network.Initialize();
            return network;
        }
    }
}
