﻿//******************************************************************************************************
//  CompanyDataGridUserControl.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  07/15/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using NetworkModelBuilder.DataModels;
using FirstFloor.ModernUI.Windows.Controls;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;
using System.Configuration;

namespace NetworkModelBuilder.DataTables
{
    /// <summary>
    /// Interaction logic for CompanyDataGridUserControl.xaml
    /// </summary>
    public partial class CompanyDataGridUserControl : UserControl
    {
        private ObservableCollection<CompanyDataModel> m_companyData;
        private CompanyDataModel m_editedCompany;

        public CompanyDataGridUserControl()
        {
            InitializeComponent();
            CompanyDataGrid.CellEditEnding += CompanyDataGrid_CellEditEnding;
            CompanyDataGrid.CurrentCellChanged += CompanyDataGrid_CurrentCellChanged;
            //Bind the DataGrid to the company data
            m_companyData = new ObservableCollection<CompanyDataModel>();
            CompanyDataGrid.DataContext = m_companyData;
            LoadCurrentCompanyView(LoadNetworkModel());
        }

        private void CompanyDataGrid_CurrentCellChanged(object sender, EventArgs e)
        {
            if (m_editedCompany != null)
            {
                Network network = LoadNetworkModel();

                foreach (Company company in network.Model.Companies)
                {
                    if (company.InternalID == m_editedCompany.InternalID)
                    {
                        company.Number = m_editedCompany.Number;
                        company.Name = m_editedCompany.Name;
                        company.Acronym = m_editedCompany.Acronym;
                        company.Description = m_editedCompany.Description;
                    }
                }

                network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

                LoadCurrentCompanyView(LoadNetworkModel());
            }

            m_editedCompany = null;
        }

        private void CompanyDataGrid_CellEditEnding(object sender, DataGridCellEditEndingEventArgs e)
        {
            if (e.EditAction == DataGridEditAction.Commit)
            {
                m_editedCompany = e.Row.Item as CompanyDataModel;
            }
        }

        private void AddButton_Click(object sender, RoutedEventArgs e)
        {
            Network network = LoadNetworkModel();

            // Figure out which InternalID to use
            List<int> companyInternalIds = new List<int>();
            foreach (Company company in network.Model.Companies)
            {
                companyInternalIds.Add(company.InternalID);
            }
            int newCompanyInternalID = companyInternalIds.Max() + 1;

            network.Model.Companies.Add(new Company()
            {
                InternalID = newCompanyInternalID,
                Number = 1,
                Acronym = "CO",
                Name = "Company Name",
                Description = "Company Description",
            });

            network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

            LoadCurrentCompanyView(LoadNetworkModel());
        }

        private void RefreshButton_Click(object sender, RoutedEventArgs e)
        {
            LoadCurrentCompanyView(LoadNetworkModel());
        }

        private void DeleteButton_Click(object sender, RoutedEventArgs e)
        {
            CompanyDataModel selectedItem = (CompanyDataModel)CompanyDataGrid.SelectedItem;

            if (selectedItem != null)
            {  
                StringBuilder stringBuilder = new StringBuilder();
                stringBuilder.AppendFormat("Are you sure you want to delete the Company and all of its children Divisions?{0}", Environment.NewLine);
                stringBuilder.AppendLine();
                stringBuilder.AppendFormat("InternalID: " + selectedItem.InternalID.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Number: " + selectedItem.Number.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Acronym: " + selectedItem.Acronym + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Name: " + selectedItem.Name + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Description: " + selectedItem.Description + "{0}", Environment.NewLine);

                MessageBoxButton confirmDeletionButtons = MessageBoxButton.YesNo;
                var confirmDeletionResult = ModernDialog.ShowMessage(stringBuilder.ToString(), "Confirm Deletion", confirmDeletionButtons);

                if (confirmDeletionResult.ToString().Equals("Yes"))
                {
                    Network network = LoadNetworkModel();

                    network.Model.Companies.RemoveAll(x => x.InternalID == selectedItem.InternalID);

                    network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

                    LoadCurrentCompanyView(LoadNetworkModel());
                }
            }

        }

        private void LoadCurrentCompanyView(Network network)
        {
            // Clear the current view
            m_companyData.Clear();

            foreach (Company company in network.Model.Companies)
            {
                m_companyData.Add(new CompanyDataModel(company.InternalID, company.Number, company.Acronym, company.Name, company.Description));
            }
        }

        private Network LoadNetworkModel()
        {

            Network network = Network.DeserializeFromXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);
            network.Initialize();
            return network;
        }

        private void AddDivisionButton_Click(object sender, RoutedEventArgs e)
        {
            Network network = LoadNetworkModel();

            // Figure out which InternalID to use
            List<int> divisionInternalIds = new List<int>();
            foreach (Company company in network.Model.Companies)
            {
                foreach (Division division in company.Divisions)
                {
                    divisionInternalIds.Add(division.InternalID);
                }
            }
            int newDivisionInternalID = divisionInternalIds.Max() + 1;

            CompanyDataModel selectedItem = (CompanyDataModel)CompanyDataGrid.SelectedItem;

            if (selectedItem == null)
            {
                    ModernDialog.ShowMessage("Select a Company to add a Division", "Add Division Failed", MessageBoxButton.OK);
            }
            else
            {
                foreach (Company company in network.Model.Companies)
                {
                    if (company.InternalID == selectedItem.InternalID)
                    {
                        company.Divisions.Add(new Division()
                        {
                            InternalID = newDivisionInternalID,
                            Number = 1,
                            Acronym = "DV",
                            Name = "Division Name",
                            Description = "Division Description",
                        });
                    }
                }
            }

            network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

        }
    }
}
