﻿//******************************************************************************************************
//  NodeDataGridUserControl.cs
//
//  Copyright © 2013, Kevin D. Jones.  All Rights Reserved.
//
//  This file is licensed to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  07/15/2013 - Kevin D. Jones
//       Generated original version of source code.
//
//******************************************************************************************************
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Configuration;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using NetworkModelBuilder.DataModels;
using FirstFloor.ModernUI.Windows.Controls;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Modeling;

namespace NetworkModelBuilder.DataTables
{
    /// <summary>
    /// Interaction logic for NodeDataGridUserControl.xaml
    /// </summary>
    public partial class NodeDataGridUserControl : UserControl
    {
        private ObservableCollection<NodeDataModel> m_nodeData;
        private List<double> m_voltageLevelList = new List<double>();
        private List<string> m_parentSubstationList = new List<string>();
        private List<string> m_parentTransmissionLineList = new List<string>();

        public NodeDataGridUserControl()
        {
            InitializeComponent();
            m_nodeData = new ObservableCollection<NodeDataModel>();
            NodeDataGrid.DataContext = m_nodeData;
            VoltageLevelComboBox.ItemsSource = m_voltageLevelList;
            ParentSubstationComboBox.ItemsSource = m_parentSubstationList;
            ParentTransmissionLineComboBox.ItemsSource = m_parentTransmissionLineList;
            LoadNodeView(LoadNetworkModel());
        }

        private void AddButton_Click(object sender, RoutedEventArgs e)
        {
        }

        private void RefreshButton_Click(object sender, RoutedEventArgs e)
        {
            LoadNodeView(LoadNetworkModel());
        }

        private void DeleteButton_Click(object sender, RoutedEventArgs e)
        {
            NodeDataModel selectedItem = (NodeDataModel)NodeDataGrid.SelectedItem;

            if (selectedItem != null)
            {
                StringBuilder stringBuilder = new StringBuilder();
                stringBuilder.AppendFormat("Are you sure you want to delete the Node?{0}", Environment.NewLine);
                stringBuilder.AppendLine();
                stringBuilder.AppendFormat("InternalID: " + selectedItem.InternalID.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Number: " + selectedItem.Number.ToString() + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Name: " + selectedItem.Name + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Description: " + selectedItem.Description + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Voltage Level: " + selectedItem.VoltageLevel + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Parent Substation: " + selectedItem.ParentSubstation + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Parent Transmission Line: " + selectedItem.ParentTransmissionLine + "{0}", Environment.NewLine);
                stringBuilder.AppendFormat("Voltage: " + selectedItem.Voltage + "{0}", Environment.NewLine);

                MessageBoxButton confirmDeletionButtons = MessageBoxButton.YesNo;
                var confirmDeletionResult = ModernDialog.ShowMessage(stringBuilder.ToString(), "Confirm Deletion", confirmDeletionButtons);

                if (confirmDeletionResult.ToString().Equals("Yes"))
                {
                    Network network = LoadNetworkModel();

                    int companyIndex = 0;
                    int divisionIndex = 0;
                    int substationIndex = 0;

                    foreach (Company company in network.Model.Companies)
                    {
                        foreach (Division division in company.Divisions)
                        {
                            foreach (Substation substation in division.Substations)
                            {
                                foreach (Node node in substation.Nodes)
                                {
                                    if (node.InternalID == selectedItem.InternalID)
                                    {
                                        companyIndex = network.Model.Companies.IndexOf(company);
                                        divisionIndex = company.Divisions.IndexOf(division);
                                        substationIndex = division.Substations.IndexOf(substation);
                                    }
                                }
                            }
                        }
                    }

                    network.Model.Companies[companyIndex].Divisions[divisionIndex].Substations[substationIndex].Nodes.RemoveAll(x => x.InternalID == selectedItem.InternalID);

                    network.SerializeToXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);

                    LoadNodeView(LoadNetworkModel());
                }
            }
        }

        private Network LoadNetworkModel()
        {
            Network network = Network.DeserializeFromXml(ConfigurationManager.AppSettings["WorkingDirectory"] + "/" + ConfigurationManager.AppSettings["DefaultConfigurationFile"]);
            network.Initialize();
            return network;
        }

        private void LoadNodeView(Network network)
        {
            m_nodeData.Clear();
            m_parentSubstationList.Clear();
            m_parentSubstationList.Add("Undefined");
            m_parentTransmissionLineList.Clear();
            m_parentTransmissionLineList.Add("Undefined");
            m_voltageLevelList.Clear();

            foreach (VoltageLevel voltageLevel in network.Model.VoltageLevels)
            {
                m_voltageLevelList.Add(voltageLevel.Value);
            }

            foreach (Company company in network.Model.Companies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        m_parentSubstationList.Add(substation.ParentDivision.ParentCompany.Acronym + " - " + substation.Name);
                        foreach (Node node in substation.Nodes)
                        {
                            if (node.ParentTransmissionLine != null && node.ParentSubstation != null)
                            {
                                m_nodeData.Add(new NodeDataModel()
                                {
                                    InternalID = node.InternalID,
                                    Number = node.Number,
                                    Name = node.Name,
                                    Description = node.Description,
                                    ParentSubstation = node.ParentSubstation.ParentDivision.ParentCompany.Acronym + " - " + node.ParentSubstation.Name,
                                    ParentTransmissionLine = node.ParentSubstation.ParentDivision.ParentCompany.Acronym + " - " + node.ParentTransmissionLine.Name,
                                    Voltage = node.Voltage.Name,
                                    VoltageLevel = node.BaseKV.Value

                                });
                            }
                            else if (node.ParentTransmissionLine != null)
                            {
                                m_nodeData.Add(new NodeDataModel()
                                {
                                    InternalID = node.InternalID,
                                    Number = node.Number,
                                    Name = node.Name,
                                    Description = node.Description,
                                    ParentSubstation = "Undefined",
                                    ParentTransmissionLine = node.ParentSubstation.ParentDivision.ParentCompany.Acronym + " - " + node.ParentTransmissionLine.Name,
                                    Voltage = node.Voltage.Name,
                                    VoltageLevel = node.BaseKV.Value
                                });
                            }
                            else if (node.ParentSubstation != null)
                            {
                                m_nodeData.Add(new NodeDataModel()
                                {
                                    InternalID = node.InternalID,
                                    Number = node.Number,
                                    Name = node.Name,
                                    Description = node.Description,
                                    ParentSubstation = node.ParentSubstation.ParentDivision.ParentCompany.Acronym + " - " + node.ParentSubstation.Name,
                                    ParentTransmissionLine = "Undefined",
                                    Voltage = node.Voltage.Name,
                                    VoltageLevel = node.BaseKV.Value
                                });
                            }
                        }
                    }

                    foreach (TransmissionLine transmissionLine in division.TransmissionLines)
                    {
                    }
                }
            }
        }
    }
}
