﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Xml.Serialization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using SynchrophasorAnalytics.Measurements;
using SynchrophasorAnalytics.Modeling;
using SynchrophasorAnalytics.Islanding;
using SynchrophasorAnalytics.Networks;
using SynchrophasorAnalytics.Testing;
using SynchrophasorAnalytics.Matrices;
using SynchrophasorAnalytics.Calibration;
using SynchrophasorAnalytics.DataConditioning;

namespace SynchrophasorAnalyticsOfflineModule
{
    public partial class OfflineModule : Form
    {
        private RawMeasurements m_rawMeasurements;
        private Network m_network;

        private bool m_networkIsInitialized;
        private bool m_measurementsAreMapped;
        private bool m_activeCurrentPhasorsHaveBeenDetermined;
        private bool m_observedBussesHaveBeenResolved;
        private bool m_stateWasComputed;

        public OfflineModule()
        {
            m_rawMeasurements = null;
            m_network = null;

            m_networkIsInitialized = false;
            m_measurementsAreMapped = false;
            m_activeCurrentPhasorsHaveBeenDetermined = false;
            m_observedBussesHaveBeenResolved = false;
            m_stateWasComputed = false;

            InitializeComponent();

            DisableControls();
        }

        private void BrowseRawMeasurementsButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();

            openFileDialog.Filter = "Model Configuration Files (.xml)|*.xml";

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                RawMeasurementsTextBox.Text = openFileDialog.FileName;
            }
        }

        private void UploadRawMeasurementsButton_Click(object sender, EventArgs e)
        {
            try
            {
                XmlSerializer deserializer = new XmlSerializer(typeof(RawMeasurements));
                StreamReader reader = new StreamReader(RawMeasurementsTextBox.Text);
                m_rawMeasurements = (RawMeasurements)deserializer.Deserialize(reader);
                reader.Close();
                StatusLabel.Text = "Raw measurements loaded successfully.";
                RawMeasurementsTextBox.Enabled = false;
                BrowseRawMeasurementsButton.Enabled = false;
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void BrowseNetworkModelButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();

            openFileDialog.Filter = "Model Configuration Files (.xml)|*.xml";

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                NetworkModelTextBox.Text = openFileDialog.FileName;
            }
        }

        private void UploadNetworkModelButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network = Network.DeserializeFromXml(NetworkModelTextBox.Text);
                m_network.SerializeData(true);
                StatusLabel.Text = "Network model loaded successfully.";
                NetworkModelTextBox.Enabled = false;
                BrowseNetworkModelButton.Enabled = false;
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ClearTestCaseButton_Click(object sender, EventArgs e)
        {
            DisableControls();

            // Clear Raw Measurements 
            m_rawMeasurements = new RawMeasurements();
            RawMeasurementsTextBox.Enabled = true;
            RawMeasurementsTextBox.Text = "";
            BrowseRawMeasurementsButton.Enabled = true;

            // Clear Network Model
            m_network = null;
            m_networkIsInitialized = false;
            NetworkModelTextBox.Enabled = true;
            NetworkModelTextBox.Text = "";
            BrowseNetworkModelButton.Enabled = true;

            m_measurementsAreMapped = false;
            m_activeCurrentPhasorsHaveBeenDetermined = false;
            m_observedBussesHaveBeenResolved = false;
            m_stateWasComputed = false;

            StatusLabel.Text = "Test case cleared.";

        }

        private void InitializeNetworkModelButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.Model.Initialize();
                m_network.Model.AcceptsMeasurements = true;
                m_networkIsInitialized = true;
                StatusLabel.Text = "Network model initialized successfully.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void SerializeButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                m_network.SerializeData(true);
                m_network.SerializeToXml(saveFileDialog.FileName);
                m_network.SerializeData(false);
            }
        }

        private void MapMeasurementsButton_Click(object sender, EventArgs e)
        {
            try
            {
                foreach (RawMeasurementsMeasurement measurement in m_rawMeasurements.Items)
                {
                    m_network.Model.RawMeasurementKeyValuePairs.Add(measurement.Key, Convert.ToDouble(measurement.Value));
                }

                m_network.Model.OnNewMeasurements();

                m_measurementsAreMapped = true;

                StatusLabel.Text = "Measurements succcessfully mapped to network model.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void DetermineActiveCurrentPhasors_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.Model.DetermineActiveCurrentPhasors();
                m_activeCurrentPhasorsHaveBeenDetermined = true;
                StatusLabel.Text = "Active current phasors determined successfully.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ResolveToObservedBussesButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.Model.ResolveToObservedBusses();
                m_observedBussesHaveBeenResolved = true;
                StatusLabel.Text = "Observability analysis resolved nodes to observed nodes.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ViewActiveCurrentPhasorsButton_Click(object sender, EventArgs e)
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendFormat(m_network.Model.ActiveCurrentPhasors.Count.ToString() + " Active Current Phasors");
            foreach (CurrentPhasorGroup currentPhasorGroup in m_network.Model.ActiveCurrentPhasors)
            {
                stringBuilder.AppendFormat(currentPhasorGroup.ToVerboseString() + "{0}", Environment.NewLine);
            }

            File.WriteAllText(Directory.GetCurrentDirectory() + "/ActiveCurrentPhasors.txt", stringBuilder.ToString());
            Process.Start(Directory.GetCurrentDirectory() + "/ActiveCurrentPhasors.txt");
        }

        private void ViewMeasuredNodesButton_Click(object sender, EventArgs e)
        {
            MessageBox.Show("This button has not yet been implemented", "Oh, snap!");
        }

        private void ViewSubstationAdjacencyList_Click(object sender, EventArgs e)
        {
            StringBuilder stringBuilder = new StringBuilder();

            foreach (Company company in m_network.Model.Companies)
            {
                foreach (Division division in company.Divisions)
                {
                    foreach (Substation substation in division.Substations)
                    {
                        stringBuilder.AppendFormat(substation.ToString() + "{0}", Environment.NewLine);
                        stringBuilder.AppendFormat(substation.Graph.AdjacencyList.ToString() + "{0}", Environment.NewLine);
                        stringBuilder.AppendLine();
                    }
                }
            }

            File.WriteAllText(Directory.GetCurrentDirectory() + "/SubstationAdjacencyLists.txt", stringBuilder.ToString());
            Process.Start(Directory.GetCurrentDirectory() + "/SubstationAdjacencyLists.txt");
        }

        private void ViewObservableNodesButton_Click(object sender, EventArgs e)
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendFormat(m_network.Model.ObservedBusses.Count.ToString() + " Observed Busses");
            foreach (ObservedBus observedBus in m_network.Model.ObservedBusses)
            {
                stringBuilder.AppendFormat(observedBus.ToVerboseString() + "{0}", Environment.NewLine);
            }

            File.WriteAllText(Directory.GetCurrentDirectory() + "/ObservedBusses.txt", stringBuilder.ToString());
            Process.Start(Directory.GetCurrentDirectory() + "/ObservedBusses.txt");
        }

        private void ViewOutputMeasurementsButton_Click(object sender, EventArgs e)
        {
            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendFormat(m_network.Model.RawEstimateKeyValuePairs.Count.ToString() + " Output Measurements");

            foreach (KeyValuePair<string, double> outputMeasurement in m_network.Model.RawEstimateKeyValuePairs)
            {
                stringBuilder.AppendFormat(outputMeasurement.Key + "\t" + outputMeasurement.Value.ToString() + "{0}", Environment.NewLine);
            }

            File.WriteAllText(Directory.GetCurrentDirectory() + "/OutputMeasurements.txt", stringBuilder.ToString());
            Process.Start(Directory.GetCurrentDirectory() + "/OutputMeasurements.txt");

        }

        private void AMatrixButton_Click(object sender, EventArgs e)
        {
            try
            {
                CurrentMeasurementBusIncidenceMatrix A = new CurrentMeasurementBusIncidenceMatrix(m_network);

                StringBuilder stringBuilder = new StringBuilder();

                stringBuilder.AppendFormat(A.ToCsvString());

                File.WriteAllText(Directory.GetCurrentDirectory() + "/A Matrix.csv", stringBuilder.ToString());
                Process.Start(Directory.GetCurrentDirectory() + "/A Matrix.csv");
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void IIMatrixButton_Click(object sender, EventArgs e)
        {
            try
            {
                VoltageMeasurementBusIncidenceMatrix II = new VoltageMeasurementBusIncidenceMatrix(m_network);

                StringBuilder stringBuilder = new StringBuilder();

                stringBuilder.AppendFormat(II.ToCsvString());

                File.WriteAllText(Directory.GetCurrentDirectory() + "/II Matrix.csv", stringBuilder.ToString());
                Process.Start(Directory.GetCurrentDirectory() + "/II Matrix.csv");
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void YMatrixButton_Click(object sender, EventArgs e)
        {
            try
            {
                SeriesAdmittanceMatrix Y = new SeriesAdmittanceMatrix(m_network);

                StringBuilder stringBuilder = new StringBuilder();

                stringBuilder.AppendFormat(Y.ToCsvString());

                File.WriteAllText(Directory.GetCurrentDirectory() + "/Y Matrix.csv", stringBuilder.ToString());
                Process.Start(Directory.GetCurrentDirectory() + "/Y Matrix.csv");
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void YsMatrixButton_Click(object sender, EventArgs e)
        {
            try
            {
                ShuntSusceptanceMatrix Ys = new ShuntSusceptanceMatrix(m_network);

                StringBuilder stringBuilder = new StringBuilder();

                stringBuilder.AppendFormat(Ys.ToCsvString());

                File.WriteAllText(Directory.GetCurrentDirectory() + "/Ys Matrix.csv", stringBuilder.ToString());
                Process.Start(Directory.GetCurrentDirectory() + "/Ys Matrix.csv");
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ComputeSystemStateButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.ComputeSystemState();
                m_stateWasComputed = true;
                StatusLabel.Text = "System state computed successfully.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ComputeLineFlowsButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.Model.ComputeEstimatedCurrentFlows();
                StatusLabel.Text = "Estimated line flows computed successfully.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void ComputePowerFlowsButton_Click(object sender, EventArgs e)
        {
            MessageBox.Show("This button has not yet been implemented", "Oh, snap!");
        }

        private void ComputeSequenceComponentsButton_Click(object sender, EventArgs e)
        {
            try
            {
                m_network.Model.ComputeSequenceValues();
                StatusLabel.Text = "Sequence values computed successfully.";
                EnableControls();
            }
            catch (Exception exception)
            {
                MessageBox.Show(exception.ToString(), "Error!");
            }
        }

        private void DisableControls()
        {
            InitializeNetworkModelButton.Enabled = false;
            SerializeButton.Enabled = false;
            MapMeasurementsButton.Enabled = false;
            DetermineActiveCurrentPhasorsButton.Enabled = false;
            ResolveToObservedBussesButton.Enabled = false;
            ViewActiveCurrentPhasorsButton.Enabled = false;
            ViewMeasuredNodesButton.Enabled = false;
            ViewObservableNodesButton.Enabled = false;
            ViewSubstationAdjacencyList.Enabled = false;
            ViewOutputMeasurementsButton.Enabled = false;
            AMatrixButton.Enabled = false;
            IIMatrixButton.Enabled = false;
            YMatrixButton.Enabled = false;
            YsMatrixButton.Enabled = false;
            ComputeSystemStateButton.Enabled = false;
            ComputeLineFlowsButton.Enabled = false;
            ComputePowerFlowsButton.Enabled = false;
            ComputeSequenceComponentsButton.Enabled = false;
        }

        private void EnableControls()
        {
            if (m_rawMeasurements != null && m_network != null)
            {
                InitializeNetworkModelButton.Enabled = true;
            }
            if (m_networkIsInitialized)
            {
                SerializeButton.Enabled = true;
                MapMeasurementsButton.Enabled = true;
            }
            if (m_measurementsAreMapped)
            {
                DetermineActiveCurrentPhasorsButton.Enabled = true;
            }
            if (m_activeCurrentPhasorsHaveBeenDetermined)
            {
                ResolveToObservedBussesButton.Enabled = true;
                ViewActiveCurrentPhasorsButton.Enabled = true;
            }
            if (m_observedBussesHaveBeenResolved)
            {
                ViewMeasuredNodesButton.Enabled = true;
                ViewObservableNodesButton.Enabled = true;
                ViewSubstationAdjacencyList.Enabled = true;
                AMatrixButton.Enabled = true;
                IIMatrixButton.Enabled = true;
                YMatrixButton.Enabled = true;
                YsMatrixButton.Enabled = true;
                ComputeSystemStateButton.Enabled = true;
            }
            if (m_stateWasComputed)
            {
                ComputeLineFlowsButton.Enabled = true;
                ComputePowerFlowsButton.Enabled = true;
                ViewOutputMeasurementsButton.Enabled = true;
                if (m_network.PhaseConfiguration == PhaseSelection.ThreePhase)
                {
                    ComputeSequenceComponentsButton.Enabled = true;
                }
            }
        }


    }
}
